#include<iostream>
#include<vector>
#include<string>
#include<fstream>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <error.h>
#include <netinet/tcp.h>
#include <sys/stat.h>

#include "log.h"
#include "debug.h"
#include "FileTransfer.h"
#include "BootStrapServer.h"
#include "digiSig.h"
#include "EventMessage.h"
#include "util.h"

#define MAXCHAR 1500
#define SOCKET_ERROR -1
#define POLL_TIMEOUT (4*1000)


CBootStrapServer::CBootStrapServer(const std::string& sConfigFile):

	m_sConfigFile(sConfigFile),
	m_iDaemonSockFd(SOCKET_ERROR),
	m_uServerIP(0),
        m_uServerPort(0),
	m_sServerPrivKey(),
	m_sServerCert(),
	m_sLogFile(),
	m_sNodeFile(),
	m_sSignedNodeFile(),
	m_vNodes(),
	m_bRun(true){
}

CBootStrapServer::~CBootStrapServer(){

	m_bRun = false;
	Stop();
	for(int i=0;i<m_filetrans.size();i++){
		delete m_filetrans[i];
	}
	DestroySocket();
}

int CBootStrapServer::Start(){

	if(-1 == ParseConfig()){
		return -1;
	}

	LOG_OPEN(m_sLogFile);

	if(-1 == CreateDaemonSocket()){
		return -1;
	}
	if(-1 == InitSocketPair()){
		return -1;
	}
	if(-1 == LoadNodes()){
		return -1;
	}

	if(SignNodeFile()==-1)
		return -1;

	Run();

return 0;
}

int CBootStrapServer::Stop()const{

	DB_AND_LOG("Closing Server..");
	//StopPollLoop();	
	LOG_CLOSE();
return 0;
}

int CBootStrapServer::Run(){


	struct pollfd pollFd[2];

	pollFd[0].fd = m_iDaemonSockFd;
	pollFd[0].events = POLLIN;

	pollFd[1].fd = m_iSockPairFd[0];
	pollFd[1].events = POLLIN;


	while (m_bRun){
		int iStatus = poll(pollFd, 2, POLL_TIMEOUT );
		if( iStatus < 0){
			cout<<"poll failed in main loop"<<iStatus<<endl;;
		}
		else if( 0 == iStatus){
#if 0
		//	DB("Timeout in poll");
			if(SignNodeFile()==-1){
				DB_AND_LOG("Error in signing node file");
			}
#endif
		}
		else if(pollFd[0].revents & POLLIN){
			HandleConnection(pollFd[0].fd);
		} 
		else if(pollFd[1].revents & POLLIN){
			DB("in sock pair");
			if(0!=HandleNotification(pollFd[1].fd)){
				DB("msg for shutdown!");
				break;
			}
		}
		else{
			DB_AND_LOG("error in poll");
			break;
		}
	    
	}//while loop

return 0;
}


int CBootStrapServer::HandleConnection(int iFd){

	struct sockaddr_in FromAddr;
	socklen_t FromAddrLen=sizeof(FromAddr);

	int iClientSock = accept(iFd, (struct sockaddr *) &FromAddr, &FromAddrLen);
	if(iClientSock<0){
		DB_AND_LOG("Error in accept");
	}
	else{
		uint32_t uIP = ntohl(FromAddr.sin_addr.s_addr);
		uint16_t uPort = ntohs(FromAddr.sin_port);
		DB_AND_LOG("Request for a BootStrap Node by: "<< GetDotFormatIP(uIP)<<":"<< uPort);
	}	

	CFileTransfer* filetrans = new CFileTransfer(iClientSock,"","");
	filetrans->SetSocketPair(m_iSockPairFd[1]);

	filetrans->HandShake(false,true,m_sSignedNodeFile, GetFileSize(m_sSignedNodeFile));
	m_filetrans.push_back(filetrans);
	printf("file ptr=%p\n",filetrans);

return 0;
}

int CBootStrapServer::HandleNotification(int iFd){
	//we need to delete the file transfer pointer..
	DB_AND_LOG("Notification arrived in fd="<<iFd);
	int iRet = 0;
	char szBuf[MAXCHAR];

        int iReceived = recv(iFd, (void*)szBuf, MAXCHAR, 0 );
        if(0== iReceived ){
                return 0;
        }
        int iRead = 0;
        while(iRead<iReceived){

                const char* pBuf = szBuf;
                pBuf+=iRead;

                EventMessage msg;
                std::string sBuf(pBuf,iReceived-iRead);
                iRead += msg.deserialize(sBuf);

                std::cout<<"iRead="<<iRead<<std::endl;

                switch(msg.GetStatus()){
			case EVT_SUCCESS:
			{
					DB("Success received in Notification");
					const CFileTransfer* filetrans = reinterpret_cast<CFileTransfer*>(msg.GetSearchID());
					SearchAndDeleteFileObj(filetrans);
					break;
			}
			case EVT_ERROR:
					DB("Error received in Notification");
					break;
			case EVT_SHUTDOWN:
					DB("Stopping down the loop");
					iRet = -1;
					break;
			default:
					DB("Unknown status code received in Notification");
					break;
		}
	}//while



return iRet;
}

int CBootStrapServer::CreateDaemonSocket(){
	struct sockaddr_in kAddr;
	int iRet;

	m_iDaemonSockFd = ::socket(AF_INET, SOCK_STREAM, 0);
	if (m_iDaemonSockFd == -1){
		DB_AND_LOG("Could not create tcp socket.");
		return -1;
	}

	reuse(m_iDaemonSockFd);

	bzero((char *) &kAddr, sizeof(kAddr));
	kAddr.sin_family = AF_INET;
	kAddr.sin_addr.s_addr = htonl(m_uServerIP); 
	kAddr.sin_port = htons(m_uServerPort);

printf("ip=%x\n", kAddr.sin_addr.s_addr);
printf("port=%d\n", kAddr.sin_port);

	iRet = bind(m_iDaemonSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr));
	if(iRet<0){
		DB_AND_LOG("ERROR on binding in tcp sock");
		return -1;
	}

	/*Listen for maximum Ten client*/
	iRet = listen(m_iDaemonSockFd, 10) ;
	if(iRet < 0 ){
		DB_AND_LOG("Error in Listen");
		return -1;
	}

return 0;
}

int CBootStrapServer::LoadNodes(){

	char szBuf[MAXCHAR];

	std::ifstream fileHandler(m_sNodeFile.c_str());
	if(!fileHandler.is_open()){
		DB_AND_LOG("Unable to open Input file"<<m_sNodeFile);
		return -1;
	}

	do{
		bzero(szBuf,MAXCHAR);
		fileHandler.getline(szBuf,MAXCHAR);
		std::string sTemp(szBuf);
		StrTrim(sTemp);
		if(sTemp.size()!=0){
			uint32_t ip = ntohl( inet_addr(sTemp.c_str())); //keep in host byte order
			m_vNodes.push_back(ip);
			DB("ip="<<ip);
		}

	}while(!fileHandler.eof());
	fileHandler.close();

return 0;
}

int CBootStrapServer::InitSocketPair(){

	int rc = socketpair( AF_UNIX, SOCK_STREAM, 0, m_iSockPairFd );
	if ( rc < 0 ) {
		DB_AND_LOG("Cannot open socketpair");
		return -1;
	}

return 0;
}

int CBootStrapServer::ParseConfig(){

	std::vector<std::string> vConf;
	char szBuf[MAXCHAR];
	std::string sTemp;

	std::ifstream fileHandler(m_sConfigFile.c_str());
	if(!fileHandler.is_open()){
		std::cerr<<"Unable to open Input file"<<m_sConfigFile<<std::endl;
		return -1;
	}

	do{
		bzero(szBuf,MAXCHAR);
		fileHandler.getline(szBuf,MAXCHAR);
		vConf.push_back(szBuf);
	}while(!fileHandler.eof());
	fileHandler.close();


	//Now, parse config file
        sTemp = GetString(vConf, "server_ip") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_uServerIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

	sTemp = GetString(vConf, "server_port") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_uServerPort = atoi(sTemp.c_str());

	sTemp = GetString(vConf, "server_private_key") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_sServerPrivKey = sTemp;

	sTemp = GetString(vConf, "server_certificate") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_sServerCert = sTemp;

	sTemp = GetString(vConf, "log_file") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_sLogFile = sTemp;

	sTemp = GetString(vConf, "node_file") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_sNodeFile = sTemp;
	m_sSignedNodeFile = "signed_"+m_sNodeFile;

return 0;
}

void CBootStrapServer::StrTrim(std::string& str){

	std::string::size_type pos = str.find_last_not_of(' ');
	if(pos != std::string::npos) {
		str.erase(pos + 1);
		pos = str.find_first_not_of(' ');
		if(pos != std::string::npos)
			str.erase(0, pos);
		}
		else
			str.erase(str.begin(), str.end());
}

void CBootStrapServer::DestroySocket()
{
	if(m_iDaemonSockFd != SOCKET_ERROR){
		::close(m_iDaemonSockFd);
		m_iDaemonSockFd = SOCKET_ERROR;
	}
	if(m_iSockPairFd[0] != SOCKET_ERROR){
		::close(m_iSockPairFd[0]);
		m_iSockPairFd[0] = SOCKET_ERROR;
	}
	if(m_iSockPairFd[1] != SOCKET_ERROR){
		::close(m_iSockPairFd[1]);
		m_iSockPairFd[1] = SOCKET_ERROR;
	}

}

int CBootStrapServer::GetFileSize(const std::string& sFileName)const{

	struct stat fStat;
	stat( sFileName.c_str(), &fStat );
		return fStat.st_size;

}

std::string CBootStrapServer::GetString(const std::vector<std::string>& vConf, const std::string& sLine)const{

	std::vector<std::string>::const_iterator vIter;
	std::string sTemp;
	int iPos=-1;

	for(vIter=vConf.begin(); vIter!=vConf.end(); vIter++){
		sTemp = *vIter;
		iPos = sTemp.find(sLine);
		if(iPos == std::string::npos)
			continue;

		iPos = sTemp.find("=");
		return sTemp.substr(iPos+1);
	}

return "";
}

int CBootStrapServer::SignNodeFile(){


	CDigiSig obj;
#if 0
        /*Now, generate the smime file*/
        /*We put a random uuid as an unique transaction in subject field*/
        boost::uuids::uuid Tag = boost::uuids::random_generator()();
        std::stringstream ss;
        ss<<Tag;
        std::string sSub;
        ss>>sSub;
        DB("subject="<<sSub);
#endif

	std::string sTo("Client");
	std::string sServerName("BootStrapping Server");
	std::string sSub("Node list");

        int iRet = obj.SignSMIME(m_sServerCert, m_sServerPrivKey, m_sNodeFile, m_sSignedNodeFile, sTo, sServerName, sSub);
        if(iRet!=0){
                LOG("Signature failed");
                return -1;
        }

return 0;
}

void CBootStrapServer::SearchAndDeleteFileObj(const CFileTransfer* obj){

	FileTransfer::iterator vIter;	
	for(vIter=m_filetrans.begin();vIter!=m_filetrans.end();vIter++){
		if(*vIter == obj ){
			DB("Obj found in the vector");
			delete obj;
			m_filetrans.erase(vIter);
			break;
		}
	}
}

void CBootStrapServer::StopPollLoop()const{

        EventMessage msg;
        msg.SetStatus(EVT_SHUTDOWN);

        std::string sBuf;
        msg.serialize(sBuf);

        int iStatus = ::send(m_iSockPairFd[1], sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                LOG("Failed sending shutdown command in sockpair");
        }
	DB("shutdown msg sent");
}


