/*
Author: Md Sazzadur Rahman
rahmanm@cs.ucr.edu
*/

#include <iostream>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <error.h>

#include "FileTransfer.h"
#include "EventMessage.h"
#include "util.h"


#define SOCKET_ERROR 	-1
#define POLL_TIMEOUT	1000
#define MAX_TIMEOUT	5


CFileTransfer::CFileTransfer(int iSockFd, const std::string& sSharedDir, const std::string& sDownloadDir, int iSearchID):
	m_iSockFd(iSockFd),
	m_sSharedDir(sSharedDir),
	m_sDownloadDir(sDownloadDir),
	m_iSearchID(iSearchID),
	m_8FileType(0),
	m_bCompleted(false),
	m_bPassive(false),
	m_bPush(false)
{

}


CFileTransfer::~CFileTransfer()
{

}

int CFileTransfer::HandShake(bool bPassive, bool bPush, const std::string& sFileName, int iFileSize){

int iRet = -1;
	if( !sFileName.empty() && iFileSize!=0){
		m_sFileName = sFileName;
		m_iFileSize = iFileSize;
	}
	m_bPassive=bPassive;
	m_bPush=bPush;

	iRet = StartFileTransferThread();
	if(iRet!=0){
		CloseSocket();
		return -1;
	}

return 0;
}

int CFileTransfer::HandShake(bool bPassive, bool bPush, int iSockFd){

	m_iSockFd = iSockFd;
	if(m_sFileName.empty()){
		std::cout<<"file name is empty\n ";
		CloseSocket();
		return -1;
	}

	return HandShake(bPassive, bPush, m_sFileName, m_iFileSize);
}

int CFileTransfer::StartFileTransferThread(){

        pthread_attr_t threadAttr;
        if( pthread_attr_init(&threadAttr) != 0 )
                std::cout<<"Error in pthread_attr_init\n";

        if( pthread_attr_setdetachstate(&threadAttr, PTHREAD_CREATE_DETACHED) != 0 )
                std::cout<<"Error in pthread_attr_setdetachstate\n";

        int iRet = pthread_create(&m_pTID,&threadAttr,CFileTransfer::FileTransferThread, this);
        if (iRet){
                 std::cout<<"ERROR; return code from pthread_create() is "<<iRet<<std::endl;
                 perror("error in creating receiver thread:");
                 return -1;
        }

        if( pthread_attr_destroy(&threadAttr) != 0 )
                std::cout<<"Error in pthread_attr_destroy\n";

return 0;
}
void* CFileTransfer::FileTransferThread(void* arg){


        CFileTransfer& filetrans = *(CFileTransfer*)arg;

	filetrans.StartFileTransfer();
	filetrans.Cleanup();

return NULL;
}

int CFileTransfer::StartFileTransfer(){

int iRet=-1;

        if(m_bPassive){
                //wait for handshake packet
                //depending on the packet, start sender or receiver thread..
                bool bPull;
                std::cout<<"Waiting for handshake packet : sock="<<m_iSockFd<<std::endl;
                iRet = WaitForHandShakePacket(bPull);
                if(iRet!=0){
                        std::cout<<"Failed to receive handshake packet \n";
                        return iRet;
                }

                if(bPull){
                        //send ACK for Handshake pkt: sender should start after that..
                        std::cout<<"Sending READY packet :sock="<<m_iSockFd<<std::endl;
                        SendPacket(m_iSockFd, "READY", 5 );

                        std::cout<<"Going to start Receiver thread \n";
                        ReceiveFile();
                }
                else{
                        std::cout<<"Going to start Sender thread \n";
                        SendFile();
                }
        }

        else{
                //send handshake packet
                std::cout<<"Sending handshake packet :sock="<<m_iSockFd<<std::endl;
                iRet = SendHandShakePacket(m_bPush);
                if(iRet!=0){
                        std::cout<<"Failed to send handshake packet \n";
                        return iRet;
                }

                //depending on push or pull, start sender or receiver thread..
                if(m_bPush){
                        //should wait for Handshake ACK
                        iRet = WaitForHandShakePacketAck();
                        if(iRet!=0){
                                std::cout<<"Failed to receive handshake packet ack \n";
                                return iRet;
                        }
                        //wants to disseminate gossip packet
                        std::cout<<"Going to start Sender thread \n";
                        SendFile();
                }
                else{
                        //tries to get some file from network
                        std::cout<<"Going to start Receiver thread \n";
                        ReceiveFile();
                }
        }

return 0;
}


void CFileTransfer::ReceiveFile(){

int iSockFd = -1;
int iReceived = -1;
uint8_t szBuffer[MAXSIZE];

	iSockFd = GetSockFd();
	std::string sFileName = GetFileName();

	std::string sPath;
	FILE* pFile;

	switch(GetFileType()){

		case FILE_GOSSIP:
				sPath = GetDownloadDir()+"/"+sFileName;
				break;
		case FILE_OTHER:
				sPath = GetSharedDir()+"/"+sFileName;
				break;
		default:
				std::cout<<"Unknown file type:"<<std::endl;
				break;
	}
	SetFilePath(sPath);

	pFile = fopen(sPath.c_str(), "wb");
	if( pFile == NULL){
		std::cout<<"Unable to open file for write\n";
		return ;
	}

	unsigned int iFileSize = 0;
	bool bFlag = true;
	bool bNotify = false;
	int iCounter = 0;
#if 0
	//do we have m_sExtraData?
	const std::string& sExtraData = filetrans.GetExtraData();
	if(!sExtraData.empty()){
		fwrite(sExtraData.c_str(),1,sExtraData.size(),pFile);
		iFileSize += sExtraData.size();
		if(iFileSize>=filetrans.GetFileSize()){
			filetrans.SendPacket(iSockFd, "End", 3 );	
			bNotify = true;
			bFlag = false;
		}
	}
#endif
        struct pollfd pollFd[1];
        pollFd[0].fd = iSockFd;
        pollFd[0].events = POLLIN;


	//std::cout<<"File Size="<<filetrans.GetFileSize()<<std::endl;

	//wait for whole data file and receive
	while(bFlag){

                int iStatus = poll(pollFd, 1, POLL_TIMEOUT );

                if( iStatus < 0){
			std::cout<<"Poll failed in receiver thread\n";
                }
                switch (pollFd[0].revents) {

                        case POLLIN:
				bzero(szBuffer, MAXSIZE);
				iReceived = recv(iSockFd, (void*)szBuffer, MAXSIZE, 0 );
				if(0==iReceived){
					std::cout<<"Peer is down: shutting down the receiver\n";
					bFlag = false;
					break;
				}
				//std::cout<<"received data: "<<std::string((char*)szBuffer,iReceived)<<std::endl;
                                fwrite(szBuffer,1,iReceived,pFile);
				iFileSize+=iReceived;

				//std::cout<<"iFileSize="<<iFileSize<<std::endl;
				if(iFileSize>=GetFileSize()){
					//std::cout<<"Whole File Received: sending End"<<std::endl;
					SendPacket(iSockFd, "End", 3 );	
					bNotify = true;
					bFlag = false;
				}
                                break;

                        case 0:
				std::cout<<"Poll timeout in receiver thread: sock="<<iSockFd<<std::endl;
				iCounter++;
				if(iCounter>MAX_TIMEOUT){
					//std::cout<<"Waited too long: terminating..receiver"<<std::endl;
					bFlag = false;
				}
			
                                break;

                        case -1:
				std::cout<<"Poll returned error in receiver thread\n";
                                break;

                }

	}	
	
	fclose(pFile);
	if(bNotify){
		NotifyComplete(true);
	}
	else{
		NotifyComplete(false);
	}

std::cout<<"Receiver terminating\n";

}

void CFileTransfer::SendFile(){

int iReceived;
int iSockFd=-1;
FILE* pFile=NULL;
uint8_t szBuffer[MAXSIZE];

	iSockFd = GetSockFd();
	std::string sFileName = GetFileName();

	//waits for filename and then starts sends

	std::string sPath;
        switch(GetFileType()){

                case FILE_GOSSIP:
                                sPath = sFileName;
                                break;
                case FILE_OTHER:
                                sPath = GetSharedDir()+"/"+sFileName;
                                break;
                default:
                                std::cout<<"Unknown file type:"<<std::endl;
                                break;
        }

	pFile = fopen(sPath.c_str(), "rb");
	if( pFile == NULL){
		std::cout<<"Unable to open file for read: "<<sPath<<std::endl;
		return ;
	}

//now, send the file
        while (1){
		bzero(szBuffer, MAXSIZE);

		int iReadSize=fread(szBuffer,1,MAXSIZE,pFile);
		if(iReadSize<=0){
			break;
		}
		//std::cout<<"Sending PAYLOAD: "<<iReadSize<<std::endl;
		std::cout<<"sending payload \n";
		SendPacket(iSockFd, (char*)szBuffer, iReadSize );
        }

	//wait for ack
	struct pollfd pollFd[1];
        pollFd[0].fd = iSockFd;
        pollFd[0].events = POLLIN;

	bool bFlag = true;
	int iCounter = 0;
        while (bFlag){
                int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
                if( iStatus < 0){
                        std::cout<<"poll failed in sender thread"<<std::endl;;
			std::cout<<"Poll failed \n";
                }

                switch (pollFd[0].revents) {

                        case POLLIN:
				iReceived = recv(iSockFd, (void*)szBuffer, MAXSIZE, 0 );
				std::cout<<"Received waiting packet \n";
				bFlag = false;
                                break;

                        case 0:
				std::cout<<"poll timeout in sender thread: sock="<<iSockFd<<std::endl;
				iCounter++;
				if(iCounter>MAX_TIMEOUT){
					std::cout<<"Waited too long: terminating..sender \n";
					bFlag = false;
				}
                                break;

                        case -1:
				std::cout<<"Poll returned error\n";
                                break;

                }
        }


	fclose(pFile);

std::cout<<"Sender is terminating\n";

}


int CFileTransfer::SendPacket(int iSockFd, const char* szPayload, int iSize){

        int iRet = ::send(iSockFd, (void*)szPayload, iSize , 0);

        if(iRet < 0){
		std::cout<<"Send failed in file transfer\n";
	}

return 0;
}

void CFileTransfer::Cleanup(){

	m_bCompleted = true;
	CloseSocket();


	//int iStatus = 0;
	//pthread_exit((void*)&iStatus);
	pthread_exit(0);

}

void CFileTransfer::CloseSocket(){

	if(m_iSockFd != SOCKET_ERROR){
              ::close(m_iSockFd);
              m_iSockFd = SOCKET_ERROR;
	}
}

int CFileTransfer::ConnectPeer(uint32_t iRemoteIP, uint16_t iRemotePort){

struct sockaddr_in kAddr;

int iRet;
int iSockFd;

        iSockFd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (iSockFd == -1){
                std::cout<<"Could not create socket."<<std::endl;
                return -1;
        }
#if 0//for testing only..this bind
        bzero((char *) &kAddr, sizeof(kAddr));
        kAddr.sin_family = AF_INET;
        kAddr.sin_addr.s_addr = INADDR_ANY;
        kAddr.sin_port = htons(0);

        if (bind(iSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr)) < 0) {
                std::cout<<"ERROR on binding in connectPeer in file transfer sock"<<std::endl;
                return -1;
        }

#endif
	std::cout<<"Connected port="<<ntohs(GetSockName(iSockFd).second)<<std::endl;

        //connect
        struct sockaddr_in destAddr;
        socklen_t addrLen = sizeof(destAddr);
        bzero(&destAddr, sizeof(destAddr));
        destAddr.sin_family = AF_INET;
        destAddr.sin_addr.s_addr = htonl(iRemoteIP);
        destAddr.sin_port = htons(iRemotePort);

        iRet = ::connect(iSockFd, (struct sockaddr *) &destAddr, addrLen) ;
        if (iRet < 0) {
                std::cout<<"ERROR on connect"<<std::endl;
                return -1;
        }
        else{
		std::cout<<"connect successfully\n";
	}

return iSockFd;
}

/*Notify with EventMessage*/
void CFileTransfer::NotifyComplete(bool bSuccess){
	
	EventMessage msg;
	if(bSuccess){
		msg.SetStatus(EVT_SUCCESS);
	}
	else{
		msg.SetStatus(EVT_ERROR);
	}

	msg.SetFileSize(m_iFileSize);
	msg.SetFileName(m_sFileName);
	msg.SetFilePath(m_sFilePath);
	msg.SetSearchID(m_iSearchID);
	msg.SetFileType(m_8FileType);

	switch(m_8FileType){
	
	case FILE_GOSSIP:
			msg.SetRelatedModule(MODULE_GOSSIP);
			break;
	case FILE_OTHER:
			msg.SetRelatedModule(MODULE_KAD);
			break;
	default:
			break;
	}

	std::string sBuf;
	msg.serialize(sBuf);

	std::cout<<"Notification for file:  "<<m_sFileName<<std::endl;

        int iStatus = ::send(GetSocketPair(), sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                std::cout<<"Failed sending complete notification to kademlia\n";
                perror("error in notify complete..");
        }

}


int CFileTransfer::SendHandShakePacket(bool bPush){

        EventMessage msg;

	msg.SetStatus(EVT_SUCCESS);
	if(bPush){
		m_8FileType = FILE_GOSSIP;

		msg.SetReasonCode(FILE_RECV_REQ);
		msg.SetFileType(FILE_GOSSIP);
	}
	else{
		m_8FileType = FILE_OTHER;

		msg.SetReasonCode(FILE_SEND_REQ);
		msg.SetFileType(FILE_OTHER);
	}

	//send only the name part of the file, not the path
        std::string sName  = FileNameFromPath(m_sFileName);


        msg.SetFileSize(m_iFileSize);
        msg.SetFileName(sName);
        msg.SetSearchID(-1);

        std::string sBuf;
        msg.serialize(sBuf);

        std::cout<<"Sending HandShakePacket file:  "<<m_sFileName<<std::endl;

        int iStatus = ::send(m_iSockFd, sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                std::cout<<"Failed sending HandShakePacket\n";
                perror("error in send in HandShakePacket");
		return -1;
        }

return 0;
}

int CFileTransfer::WaitForHandShakePacketAck(){
	bool bPull;
	return WaitForHandShakePacket(bPull, true);
}

int CFileTransfer::WaitForHandShakePacket(bool& bPull, bool bWaitForAck){

int iReceived = -1;
uint8_t szBuffer[MAXSIZE];
struct pollfd pollFd[1];
pollFd[0].fd = m_iSockFd;
pollFd[0].events = POLLIN;
bool bFlag = true;
int iCounter=0;


	while(bFlag){

		int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
		if( iStatus < 0){
			std::cout<<"Poll failed in in waithandshakedata\n";
			return -1;
		}

		switch (pollFd[0].revents) {
		case POLLIN:
		{
			bzero(szBuffer, MAXSIZE);
			iReceived = recv(m_iSockFd, (void*)szBuffer, MAXSIZE, 0 );
			const char* pBuf = (char*)szBuffer;

			if(0==iReceived){
				std::cout<<"Failed to receive data..peer is down\n";
				return -1;
			}

			if(iReceived<0) {
				std::cout<<"Failed to refceived data \n";
				return -1;
			}
			else{
				bFlag = false;
				if(bWaitForAck){
					std::cout<<"Received ACK data: "<<iReceived<<std::endl;
					//should we check the payload?
					if(strncmp(pBuf,"READY",5)!=0){
						std::cout<<"HandShake ACK received does not match with READY! \n";
						return -1;
					}
				}
				else{
					if(ParseHandShakePacket((char*)szBuffer, iReceived, bPull) !=0 )
						return -1;
				}
			}
			break;

		}
		case 0:
				if(bWaitForAck)
					std::cout<<"Poll timeout in waitforpktACK; sock="<<m_iSockFd<<std::endl;
				else
					std::cout<<"Poll timeout in waitforpkt; sock="<<m_iSockFd<<std::endl;
                                iCounter++;
                                if(iCounter>MAX_TIMEOUT*10){
					if(bWaitForAck)
						std::cout<<"Waited too long: terminating..waitforpktACK: port="<<m_iSockFd<<std::endl;
					else
						std::cout<<"Waited too long: terminating..waitforpkt; port="<<m_iSockFd<<std::endl;
                                        //bFlag = false;
					return -1;
                                }
                                break;

		case -1:
                                std::cout<<"Poll returned error\n";
				return -1;

		}//switch
	}//while

return 0;
}

int CFileTransfer::ParseHandShakePacket(const char* szBuffer, int iReceived, bool& bPull){

	int iRead = 0;
	const char* pBuf = szBuffer;
	pBuf+=iRead;

	EventMessage msg;
	std::string sBuf(pBuf,iReceived-iRead);
	iRead += msg.deserialize(sBuf);

	m_sFileName = msg.GetFileName();
	m_iFileSize = msg.GetFileSize();
	m_8FileType = msg.GetFileType();
	

	switch(msg.GetReasonCode()){

	case  FILE_SEND_REQ:
			/*Peer wants me to send him a file*/
			bPull = false;
			break;
	case FILE_RECV_REQ:
			/*Peer wants me to receive a file from him*/
			bPull = true;
			break;
	default:
		std::cout<<"Invalid Reason Code in HandShake Packet"<<std::endl;
		return -1;
	}

	if(iRead != iReceived){
		std::cout<<"Received extra data \n";
		 std::cout<<"Received extra data: "<<iRead<<std::endl;
		pBuf +=iRead;
		m_sExtraData = std::string(pBuf, iReceived-iRead);
		std::cout<<"Extra data="<<m_sExtraData<<std::endl;
	}

return 0;
}
std::string CFileTransfer::FileNameFromPath(const std::string& sPath)const{

          size_t pos = sPath.find_last_of('/');
          if(pos==std::string::npos)
                  return sPath;

return sPath.substr(pos+1);
}

