#ifndef FILE_TRANSFER_H
#define FILE_TRANSFER_H

#ifdef CLIENT
#include "MD4Hash.h"
#endif

#define MAXSIZE         1024
#if 0
#define FILEREQ		0xA2
#define FILERES		0xA3
#define FILEFINISH	0xA4
#define FILEERR		0xA5
#define FILEFINISHACK	0xA6

#define FILE_SEND_REQ	101
#define FILE_RECV_REQ	102

#define FILE_GOSSIP	0xB1
#define FILE_OTHER	0xB2
#endif

class CFileTransfer{

public:
	CFileTransfer(int iSockFd, const std::string& sSharedDir, const std::string& sSharedDir, int iSearchID=-1);
	virtual ~CFileTransfer();

	static int ConnectPeer(uint32_t iRemoteIP, uint16_t iRemotePort);
	int GetSockFd(){return m_iSockFd;}
	std::string GetFileName(){return m_sFileName.c_str();}
	int GetFileSize(){return m_iFileSize;}

	int GetSearchID(){return m_iSearchID;}
	std::string GetSharedDir(){return m_sSharedDir;}
	std::string GetDownloadDir(){return m_sDownloadDir;}
	void Cleanup();
	void CloseSocket();
#ifdef CLIENT
	void SetFileHash(const CMD4Hash& fHash){ m_fileHash = fHash;}
	CMD4Hash GetFileHash(){return m_fileHash;}
#endif
	void SetSocketPair(int iFd){m_iSocketPairFd = iFd;}
	int GetSocketPair(){return m_iSocketPairFd;}
	void NotifyComplete(bool bSuccess);

	int HandShake(bool bPassive, bool bPush=false, const std::string& sFileName=std::string(), int iFileSize=0);
	int HandShake(bool bPassive, bool bPush, int iSockFd);
	uint8_t GetFileType(){return m_8FileType;}
	void SetFilePath(const std::string& sPath){m_sFilePath = sPath;}
	bool IsCompleted(){return m_bCompleted;}
	const std::string GetExtraData(){return m_sExtraData;}
	int StartFileTransfer();






private:
	int SendPacket(int iSockFD, const char* szPayload, int iSize); /*code is unused at this time*/
	int ParseHandShakePacket(const char* szBuffer, int iReceived, bool& bPull);
	int WaitForHandShakePacket(bool& bPull, bool bWaitForAck=false);
	int WaitForHandShakePacketAck();
	int SendHandShakePacket(bool bPush);
	int StartFileTransferThread();
	static void* FileTransferThread(void* arg);
	void ReceiveFile();
	void SendFile();
        std::string FileNameFromPath(const std::string&)const;










private:
	int m_iSockFd;
        std::string m_sFileName;
        std::string m_sFilePath;
        int m_iFileSize;
	pthread_t m_pTID;
	int m_iSearchID;
	std::string m_sSharedDir;
	std::string m_sDownloadDir;
#ifdef CLIENT
	CMD4Hash m_fileHash;
#endif
	int m_iSocketPairFd;
	uint8_t	m_8FileType;
	std::string m_sExtraData;
	bool m_bCompleted;
	bool m_bPassive;
	bool m_bPush;

};


#endif //FILE_TRANSFER_H
