
#include<iostream>
#include<fstream>
#include<set>

#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/x509.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>

//for stat
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "digiSig.h"
#include "verify.h"
#include "debug.h"



/*
generate private key:
$openssl genrsa -out privkey.pem 2048

genertae a certificate from the private key:
$openssl req -new -x509 -key privkey.pem -out cacert.pem -days 1095

*/

#ifdef __cplusplus
extern "C" {
#define NON_MAIN
#include "apps.h"
#undef NON_MAIN

}
#endif


using std::cerr;
using std::endl;



CDigiSig::CDigiSig(){


  /* Just load the crypto library error strings,
   * SSL_load_error_strings() loads the crypto AND the SSL ones */
  /* SSL_load_error_strings();*/
  ERR_load_crypto_strings();

}

CDigiSig::~CDigiSig(){

}

int CDigiSig::SignSMIME(const std::string& sCert, const std::string& sPrivateKey, const std::string& sCommandFile, const std::string& sOutFile, const std::string& sTo, const std::string& sFrom, const std::string& sSubject, bool bAddNoCert){

        BIO *in = NULL, *out = NULL, *tbio = NULL;
        X509 *scert = NULL;
        EVP_PKEY *skey = NULL;
        PKCS7 *p7 = NULL;
        FILE* fp = 0;
        int ret = 1;
        char *p = 0;
        int iLen = -1;

        /* For simple S/MIME signing use PKCS7_DETACHED.
         * On OpenSSL 0.9.9 only:
         * for streaming detached set PKCS7_DETACHED|PKCS7_STREAM
         * for streaming non-detached set PKCS7_STREAM
         */
        int flags = 0;//PKCS7_DETACHED|PKCS7_STREAM;
	if(bAddNoCert)
		flags |= PKCS7_NOCERTS;

        OpenSSL_add_all_algorithms();
        ERR_load_crypto_strings();

        if(sPrivateKey.empty())
                return -1;

        if(sCert.empty())
                return -1;

        fp = fopen (sCert.c_str(), "r");
        if (fp == NULL)
                return -1;

        scert = PEM_read_X509(fp, NULL, NULL, NULL);
        fclose (fp);

        if (scert == NULL) {
                ERR_print_errors_fp (stderr);
                return -1;
        }

        fp = fopen (sPrivateKey.c_str(), "r");
        if (fp == NULL)
                return -1;

        skey = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
        fclose (fp);
        if (skey == NULL) {
                ERR_print_errors_fp (stderr);
                return -1;
        }



        if (!scert || !skey)
                goto err;
        /* Open content being signed */

        in = BIO_new_file(sCommandFile.c_str(), "r");

        if (!in)
                goto err;

        /* Sign content */
        p7 = PKCS7_sign(scert, skey, NULL, in, flags);

        if (!p7)
                goto err;

        out = BIO_new_file(sOutFile.c_str(), "w");
        //out = BIO_new(BIO_s_mem());
        if (!out)
                goto err;

        if (!(flags & PKCS7_STREAM))
                BIO_reset(in);

        if (!sTo.empty())
                BIO_printf(out, "To: %s\n", sTo.c_str());
        if (!sFrom.empty())
                BIO_printf(out, "From: %s\n", sFrom.c_str());
        if (!sSubject.empty())
                BIO_printf(out, "Subject: %s\n", sSubject.c_str());


        /* Write out S/MIME message */
        if (!SMIME_write_PKCS7(out, p7, in, flags))
                goto err;

        ret = 0;

err:

        if (ret)
                {
                fprintf(stderr, "Error Signing Data\n");
                ERR_print_errors_fp(stderr);
                }

        if (p7)
                PKCS7_free(p7);
        if (scert)
                X509_free(scert);
        if (skey)
                EVP_PKEY_free(skey);

        if (in)
                BIO_free(in);
        if (out)
                BIO_free(out);
        if (tbio)
                BIO_free(tbio);

        return ret;
}

int CDigiSig::VerifySMIME(const std::string& sFileToVerify, const std::string& sCAChain, const std::string& sDataFile, const std::string& sSignerCert){

	DB("In verifySMIME: fileToVerify="<<sFileToVerify<<std::endl);
	DB("In verifySMIME: CAchain="<<sCAChain<<std::endl);
	DB("In verifySMIME: data file="<<sDataFile<<std::endl);
	DB("In verify SMIME: signer cert="<<sSignerCert<<std::endl);

	int ret = 0;
        X509_STORE *store = NULL;
        PKCS7 *p7 = NULL;

        BIO *cont = NULL;
        BIO *out = NULL;
        BIO *in = NULL;

        STACK_OF(X509) *other = NULL;

        apps_startup();
	

        /* Open content being signed */
        in = BIO_new_file(sFileToVerify.c_str(), "r");
        if (!in){
		std::cout<<"failed to get mem for in\n";
                ERR_print_errors_fp(stderr);
                goto err;
	}


        p7 = SMIME_read_PKCS7(in, &cont);
        if (!p7){
		std::cout<<"failed to read p7\n";
                ERR_print_errors_fp(stderr);
		print_file(sFileToVerify);
                goto err;
	}


        if (!(store = setup_verify(bio_err, sCAChain.c_str(), NULL))){
		std::cout<<"failed to store\n";
                ERR_print_errors_fp(stderr);
                        goto err;
	}

        /*Read the signer certificate*/
        if(!sSignerCert.empty()){
                if (!(other = load_certs(bio_err,sSignerCert.c_str(),FORMAT_PEM, NULL, NULL, "certificate file"))){
			ERR_print_errors_fp(stderr);
                        std::cout<<"other is failed to load\n";
                        goto err;
                }
                else{
                        DB("other is loaded");
		}
        }

        out = BIO_new_file(sDataFile.c_str(), "w");
        if (!out)
                goto err;

        if (PKCS7_verify(p7, other, store, cont, out, 0)){
                fprintf(stderr, "Verification Successful\n");
                ret = 1;
        }
        else{
                fprintf(stderr, "Verification failure\n");
                ERR_print_errors_fp(stderr);
                goto err;
        }


err:
        X509_STORE_free(store);
        sk_X509_pop_free(other, X509_free);

        if (p7)
                PKCS7_free(p7);
        if (out)
                BIO_free(out);
        if (in)
                BIO_free(in);
return ret;

#if 0
        BIO *in = NULL, *out = NULL, *tbio = NULL, *cont = NULL, *sbio = NULL;
        X509_STORE *st = NULL;
        X509 *cacert = NULL;
        X509 *signercert = NULL;
        PKCS7 *p7 = NULL;
        STACK_OF(X509) *st_x509 = NULL;

//test
        std::string s;
//

        int ret = 1;

        OpenSSL_add_all_algorithms();
        ERR_load_crypto_strings();

        /* Set up trusted CA certificate store */

        st = X509_STORE_new();

        /* Read in signer certificate and private key */
        tbio = BIO_new_file(sCAChain.c_str(), "r");

        if (!tbio)
                goto err;


        cacert = PEM_read_bio_X509(tbio, NULL, 0, NULL);

        if (!cacert)
                goto err;

        if (!X509_STORE_add_cert(st, cacert))
                goto err;
        /*Read the signer certificate*/
        if(!sSignerCert.empty()){

                sbio = BIO_new_file(sSignerCert.c_str(), "r");

                if (!sbio)
                        goto err;

                signercert = PEM_read_bio_X509(sbio, NULL, 0, NULL);

                if (!signercert)
                        goto err;

                st_x509=sk_X509_new_null();
                sk_X509_push(st_x509,signercert);
        }


        /* Open content being signed */

        in = BIO_new_file(sFileToVerify.c_str(), "r");

        if (!in)
                goto err;

        /* Sign content */
        p7 = SMIME_read_PKCS7(in, &cont);

        if (!p7)
                goto err;
        out = BIO_new_file(sDataFile.c_str(), "w");
        if (!out)
                goto err;

        if (PKCS7_verify(p7, st_x509, st, cont, out, 0)!=1)
        //if (!PKCS7_verify(p7, NULL, st, cont, out, 0))
                {
                fprintf(stderr, "Verification Failure\n");
                goto err;
                }

        fprintf(stderr, "Verification Successful\n");

//test
        //GetParam(sFileToVerify, s, ISSUER);
        GetParamsFromPKCS7(p7, SUBJECT, s);
        std::cout<<s<<std::endl;

        GetParamsFromPKCS7(p7, ISSUER, s);
        std::cout<<s<<std::endl;

//

        ret = 0;
        err:

        if (ret)
                {
                fprintf(stderr, "Error Verifying Data\n");
                ERR_print_errors_fp(stderr);
                }

        if (p7)
                PKCS7_free(p7);

        if (cacert)
                X509_free(cacert);

        if (in)
                BIO_free(in);
        if (out)
                BIO_free(out);
        if (tbio)
                BIO_free(tbio);

        if(st_x509)
             sk_X509_free(st_x509);


return ret;
#endif
}
int CDigiSig::GetParamsFromPKCS7(const PKCS7 *p7, ReqType req, std::string& sOut){


        STACK_OF(X509) *certs=NULL;
        STACK_OF(X509_CRL) *crls=NULL;
        char* str;

        int i=OBJ_obj2nid(p7->type);

        switch (i)
        {
        case NID_pkcs7_signed:
                certs=p7->d.sign->cert;
                crls=p7->d.sign->crl;
                break;
        case NID_pkcs7_signedAndEnveloped:
                certs=p7->d.signed_and_enveloped->cert;
                crls=p7->d.signed_and_enveloped->crl;
                break;
        default:
                break;
        }

        if (certs == NULL)
                return -1;

        X509 *x509;

        printf("no. of certs=%d", sk_X509_num(certs));

        for (i=0; i<sk_X509_num(certs); i++){

                x509=sk_X509_value(certs,i);
                //      X509_print(out, x);

                switch(req){

                        case SUBJECT:
                        {
                                str = X509_NAME_oneline (X509_get_subject_name (x509),0,0);
                                break;
                        }

                        case ISSUER:
                        {
                                str = X509_NAME_oneline (X509_get_issuer_name  (x509),0,0);
                                break;
                        }

                }

                if(str == NULL){

                        return -1;
                }
                //printf ("\t param: %s\n", str);
                std::string sName(str);
                OPENSSL_free (str);

                int iPos = sName.find("CN=");
                sOut = sName.substr(iPos+3);
        }

#if 0
        if (crls != NULL){
                
                X509_CRL *crl;

                for (i=0; i<sk_X509_CRL_num(crls); i++){

                        crl=sk_X509_CRL_value(crls,i);

                        X509_CRL_print(out, crl);

                        if(1)
                                PEM_write_bio_X509_CRL(out,crl);
                        BIO_puts(out,"\n");
                }
        }


//        if (in != NULL) BIO_free(in);
//        if (out != NULL) BIO_free_all(out);
#endif

return 0;
}


int CDigiSig::Sign(const std::string& sPrivateKey, const std::string& sCommand, uint8_t* sig_buf, unsigned int & iSigLen){

FILE *fp=0;
EVP_PKEY * pkey;
EVP_MD_CTX     md_ctx;
int sig_len;
int err;



	/* Read private key */
	if(sPrivateKey.empty())
		return -1;

	fp = fopen (sPrivateKey.c_str(), "r");
	if (fp == NULL) 
		return -1;

	pkey = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
	fclose (fp);

	if (pkey == NULL) {
		ERR_print_errors_fp (stderr);
		return -1;
	}

	/* Do the signature */

	EVP_SignInit   (&md_ctx, EVP_sha1());
	EVP_SignUpdate (&md_ctx, sCommand.c_str(), sCommand.size());
	err = EVP_SignFinal (&md_ctx, sig_buf, &iSigLen, pkey);

	if (err != 1) {
		ERR_print_errors_fp(stderr);
		return -1;
	}

	EVP_PKEY_free (pkey);


return 0;
}

bool CDigiSig::VerifyByCA(const std::string& sCert, const std::string& sFile){

	bool bRet = false;

        char **sArg=0;
        const char* param = "-CAfile";

        sArg = new char*[4];

        sArg[0] = new char[1];

        sArg[1] = new char[8];
	bzero(sArg[1], 8);
        memcpy(sArg[1], param, 7 );

        sArg[2] = new char[sCert.size()+1];
	bzero(sArg[2], sCert.size()+1);
        memcpy(sArg[2], sCert.c_str(), sCert.size());

        sArg[3] = new char[sFile.size()+1];
	bzero(sArg[3], sFile.size()+1);
        memcpy(sArg[3], sFile.c_str(), sFile.size());

        int iRet = MAIN2(4, sArg);
        printf("iRet=%d\n",iRet);

        for(int i=4;i>0;--i){
                delete [] sArg[i-1];
		sArg[i-1] = 0;
        }
        delete [] sArg;
	sArg = 0;

	if(iRet==1)
		bRet = true;


return bRet;
}

bool CDigiSig::GetSubjectName(const std::string& sPublicKey, std::string& sSubjectName ){

	return GetParam(sPublicKey, sSubjectName, SUBJECT);
}

bool CDigiSig::GetIssuerName(const std::string& sPublicKey, std::string& sIssuerName ){

	return GetParam(sPublicKey, sIssuerName, ISSUER);
}


bool CDigiSig::GetParam(const std::string& sPublicKey, std::string& sParamName, const ReqType eReq ){

FILE *          fp;
X509 *        x509;
EVP_PKEY *      pkey;
EVP_MD_CTX     md_ctx;
int err;
char* str;

        /* Read public key */
        if(sPublicKey.empty())
                return false;

        fp = fopen (sPublicKey.c_str(), "r");
        if (fp == NULL)
                return false;

        x509 = PEM_read_X509(fp, NULL, NULL, NULL);
        fclose (fp);

        if (x509 == NULL) {
                ERR_print_errors_fp (stderr);
                return false;
        }

	switch(eReq){

		case SUBJECT:
			{
			str = X509_NAME_oneline (X509_get_subject_name (x509),0,0);
			break;
			}

		case ISSUER:
			{
			str = X509_NAME_oneline (X509_get_issuer_name  (x509),0,0);
			break;
			}

	}

	if(str == NULL){
		return false;
	}
	//printf ("\t param: %s\n", str);
	std::string sName(str);
	OPENSSL_free (str);

	int iPos = sName.find("CN=");
	sParamName = sName.substr(iPos+3);

	//sazzad: don't know if this is necessary..
	X509_free (x509);

return true;
}

bool CDigiSig::Verify(const std::string& sPublicKey, const std::string& sCommand, const uint8_t* sig_buf, unsigned int iSigLen){

FILE *          fp;
X509 *        x509;
EVP_PKEY *      pkey;
EVP_MD_CTX     md_ctx;
int err;



	/* Read public key */
	if(sPublicKey.empty())
		return false;

	fp = fopen (sPublicKey.c_str(), "r");
	if (fp == NULL) 
		return false;

	x509 = PEM_read_X509(fp, NULL, NULL, NULL);
	fclose (fp);

	if (x509 == NULL) {
		ERR_print_errors_fp (stderr);
		return false;
	}

	/* Get public key - eay */
	pkey=X509_get_pubkey(x509);
	if (pkey == NULL) {
		ERR_print_errors_fp (stderr);
		return false;
	}

	/* Verify the signature */

	EVP_VerifyInit   (&md_ctx, EVP_sha1());
	EVP_VerifyUpdate (&md_ctx, sCommand.c_str(), sCommand.size());
	err = EVP_VerifyFinal (&md_ctx, sig_buf, iSigLen, pkey);
	EVP_PKEY_free (pkey);

	if (err != 1) {
		ERR_print_errors_fp (stderr);
		return false;
	}

	//printf ("Signature Verified Ok.\n");

return true;
}

int CDigiSig::ParseMIME(const std::string& sFile, ReqType req, std::string& sOut){

std::string sSearch;
int iLen;

	switch(req){
		case FROM:
			sSearch = "From: ";
			iLen=6;
			break;
		case TO:
			sSearch = "To: ";
			iLen=4;
			break;
		case SUBJECT:
			sSearch = "Subject: ";
			iLen=9;
			break;
		default:
			std::cout<<"Unknown Request type\n";
			break;
	}

	std::ifstream ifFile(sFile.c_str());

	if (!ifFile.is_open()){
		std::cout<<"Unable to open file\n";
		return -1;
	}

	#define MAXCHAR 1500
        char szBuf[MAXCHAR];
	do{
                bzero(szBuf,MAXCHAR);
		ifFile.getline(szBuf, MAXCHAR);
		std::string sLine(szBuf);
		DB(sLine);
		int iPos = sLine.find(sSearch);
		if(iPos != std::string::npos){
			sOut = sLine.substr(iPos+iLen);
			break;
		}
	}while(!ifFile.eof());
	
	ifFile.close();

	if(sOut.empty()){

		std::cout<<"Failed to parseMIME:"<<sFile<<std::endl;
		return -1;
	}

return 0;
}

int CDigiSig::GetAllIssuer(const std::string& sCert, std::set<std::string>& sIssuerSet){


char* str;

        STACK_OF(X509) *certs = NULL;

        if (!(certs = load_certs(bio_err,sCert.c_str(),FORMAT_PEM, NULL, NULL, "certificate file"))){
                ERR_print_errors(bio_err);
                return -1;
        }

        for(int i = 0; i < sk_X509_num(certs); i++){

                X509* x509 = sk_X509_value(certs, i);

                if (x509 == NULL) {
                        ERR_print_errors_fp (stderr);
                        return -1;
                }
                str = X509_NAME_oneline (X509_get_issuer_name  (x509),0,0);

                if(str != NULL){
                        std::string sName(str);
                        OPENSSL_free (str);

                        int iPos = sName.find("CN=");
                        std::string sIName = sName.substr(iPos+3);
                        DB(sIName);
                        sIssuerSet.insert(sIName);
                }

        }


        sk_X509_pop_free(certs, X509_free);

return 0;
}

int CDigiSig::save_certs(const char *signerfile, STACK_OF(X509) *signers)
        {
        int i;
        BIO *tmp;
        if (!signerfile)
                return 1;
        tmp = BIO_new_file(signerfile, "w");
        if (!tmp) return 0;
        for(i = 0; i < sk_X509_num(signers); i++)
                PEM_write_bio_X509(tmp, sk_X509_value(signers, i));
        BIO_free(tmp);
        return 1;
        }


int CDigiSig::GetSenderCertificate(const std::string& sMsg, const std::string& sFileToSave){

        PKCS7 *p7 = NULL;
        STACK_OF(X509) *signers;
        BIO *in = NULL;
        BIO *cont = NULL;
        int iRet=-1;

        /* Open content being signed */ 
        in = BIO_new_file(sMsg.c_str(), "r");
        if (!in)
                goto err;

        p7 = SMIME_read_PKCS7(in, &cont);
        if (!p7){
                goto err;
        }

        signers = PKCS7_get0_signers(p7, NULL, 0);

        if (!save_certs(sFileToSave.c_str(), signers)){
                BIO_printf(bio_err, "Error writing signers to %s\n",sFileToSave.c_str());
                iRet = 5;
                goto err;
        }

        sk_X509_free(signers);

        if (p7)
                PKCS7_free(p7);
        if (in)
                BIO_free(in);

        iRet=0;
err:

return iRet;
}

void CDigiSig::print_file(const std::string& sFileName){


        struct stat fStat;
        stat( sFileName.c_str(), &fStat );
        //fStat.st_size;

FILE* fp = fopen(sFileName.c_str(),"r");
if(!fp)
        return ;

char szBuffer[4096]={0};
int iReadSize=0;
        while (1){
                bzero(szBuffer, 4096);

                iReadSize=fread(szBuffer,1,4096,fp);
                if(iReadSize<=0){
                        break;
                }
                break;
        }
	std::string s(szBuffer, iReadSize);
	DB("File Size="<<fStat.st_size<<std::endl);
	DB("File: "<<s<<std::endl);

fclose(fp);
}


#if 0 //only testing

int main(){

unsigned char sig_buf [4096];
static char publicfile[] = "cert.pem";
static char privatefile[]  = "key.pem";
static char data[]     = "I owe you...";
unsigned int iSize;

	CDigiSig obj;
	int iRet = obj.Sign(privatefile, data, sig_buf, iSize );
	if(iRet<0)
		std::cout<<"signature failed\n";

	else
	{

		std::cout<<"sign done: size="<<iSize<<std::endl;
		std::string s((char*)sig_buf, iSize);
		std::cout<<"sig="<<s<<std::endl;
		bool bRet = obj.Verify(publicfile, data, sig_buf, iSize);
		if(bRet)
			std::cout<<"sig varified\n";
		else
			std::cout<<"sig NOT varified\n";
	}

return 0;
}

#endif
