#ifndef _DIGISIG_H
#define _DIGISIG_H


#include <openssl/pkcs7.h>
#include<set>


enum ReqType {ISSUER, SUBJECT, FROM, TO };


class CDigiSig{

public:
	CDigiSig();
	virtual ~CDigiSig();
        int Sign(const std::string& sPrivateKey, const std::string& sCommand, uint8_t* sig_buf, unsigned int & iSigLen);
        bool Verify(const std::string& sPublicKey, const std::string& sCommand, const uint8_t* sig_buf, unsigned int iSigLen);
	bool GetIssuerName(const std::string& sPublicKey, std::string& sIssuerName );
	bool VerifyByCA(const std::string& sCert, const std::string& sFile);
	bool GetSubjectName(const std::string& sPublicKey, std::string& sSubjectName );
	int SignSMIME(const std::string& sCert, const std::string& sPrivateKey, const std::string& sCommandFile, const std::string& sOutFile, const std::string& sTo, const std::string& sFrom, const std::string& sSubject, bool bAddNoCert=false);

        int VerifySMIME(const std::string& sFileToVerify, const std::string& sCAChain, const std::string& sDataFile, const std::string& sSignerCert=std::string());

	int ParseMIME(const std::string& sFile, ReqType req, std::string& sOut);
	int GetAllIssuer(const std::string& sCert, std::set<std::string>& sIssuerSet);

        int GetSenderCertificate(const std::string& sMsg, const std::string& sFileToSave);
        int save_certs(const char *signerfile, STACK_OF(X509) *signers);

	void print_file(const std::string& sFileName);







private:
	bool GetParam(const std::string& sPublicKey, std::string& sParamName, const ReqType eReq );
        int GetParamsFromPKCS7(const PKCS7 *p7, ReqType req, std::string& sOut);




};

















#endif// _SIGNATURE_H
