#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>


#include "BootStrapClient.h"
#include "FileTransfer.h"
#include "EventMessage.h"
#include "serializable.h"
//#include "log.h"
//#include "debug.h"
#include "Logger.h"


#define SOCKET_ERROR -1
#define POLL_TIMEOUT	(4*1000) //sec
#define MAX_POLL_TIMEOUT	15
#define MAXCHAR 1500


CBootStrapClient::CBootStrapClient():
        m_uServerIP(0),
        m_uServerPort(0),
        m_sDownloadDir(),
	m_pEventMsg(0){

	m_pEventMsg = new EventMessage();

}

CBootStrapClient::~CBootStrapClient(){

	if(m_iSockPairFd[0] != SOCKET_ERROR){
		::close(m_iSockPairFd[0]);
		m_iSockPairFd[0] = SOCKET_ERROR;
	}
	if(m_iSockPairFd[1] != SOCKET_ERROR){
		::close(m_iSockPairFd[1]);
		m_iSockPairFd[1] = SOCKET_ERROR;
	}

	delete m_pEventMsg;
	m_pEventMsg = NULL;

	for(int i=0;i<m_filetrans.size();i++){
		delete m_filetrans.at(i);
	}
}

int CBootStrapClient::Connect(){


	int iSockFd = CFileTransfer::ConnectPeer(m_uServerIP, m_uServerPort);
	if(iSockFd==-1){
		AddLogLineM(false, wxT("Failed to create tcp service socket"));
		return -1;
	}

	CFileTransfer* filetrans = new CFileTransfer(iSockFd,"", m_sDownloadDir);
	filetrans->SetSocketPair(m_iSockPairFd[1]);
	filetrans->HandShake(true, false, "", 0);
	m_filetrans.push_back(filetrans);

return 0;
}

int CBootStrapClient::Download(uint32_t server_ip, uint16_t server_port, const std::string& sDownloadDir){

	m_uServerIP = server_ip;
	m_uServerPort = server_port;
	m_sDownloadDir = sDownloadDir;

	struct pollfd pollFd[1];


	int iRet = InitSocketPair();
	if(iRet==-1)
		return -1;

	iRet = Connect();
	if(iRet==-1)
		return -1;

	pollFd[0].fd = m_iSockPairFd[0];
	pollFd[0].events = POLLIN;


	int iTotalTimeout = 0;
	iRet=-1;
	while (1){
		int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
		if( iStatus < 0){
			AddLogLineM(false, wxT("poll failed in main loop"));
		}
		else if( 0 == iStatus){
			AddLogLineM(false, wxT("Timeout in poll"));
			iTotalTimeout++;
			if(iTotalTimeout>=MAX_POLL_TIMEOUT){
				AddLogLineM(false, wxT("poll waited too long for node file list. Terminating"));
				iRet = -2;
				break;
			}
			
		}
		else if(pollFd[0].revents & POLLIN){
		        iRet = HandleNotification(pollFd[0].fd);
			break;
		}
		else{
			AddLogLineM(false, wxT("Unknown poll return "));
		}
	}

return iRet;
}

int CBootStrapClient::InitSocketPair(){

	int rc = socketpair( AF_UNIX, SOCK_STREAM, 0, m_iSockPairFd );
	if ( rc < 0 ) {
		AddLogLineM(false, wxT("Cannot open socketpair "));
		return -1;
	}

return 0;
}

int CBootStrapClient::HandleNotification(int iFd){

	char szBuffer[MAXCHAR];
	int iRet = -1;

	sockaddr_in sFrom;
	memset((void*)&sFrom, 0, sizeof(sFrom));

	int fromlen = sizeof(sFrom);

	int iReceived = recvfrom(iFd, (void*)szBuffer, MAXCHAR, 0,
					(sockaddr*)&sFrom, (socklen_t*)&fromlen );

	uint32_t uAddr = ntohl(sFrom.sin_addr.s_addr);
	uint16_t uPort = ntohs(sFrom.sin_port);

	const char* pBuf = szBuffer;

	if(0==iReceived){
		AddLogLineM(false, wxT("sock pair is down: "));
	}

	if(iReceived<0) {
		AddLogLineM(false, wxT("Failed to refceived data from sock pair!"));
	}
	else{
		int iRead = 0;
		while(iRead<iReceived){
			const char* pBuf = szBuffer;
			pBuf+=iRead;

			EventMessage msg;
			std::string sBuf(pBuf,iReceived-iRead);
			iRead += msg.deserialize(sBuf);
			switch(msg.GetStatus()){

				case  EVT_SUCCESS:
			        {
					AddLogLineM(false, wxT("Node file download succeed!"));
					iRet = 0;
					*m_pEventMsg = msg;
					break;
				}
				case EVT_ERROR:
				{
					AddLogLineM(false, wxT("Node file download Failed!"));
					break;

				}

			}
		}

	}
return iRet;
}
