#ifndef EVENTMESSAGE_H
#define EVENTMESSAGE_H

#include <stdint.h>
#include "serializable.h"

//iStatus
#define EVT_SUCCESS	0
#define EVT_ERROR	-1
#define EVT_SHUTDOWN	-2
#define EVT_SEND_REPORT	-3


//iReasonCode
#define FILE_SEND_REQ   	101
#define FILE_RECV_REQ   	102
#define FILE_FIND_REQ   	103
#define FILE_PUBLISH_REQ	104
#define FILE_GOSSIP_REQ		105
#define FILE_BOOTSTRAP_SRV_REQ	106
#define FILE_BOOTSTRAP_SRV_RES	107
#define FILE_BOOTSTRAP_KAD_REQ	109
//for validating publish cert.
#define FILE_REPORT_CERT_REQ	110

#define FILE_SEND_REPORT	111
#define FILE_RECV_REPORT	112

//8FileType
#define FILE_GOSSIP     	0xB1
#define FILE_OTHER      	0xB2
#define FILE_MULTICAST  	0xB3
#define FILE_BOOTSTRAP  	0xB4
#define FILE_UNKNOWN	  	0xB5
#define FILE_ZERO_CONTACT  	0xB6
#define FILE_ALL_CONTACTS  	0xB7

#define USE_TCP		201
#define USE_UDP		202

//8RelatedModule
#define MODULE_MULTICAST			0xC1
#define MODULE_GOSSIP				0xC2
#define MODULE_KAD				0xC3
#define MODULE_UNKNOWN				0xC4
#define MODULE_GOSSIP_AND_MULTICAST		0xC5

#define LOG_ONLY_TRUE	1001
#define LOG_ONLY_FALSE	1002

#define SAFE_DELETE(X) {if (X) { delete (X); (X) = NULL; } }


class EventMessage: public ISerializable{


public:
	EventMessage();
	virtual ~EventMessage();

	/*EVT_SUCCESS or EVT_ERROR?*/
	void SetStatus(int i){m_iStatus = i;}

	/*Future use*/
	void SetReasonCode(int i){m_iReasonCode = i;}
	void SetReasonString(const std::string& s){m_sReasonString = s; m_iReasonStringLen = s.size();}

	void SetFileType(uint8_t t){m_8FileType = t;}
	void SetFileSize(int t){m_iFileSize = t;}
	void SetFileName(const std::string& s){m_sFileName = s; m_iFileNameLen = s.size();}
	void SetFilePath(const std::string& s){m_sFilePath = s; m_iFilePathLen = s.size();}
	void SetSearchID(int i){m_iSearchID = i;}

	int GetStatus(){return m_iStatus;}
	int GetReasonCode(){return m_iReasonCode;}
	std::string GetReasonString(){return m_sReasonString ; }
	uint8_t GetFileType(){return m_8FileType;}
	int GetFileSize(){return m_iFileSize;}
	std::string GetFileName(){return m_sFileName;}
	std::string GetFilePath(){return m_sFilePath ;}
	int GetSearchID(){return m_iSearchID ;}
	void SetRelatedModule(uint8_t t){m_8RelatedModule = t;}
	uint8_t GetRelatedModule(){return m_8RelatedModule;}




        virtual void serialize(std::string& sBuf) const ;
        virtual int deserialize(const std::string& sBuf);

private:
	/*Is this a success or a failure notification*/
	int m_iStatus;

	/*Reason Code*/
	int m_iReasonCode;

	int m_iReasonStringLen;
	/*Reason string for human readability*/
	std::string m_sReasonString;

	/*File Information*/

	/*Goosip file, or certificate?*/
	uint8_t m_8FileType;

	int m_iFileSize;

	int m_iFileNameLen;
	std::string m_sFileName;

	int m_iFilePathLen;
	std::string m_sFilePath;

	int m_iSearchID;

	/*Goosip file, or certificate?*/
	uint8_t m_8RelatedModule;



};

#endif // EVENTMESSAGE_H
