/*
Author: Md Sazzadur Rahman
rahmanm@cs.ucr.edu
*/

#include <iostream>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <error.h>

#include "../aMule-2.2.4/src/Logger.h"
#include "FileTransfer.h"
#include "EventMessage.h"
#include "util.h"


#define SOCKET_ERROR 	-1
#define POLL_TIMEOUT	1000
#define MAX_TIMEOUT_HANDSHAKE	20
//#define MAX_TIMEOUT_HANDSHAKE	100
#define MAX_TIMEOUT	10
//#define MAX_TIMEOUT	5


CFileTransfer::CFileTransfer(int iSockFd, const std::string& sSharedDir, const std::string& sDownloadDir, int iSearchID):
	m_iSockFd(iSockFd),
	m_sSharedDir(sSharedDir),
	m_sDownloadDir(sDownloadDir),
	m_iSearchID(iSearchID),
	m_8FileType(0),
	m_bCompleted(false),
	m_bPassive(false),
	m_bPush(false),
	m_sGossipId(),
	m_bCheckGossipId(true)
{

}


CFileTransfer::~CFileTransfer()
{

}

int CFileTransfer::HandShake(bool bPassive, bool bPush, const std::string& sFileName, int iFileSize){

int iRet = -1;
	if( !sFileName.empty() && iFileSize!=0){
		m_sFileName = sFileName;
		m_iFileSize = iFileSize;
	}
	m_bPassive=bPassive;
	m_bPush=bPush;

	iRet = StartFileTransferThread();
	if(iRet!=0){
		CloseSocket();
		return -1;
	}

return 0;
}

int CFileTransfer::HandShake(bool bPassive, bool bPush, int iSockFd){

	m_iSockFd = iSockFd;
	if(m_sFileName.empty()){
		AddLogLineM(false, wxT("file name is empty "));
		CloseSocket();
		return -1;
	}

	return HandShake(bPassive, bPush, m_sFileName, m_iFileSize);
}

int CFileTransfer::StartFileTransferThread(){

        pthread_attr_t threadAttr;
        if( pthread_attr_init(&threadAttr) != 0 )
                AddLogLineM(false, wxT("Error in pthread_attr_init"));

        if( pthread_attr_setdetachstate(&threadAttr, PTHREAD_CREATE_DETACHED) != 0 )
                AddLogLineM(false, wxT("Error in pthread_attr_setdetachstate"));

        int iRet = pthread_create(&m_pTID,&threadAttr,CFileTransfer::FileTransferThread, this);
        if (iRet){
                 AddLogLineM(false, wxT("ERROR; return code from pthread_create() is ")+ wxString::Format(wxT("%i"),iRet));
                 perror("error in creating receiver thread:");
                 return -1;
        }

        if( pthread_attr_destroy(&threadAttr) != 0 )
                AddLogLineM(false,wxT("Error in pthread_attr_destroy"));

return 0;
}
void* CFileTransfer::FileTransferThread(void* arg){


        CFileTransfer& filetrans = *(CFileTransfer*)arg;


	filetrans.StartFileTransfer();
	filetrans.Cleanup();

return NULL;
}

int CFileTransfer::StartFileTransfer(){

int iRet=-1;
	//we check the socket fd; if it is not set, we do connect..
	if(m_iSockFd==-1){
		m_iSockFd = ConnectPeer(m_iRemoteIP, m_iRemotePort);
		if(m_iSockFd==-1){
			if(!m_bPassive){
				m_8FileType = FILE_GOSSIP;
				NotifySenderReport(false);
			}
			return -1;
		}
		
	}

	AddressPortPair p = GetPeerName(m_iSockFd);
        if(m_bPassive){
                //wait for handshake packet
                //depending on the packet, start sender or receiver thread..
                bool bPull;
                AddLogLineM(false, wxT("Waiting for handshake packet : sock=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));

                iRet = WaitForHandShakePacket(bPull);
                if(iRet!=0){
                        AddLogLineM(false, wxT("Failed to receive handshake packet "));
			//if(bPull){//getting a gossip file..
				//notify ..
			m_8FileType = FILE_GOSSIP;
			NotifyComplete(false);
			//}
                        return iRet;
                }

                if(bPull){
                        //send ACK for Handshake pkt: sender should start after that..
                        AddLogLineM(false, wxT("Sending READY packet :sock=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                        SendPacket(m_iSockFd, "READY", 5 );

                        AddLogLineM(false, wxT("Going to start Receiver thread "));
                        int iRet = ReceiveFile();
			if(iRet==0)
				NotifyComplete(true);
			else
				NotifyComplete(false);
                }
                else{
                        AddLogLineM(false, wxT("Going to start Sender thread "));
                        SendFile();

			// just to notify the main network thread to decrease the threadcounter..
			NotifyComplete(false);
                }
        }

        else{
                //send handshake packet
                AddLogLineM(false, wxT("Sending handshake packet :sock=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                iRet = SendHandShakePacket(m_bPush);
                if(iRet!=0){
                        AddLogLineM(false, wxT("Failed to send handshake packet "));
			if(m_bPush){ //means it is trying to send a gossip packet
				NotifySenderReport(false);
			}
                        return iRet;
                }

                //depending on push or pull, start sender or receiver thread..
                if(m_bPush){
                        //should wait for Handshake ACK
                        iRet = WaitForHandShakePacketAck();
                        if(iRet!=0){
                                AddLogLineM(false, wxT("Failed to receive handshake packet ack "));
				NotifySenderReport(false);
                                return iRet;
                        }
                        //wants to disseminate gossip packet
                        AddLogLineM(false, wxT("Going to start Sender thread "));
                        SendFile();
			NotifySenderReport(true);
                }
                else{
                        //tries to get some file from network
                        AddLogLineM(false, wxT("Going to start Receiver thread "));
                        int iRet = ReceiveFile();
			if(iRet == 0)
				//NotifyComplete(true);
				NotifySenderReport(true);
			else
				//NotifyComplete(false);
				NotifySenderReport(false);
                }
        }

return 0;
}


int CFileTransfer::ReceiveFile(){

int iSockFd = -1;
int iReceived = -1;
uint8_t szBuffer[MAXSIZE];

	iSockFd = GetSockFd();
	/*
	* We should use some randomness in file name. Otherwise multiple threads open same file for write and if one thread fails, it deletes the file and now everything is gone!
	*/
	std::string sFileName = GetFileName();

	std::string sPath;
	FILE* pFile;

	AddLogLineM(false, wxT("Receiver Starting"));

	switch(GetFileType()){

		case FILE_GOSSIP:
		{
				int iRand = rand()%10000;
				sFileName+=".";
				sFileName+=itoa(iRand);
				sPath = GetDownloadDir()+"/"+sFileName;
				break;
		}
		case FILE_OTHER:
				sPath = GetSharedDir()+"/"+sFileName;
				break;
		default:
				std::cout<<"Unknown file type:"<<std::endl;
				break;
	}
	SetFilePath(sPath);

	pFile = fopen(sPath.c_str(), "wb");
	if( pFile == NULL){
		AddLogLineM(false, wxT("Unable to open file for write"));
		//NotifyComplete(false);
		return -1 ;
	}

	unsigned int iFileSize = 0;
	bool bFlag = true;
	bool bNotify = false;
	int iCounter = 0;
#if 0
	//do we have m_sExtraData?
	const std::string& sExtraData = filetrans.GetExtraData();
	if(!sExtraData.empty()){
		fwrite(sExtraData.c_str(),1,sExtraData.size(),pFile);
		iFileSize += sExtraData.size();
		if(iFileSize>=filetrans.GetFileSize()){
			filetrans.SendPacket(iSockFd, "End", 3 );	
			bNotify = true;
			bFlag = false;
		}
	}
#endif
        struct pollfd pollFd[1];
        pollFd[0].fd = iSockFd;
        pollFd[0].events = POLLIN;
        pollFd[0].revents = 0;


	//std::cout<<"File Size="<<filetrans.GetFileSize()<<std::endl;

	//wait for whole data file and receive
	while(bFlag){
		std::cout<<"in loop: ReceiveFile\n";

                int iStatus = poll(pollFd, 1, POLL_TIMEOUT );

                if( iStatus < 0){
			AddLogLineM(false, wxT("Poll failed in receiver thread"));
			bFlag = false;
                }
		else if(0 == iStatus){
                               AddressPortPair p = GetPeerName(iSockFd);
                                AddLogLineM(false, wxT("Poll timeout in receiver thread: sock=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                                iCounter++;
                                if(iCounter>MAX_TIMEOUT){
                                        std::cout<<"Waited too long: terminating..receiver"<<std::endl;
                                        bFlag = false;
                                        //unlink(sPath.c_str();
                                }


		}
		else{
				bzero(szBuffer, MAXSIZE);
				iReceived = recv(iSockFd, (void*)szBuffer, MAXSIZE, 0 );
				if(0==iReceived){
					AddLogLineM(false, wxT("Peer is down: shutting down the receiver"));
					bFlag = false;
					//unlink(sPath.c_str();
					break;
				}
				//std::cout<<"received data: "<<std::string((char*)szBuffer,iReceived)<<std::endl;
                                fwrite(szBuffer,1,iReceived,pFile);
				iFileSize+=iReceived;

				//std::cout<<"iFileSize="<<iFileSize<<std::endl;
				if(iFileSize>=GetFileSize()){
					std::cout<<"Whole File Received: sending End"<<std::endl;
					SendPacket(iSockFd, "End", 3 );	
					bNotify = true;
					bFlag = false;
				}
                }

	}	
	
	fclose(pFile);

	//log ownload bytes for gossiping
	if(!m_sGossipId.empty()){
		std::cout<<"Receiver Thread: Download= "<<iFileSize<<" Gossip ID= "<<m_sGossipId<<std::endl;
	}
	AddLogLineM(false, wxT("Receiver terminating"));

	if(bNotify){
		//NotifyComplete(true);
		return 0;
	}
	else{
		//NotifyComplete(false);
		AddLogLineM(false, wxT("Deleating file in Receiver:"));
		unlink(sPath.c_str()); //if unsuccessfull, we need to delete the file..
	}

return -1;
}

void CFileTransfer::SendFile(){

int iReceived;
int iSockFd=-1;
FILE* pFile=NULL;
uint8_t szBuffer[MAXSIZE];

	iSockFd = GetSockFd();
	std::string sFileName = GetFileName();

	//waits for filename and then starts sends
	AddLogLineM(false, wxT("Sender is Starting"));

	std::string sPath;
        switch(GetFileType()){

                case FILE_GOSSIP:
                                sPath = sFileName;
                                break;
                case FILE_OTHER:
                                sPath = GetSharedDir()+"/"+sFileName;
                                break;
                default:
                                std::cout<<"Unknown file type:"<<std::endl;
                                break;
        }

	pFile = fopen(sPath.c_str(), "rb");
	if( pFile == NULL){
		AddLogLineM(false, wxT("Unable to open file for read: ")+wxString(sPath.c_str(), wxConvUTF8));
		return ;
	}

//now, send the file
	unsigned int iFileSize=0;
        while (1){
		std::cout<<"in loop1: SendFile\n";
		bzero(szBuffer, MAXSIZE);

		int iReadSize=fread(szBuffer,1,MAXSIZE,pFile);
		if(iReadSize<=0){
			break;
		}
		iFileSize+=iReadSize;
		//std::cout<<"Sending PAYLOAD: "<<iReadSize<<std::endl;
		AddLogLineM(false, wxT("sending payload "));
		SendPacket(iSockFd, (char*)szBuffer, iReadSize );
        }
	std::cout<<"Sender Thread: Upload= "<<iFileSize<<" Gossip ID= "<<m_sGossipId<<std::endl;

	//wait for ack
	struct pollfd pollFd[1];
        pollFd[0].fd = iSockFd;
        pollFd[0].events = POLLIN;
        pollFd[0].revents = 0;

	bool bFlag = true;
	int iCounter = 0;
        while (bFlag){
		std::cout<<"in loop2: SendFile\n";

                int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
                if( iStatus < 0){
                        std::cout<<"poll failed in sender thread"<<std::endl;;
			AddLogLineM(false, wxT("Poll failed "));
			bFlag = false;
                }
		else if(0==iStatus){
				AddressPortPair p = GetPeerName(iSockFd);
                                AddLogLineM(false, wxT("poll timeout in sender thread: sock="+ wxString::Format(wxT("%i:%i"),p.first, p.second)));
                                iCounter++;
                                if(iCounter>MAX_TIMEOUT){
                                        AddLogLineM(false, wxT("Waited too long: terminating..sender "));
                                        bFlag = false;
                                }

		}
		else{
				iReceived = recv(iSockFd, (void*)szBuffer, MAXSIZE, 0 );
				AddLogLineM(false, wxT("Received waiting packet "));
				bFlag = false;
                                break;
                }
        }


	fclose(pFile);

AddLogLineM(false, wxT("Sender is terminating"));

#if 0
        if(GetFileType() == FILE_GOSSIP)
		NotifySenderReport();
#endif

}


int CFileTransfer::SendPacket(int iSockFd, const char* szPayload, int iSize){

        int iRet = ::send(iSockFd, (void*)szPayload, iSize , 0);

        if(iRet < 0){
		AddLogLineM(false, wxT("Send failed in file transfer"));
	}

return 0;
}

void CFileTransfer::Cleanup(){

	m_bCompleted = true;
	CloseSocket();


	//int iStatus = 0;
	//pthread_exit((void*)&iStatus);
	pthread_exit(0);

}

void CFileTransfer::CloseSocket(){

	if(m_iSockFd != SOCKET_ERROR){
              ::close(m_iSockFd);
              m_iSockFd = SOCKET_ERROR;
	}
}

int CFileTransfer::ConnectPeer(uint32_t iRemoteIP, uint16_t iRemotePort){

struct sockaddr_in kAddr;

int iRet;
int iSockFd;

        iSockFd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (iSockFd == -1){
                std::cout<<"Could not create socket."<<std::endl;
                return -1;
        }

	std::cout<<"Connected port="<<ntohs(GetPeerName(iSockFd).second)<<std::endl;

        //connect
        struct sockaddr_in destAddr;
        socklen_t addrLen = sizeof(destAddr);
        bzero(&destAddr, sizeof(destAddr));
        destAddr.sin_family = AF_INET;
        destAddr.sin_addr.s_addr = htonl(iRemoteIP);
        destAddr.sin_port = htons(iRemotePort);

        iRet = ::connect(iSockFd, (struct sockaddr *) &destAddr, addrLen) ;
        if (iRet < 0) {
                std::cout<<"ERROR on connect:"<<GetDotFormatIP(iRemoteIP)<<":"<<iRemotePort<<std::endl;
                return -1;
        }
        else{
		AddLogLineM(false, wxT("connect successfully"));
	}

return iSockFd;
}

/*Notify with EventMessage to the passive threads*/
void CFileTransfer::NotifyComplete(bool bSuccess){
	
	EventMessage msg;
        msg.SetReasonCode(FILE_RECV_REPORT);
	if(bSuccess){
		msg.SetStatus(EVT_SUCCESS);
	}
	else{
		msg.SetStatus(EVT_ERROR);
	}

	msg.SetFileSize(m_iFileSize);
	msg.SetFileName(m_sFileName);
	msg.SetFilePath(m_sFilePath);
	msg.SetSearchID(m_iSearchID);
	msg.SetFileType(m_8FileType);

	switch(m_8FileType){
	
	case FILE_GOSSIP:
			msg.SetRelatedModule(MODULE_GOSSIP);
			break;
	case FILE_OTHER:
			msg.SetRelatedModule(MODULE_KAD);
			break;
	default:
			break;
	}

	std::string sBuf;
	msg.serialize(sBuf);

	AddLogLineM(false, wxT("Notification for file:  ")+wxString(m_sFileName.c_str(), wxConvUTF8));

        int iStatus = ::send(GetSocketPair(), sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                AddLogLineM(false, wxT("Failed sending complete notification to kademlia"));
                perror("error in notify complete..");
        }

}

void CFileTransfer::NotifySenderReport(bool bSuccess){

        EventMessage msg;
	//msg.SetStatus(EVT_SEND_REPORT);
        msg.SetReasonCode(FILE_SEND_REPORT);

	if(bSuccess){
		msg.SetStatus(EVT_SUCCESS);
	}
	else{
		msg.SetStatus(EVT_ERROR);
	}

        msg.SetFileSize(m_iFileSize);
        msg.SetFileName(m_sFileName);
        msg.SetFilePath(m_sFilePath);
        msg.SetSearchID(m_iSearchID);
        msg.SetFileType(m_8FileType);


        std::string sBuf;
        msg.serialize(sBuf);

        AddLogLineM(false, wxT("Notification sender for file:  ")+wxString(m_sFileName.c_str(), wxConvUTF8));

        int iStatus = ::send(GetSocketPair(), sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                AddLogLineM(false, wxT("Failed sending complete notification to kademlia"));
                perror("error in notify complete..");
        }
                                                               
}

int CFileTransfer::SendHandShakePacket(bool bPush){

        EventMessage msg;

	msg.SetStatus(EVT_SUCCESS);
	if(bPush){
		m_8FileType = FILE_GOSSIP;

		msg.SetReasonCode(FILE_RECV_REQ);
		msg.SetFileType(FILE_GOSSIP);
		//we append the gossip id of the file with the handshake packet..
		if(m_bCheckGossipId){
			msg.SetReasonString(m_sGossipId);
		}
	}
	else{
		m_8FileType = FILE_OTHER;

		msg.SetReasonCode(FILE_SEND_REQ);
		msg.SetFileType(FILE_OTHER);
	}

	//send only the name part of the file, not the path
	wxString wxStr(m_sFileName.c_str(), wxConvUTF8);
	std::string sName  = std::string( wxFileNameFromPath(wxStr).mb_str() );

        msg.SetFileSize(m_iFileSize);
        msg.SetFileName(sName);
        msg.SetSearchID(-1);

        std::string sBuf;
        msg.serialize(sBuf);

        AddLogLineM(false, wxT("Sending HandShakePacket file:  ")+wxString(m_sFileName.c_str(), wxConvUTF8));
	std::cout<<"Sending Request: Upload= "<<sBuf.size()<<" Gossip ID= "<<m_sGossipId<<std::endl;

        int iStatus = ::send(m_iSockFd, sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                AddLogLineM(false, wxT("Failed sending HandShakePacket"));
                perror("error in send in HandShakePacket");
		return -1;
        }

return 0;
}

int CFileTransfer::WaitForHandShakePacketAck(){
	bool bPull;
	return WaitForHandShakePacket(bPull, true);
}

int CFileTransfer::WaitForHandShakePacket(bool& bPull, bool bWaitForAck){

int iReceived = -1;
uint8_t szBuffer[MAXSIZE];
struct pollfd pollFd[1];

memset(pollFd, 0 , sizeof(pollFd));

pollFd[0].fd = m_iSockFd;
pollFd[0].events = POLLIN;
pollFd[0].revents = 0;

bool bFlag = true;
int iCounter=0;


	while(bFlag){

		int iStatus = poll(pollFd, 1, POLL_TIMEOUT*4 );
		if( iStatus < 0){
			AddLogLineM(false, wxT("Poll failed in in waithandshakedata"));
			return -1;
		}
                else if(iStatus == 0)
                {
                                AddressPortPair p = GetPeerName(m_iSockFd);
                                if(bWaitForAck){
                                        AddLogLineM(false, wxT("Poll timeout in waitforpktACK; sock=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                                }
                                else{
                                        AddLogLineM(false, wxT("Poll timeout in waitforpkt; sock=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                                }
                                iCounter++;
                                if(iCounter>MAX_TIMEOUT_HANDSHAKE){
                                        if(bWaitForAck)
                                                AddLogLineM(false, wxT("Waited too long: terminating..waitforpktACK: port=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                                        else
                                                AddLogLineM(false, wxT("Waited too long: terminating..waitforpkt; port=")+ wxString::Format(wxT("%i:%i"),p.first, p.second));
                                        //bFlag = false;
                                        return -1;
                                }
                }
		//else if(pollFd[0].revents & POLLIN){//case POLLIN:
		else{
			bzero(szBuffer, MAXSIZE);
			iReceived = recv(m_iSockFd, (void*)szBuffer, MAXSIZE, 0 );
			const char* pBuf = (char*)szBuffer;

			if(0==iReceived){
				AddLogLineM(false, wxT("Failed to receive data..peer is down"));
				return -1;
			}

			if(iReceived<0) {
				AddLogLineM(false, wxT("Failed to refceived data "));
				return -1;
			}
			else{
				bFlag = false;
				if(bWaitForAck){
					AddLogLineM(false, wxT("Received ACK data: ")+ wxString::Format(wxT("%i"),iReceived));
					//should we check the payload?
					if(strncmp(pBuf,"NACK",4)==0){
						AddLogLineM(false, wxT("HandShake ACK received : peer is not interesed! "));
						return -2;
					}
					
					if(strncmp(pBuf,"READY",5)!=0){
						AddLogLineM(false, wxT("HandShake ACK received does not match with READY! "));
						return -1;
					}
				}
				else{
					int iRet = ParseHandShakePacket((char*)szBuffer, iReceived, bPull) ;
					if(iRet == -2 ){
						//we should send response to the sender..
						SendPacket(m_iSockFd, "NACK", 4 );
						return -1;
					}
					else if(iRet == -1){
						//we should send response to the sender..
						SendPacket(m_iSockFd, "UNKNOWN", 7 );
						return -1;
					}
					else{
					}
				}
			}
		}
#if 0
		else{
                                AddLogLineM(false, wxT("Poll returned error"));
				break;
		}
#endif

	}//while

return 0;
}

int CFileTransfer::ParseHandShakePacket(const char* szBuffer, int iReceived, bool& bPull){

	int iRead = 0;
	const char* pBuf = szBuffer;
	pBuf+=iRead;

	EventMessage msg;
	std::string sBuf(pBuf,iReceived-iRead);
	iRead += msg.deserialize(sBuf);

	m_sFileName = msg.GetFileName();
	m_iFileSize = msg.GetFileSize();
	m_8FileType = msg.GetFileType();
	//useful for gossip
	m_sGossipId = msg.GetReasonString();
	
	if(!m_sGossipId.empty()){
		std::cout<<"Received Request: Download= "<<iReceived<<" Gossip ID= "<<m_sGossipId<<"\tat time= "<<GetGMTime()<<std::endl;
	}
	

	switch(msg.GetReasonCode()){

	case  FILE_SEND_REQ:
			/*Peer wants me to send him a file*/
			bPull = false;
			break;
	case FILE_RECV_REQ:
			/*Peer wants me to receive a file from him*/
			bPull = true;
			break;
	default:
		std::cout<<"Invalid Reason Code in HandShake Packet"<<std::endl;
		return -1;
	}

	//now..check the gossip id
	//const std::string& sGossipId = msg.GetReasonString();

	if(m_bCheckGossipId && bPull){
		if( m_vGossipList.end() != find(m_vGossipList.begin(), m_vGossipList.end(), m_sGossipId) ){
			std::cout<<"Already have this file..not interested for another download: "<<m_sGossipId<<std::endl;
			return -2;
		}
	}

	if(iRead != iReceived){
		AddLogLineM(false, wxT("Received extra data "));
		 AddLogLineM(false, wxT("Received extra data: ")+ wxString::Format(wxT("%i"),iRead));
		pBuf +=iRead;
		m_sExtraData = std::string(pBuf, iReceived-iRead);
		std::cout<<"Extra data="<<m_sExtraData<<std::endl;
	}

return 0;
}
