#include<iostream>
#include<fstream>
#include<vector>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h> 
#include <error.h>
//#include <boost/uuid/uuid_io.hpp>

#include<stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/inotify.h>

#include "../kademlia/kademlia/Kademlia.h"
#include "../kademlia/kademlia/Prefs.h"
#include "Logger.h"
#include "../kademlia/utils/KadUDPKey.h"
#include "kademlia_interface.h"
#include "protocol/Protocols.h"
#include "kademlia/utils/KadClientSearcher.h"
#include "kademlia/routing/Maps.h"
#include "kademlia/routing/Contact.h"
#include "kademlia/kademlia/Search.h"

#include "KnownFile.h"
#include "ThreadTasks.h"
#include "SharedFileList.h"
#include "KnownFileList.h"
#include "Preferences.h"                // Needed for CPreferences
#include "SearchList.h"                 // Needed for CSearchList
#include "SearchFile.h"                 // Needed for CSearchFile
#include "PartFile.h"           // Needed for CPartFile
#include "Preferences.h"                // Needed for CPreferences
#include "FileTransfer.h"
#include "EventMessage.h"

#include "util.h"



#define MAXCHAR 1500
//#define POLL_TIMEOUT	10 //msec
#define POLL_TIMEOUT	1000 //msec
#define SOCKET_ERROR	-1 
//#define MAX_RECEIVER_THREAD_ALLOWED	5


using std::vector;
using std::cout;
using std::endl;
using std::string;
using std::ifstream;

CKadInterface::CKadInterface(string sFileName):
	m_iSockFd(SOCKET_ERROR),
	m_iDaemonSockFd(SOCKET_ERROR),
        m_iPeerSockFd(SOCKET_ERROR),
	m_sConfigFileName(sFileName),
        m_sClientName(),
        m_iClientTcpSockFd(SOCKET_ERROR),
	m_bBootstrapEnable(false),
	m_bKadLogEnable(false),
        m_uClientIP(0),
        m_uClientPort(0),
        m_uBootIP(0),
        m_uBootPort(0),
        m_Thread(0),
        m_uDaemonTcpPort(0),
        m_uClientTcpPort(0),
        knownfiles(0),
        m_sSharedDir(),
        m_sDownloadDir(),
        m_bMulticastEnable(false),
        m_bGossipEnable(false),
        m_sMulticastInterface(),
        m_uMulticastIP(0),
        m_uMulticastPort(0),
	m_pMulticast(0),
	m_bInit(false),
	m_bBootstrapped(false),
	m_iMaxGossip(-1),
	m_bDoVerifyBeforeGossip(true),
	m_iFileWatchFd(-1),
	m_iFileWatchWd(-1),
	m_iReceiverThreadCount(0),
	m_bDoQueryBeforeTransfer(true),
	m_bActNonCooperative(false),
	m_iMaxSenderThread(0),
	m_iMaxReceiverThread(0),
	m_bAlreadyReformed(false),
	m_bReformTopology(false),
	m_bReplaceFailedNodes(false),
	m_bAllowCriticalLevel(false)

{
        knownfiles  = new CKnownFileList();
        sharedfiles  = new CSharedFileList(knownfiles);
        searchlist      = new CSearchList();

}

CKadInterface::~CKadInterface(){
	cout<<"Deleting CKadInterface object"<<endl;
	delete knownfiles;
	knownfiles = NULL;

	delete sharedfiles;
	sharedfiles = NULL;

        delete searchlist;
        searchlist = NULL;

	if(!m_bInit)
		return;

	if(m_bGossipEnable){
		StopKad();
	}
	DestroySocket();
        if ( !m_filelist.empty() ) {
                for ( unsigned int i = 0; i < m_filelist.size(); i++ ) {
                        printf("\rSaving PartFile %u of %u", i + 1, (unsigned int)m_filelist.size());
                        fflush(stdout);
                        delete m_filelist[i];
                }
                printf("\nAll PartFiles Saved.\n");
        }

	delete m_pMulticast;
	m_pMulticast = NULL;

        if ( !m_filetrans.empty() ) {
                for ( unsigned int i = 0; i < m_filetrans.size(); i++ ) {
                        delete m_filetrans[i];
                }
        }


	ShutdownNetworkThread();
	int iRet = pthread_join(m_Thread, NULL);
	if(iRet!=0){
		std::cout<<"pthread_jon returns error\n";
	}

        close(m_iSockPairFd[0]);
        close(m_iSockPairFd[1]);

	inotify_rm_watch( m_iFileWatchFd, m_iFileWatchWd );
	close( m_iFileWatchFd );

}


int CKadInterface::Run(){


	int iStatus = ParseConfig();
	if(iStatus != 0 )
		return -1;

	MakeDirs();

	if(m_bKadLogEnable)
		SetLogger();

	iStatus = CreateSocket();
	if(iStatus != 0 )
		return -1;

	iStatus = CreateDaemonSocket();
	if(iStatus != 0 )
		return -1;
	iStatus = CreateClientTcpSocket();
	if(iStatus != 0 )
		return -1;

	iStatus = InitSocketPair();
	if(iStatus != 0 )
		return -1;

	/*Start Network Thread*/
	StartNetworkThread();

	if(m_bGossipEnable){
		/*Start Kademlia*/
		StartKad();

		/*Only for testing right now, later, we may need to remove it and check the nodes.dat file instead*/
#if 1
		if(m_bBootstrapEnable){
			m_bBootstrapped=true;
			BootstrapKad();
		}
#else
#endif
	}

	/*Configure shared files directory*/
        //sharedfiles->Reload();

	if(m_bMulticastEnable){

		m_pMulticast = new CMulticast(m_sMulticastInterface, m_uMulticastIP, m_uMulticastPort);
		CMulticast::SetSocketPair(GetSocketPair()[1]);

	}

	m_bInit = true;

return 0;
}

int CKadInterface::ParseConfig(){

        vector<string> vConf;
        char szBuf[MAXCHAR];
	string sTemp;


        ifstream fileHandler(m_sConfigFileName.c_str());

        if(!fileHandler.is_open())
        {
                cout<<"Unable to open Input file"<<endl;
                return -1;
        }

        do{
                bzero(szBuf,MAXCHAR);
                fileHandler.getline(szBuf,MAXCHAR);
                vConf.push_back(szBuf);
        }while(!fileHandler.eof());
        fileHandler.close();

	//Now, parse config file
	sTemp = GetString(vConf, "bootstrap_ip") ;
	if(sTemp.empty()) 
		return -1;
	StrTrim(sTemp);
	m_uBootIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

	sTemp = GetString(vConf, "bootstrap_port") ;
	if(sTemp.empty()) 
		return -1;
	m_uBootPort = atoi(sTemp.c_str());

	sTemp = GetString(vConf, "client_ip") ; 
	if(sTemp.empty()) 
		return -1;
	StrTrim(sTemp);
	m_uClientIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

	sTemp = GetString(vConf, "client_port");
	if(sTemp.empty()) 
		return -1;
	m_uClientPort = atoi(sTemp.c_str());
	
	sTemp = GetString(vConf, "client_name");
	m_sClientName = sTemp;

	sTemp = GetString(vConf, "bootstrap_enable");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
	if( sTemp.find("yes") != string::npos )
		m_bBootstrapEnable = true;
	else
		m_bBootstrapEnable = false;

	sTemp = GetString(vConf, "kad_log_enable");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
	std::cout<<"log="<<sTemp<<std::endl;
	if( sTemp.find("yes") != string::npos )
		m_bKadLogEnable = true;
	else
		m_bKadLogEnable = false;


	sTemp = GetString(vConf, "daemon_tcp_port");
        if(sTemp.empty())
                return -1;
        m_uDaemonTcpPort = atoi(sTemp.c_str());

        sTemp = GetString(vConf, "shared_dir");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
	SetSharedDir(sTemp);

        sTemp = GetString(vConf, "client_tcp_port");
        if(sTemp.empty())
                return -1;
        m_uClientTcpPort = atoi(sTemp.c_str());
	thePrefs::SetPort(m_uClientTcpPort);

        sTemp = GetString(vConf, "download_dir");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        SetDownloadDir(sTemp);

        sTemp = GetString(vConf, "multicast_enable");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        if( sTemp.find("yes") != string::npos )
                m_bMulticastEnable = true;
        else
                m_bMulticastEnable = false;


        sTemp = GetString(vConf, "gossip_enable");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        if( sTemp.find("yes") != string::npos )
                m_bGossipEnable = true;
        else
                m_bGossipEnable = false;

        sTemp = GetString(vConf, "multicast_interface");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
        m_sMulticastInterface = sTemp;

        sTemp = GetString(vConf, "multicast_ip");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
	m_uMulticastIP = ntohl( inet_addr(sTemp.c_str()) );

        sTemp = GetString(vConf, "multicast_port");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
        m_uMulticastPort = atoi(sTemp.c_str());


        sTemp = GetString(vConf, "max_gossip");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
	if(sTemp == "all")
		m_iMaxGossip = -1;
	else
		m_iMaxGossip = atoi(sTemp.c_str());
	std::cout<<"max_gossip="<<m_iMaxGossip<<std::endl;

        sTemp = GetString(vConf, "verify_before_gossip");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
        m_bDoVerifyBeforeGossip = (sTemp=="true") ? true: false;
	std::cout<<"verify_before_gossip="<<m_bDoVerifyBeforeGossip<<std::endl;
	//AddDebugLogLineM(false, logClientUDP, wxT("Parsing info from config file: BootStrap: ")  + Uint32_16toStringIP_Port(m_uBootIP, m_uBootPort) + wxT("Client: ") +Uint32_16toStringIP_Port(m_uClientIP, m_uClientPort) );


        sTemp = GetString(vConf, "query_before_transfer");
        if(sTemp.empty())
                return -1;
	StrTrim(sTemp);
        m_bDoQueryBeforeTransfer = (sTemp=="true") ? true: false;
	std::cout<<"query_before_transfer="<<m_bDoQueryBeforeTransfer<<std::endl;

        sTemp = GetString(vConf, "act_non_cooperative");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_bActNonCooperative = (sTemp=="true") ? true: false;
	std::cout<<"act_non_cooperative="<<m_bActNonCooperative<<std::endl;

        sTemp = GetString(vConf, "pull_frequency");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
	m_iPullFrequency = atoi(sTemp.c_str()); //in seconds
	std::cout<<"pull frequency= "<<m_iPullFrequency<<std::endl;
	Kademlia::CKademlia::SetPullFrequency(m_iPullFrequency);

        sTemp = GetString(vConf, "randomize_contacts");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_bRandomizeContact = (sTemp=="true") ? true: false;
	std::cout<<"randomize_contacts="<<m_bRandomizeContact<<std::endl;

        sTemp = GetString(vConf, "max_sender_thread");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_iMaxSenderThread = atoi(sTemp.c_str());
	std::cout<<"max_sender_thread="<<m_iMaxSenderThread<<std::endl;

        sTemp = GetString(vConf, "max_receiver_thread");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_iMaxReceiverThread = atoi(sTemp.c_str()); 


        sTemp = GetString(vConf, "reform_topology");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_bReformTopology = (sTemp=="true") ? true : false;
	std::cout<<"m_bReformTopology="<<m_bReformTopology<<std::endl;

	if(m_bReformTopology)
		ReformTopology();

        sTemp = GetString(vConf, "replace_failed_nodes");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_bReplaceFailedNodes = (sTemp=="true") ? true : false;
	std::cout<<"m_bReplaceFailedNodes="<<m_bReplaceFailedNodes<<std::endl;
//
#if 0
        sTemp = GetString(vConf, "degree_lb");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_iDegreeLB = atoi(sTemp.c_str());

        sTemp = GetString(vConf, "degree_ub");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_iDegreeUB = atoi(sTemp.c_str());
#endif

        sTemp = GetString(vConf, "reset_reform_topology");
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        bool bResetReform = (sTemp=="true") ? true : false;
	if(bResetReform)
		m_bAlreadyReformed=false;
        std::cout<<"bResetReform="<<bResetReform<<std::endl;

        sTemp = GetString(vConf, "allow_critical_level");
        if(sTemp.empty()){
        //        return -1;
	}
        StrTrim(sTemp);
        bool m_bAllowCriticalLevel = (sTemp=="true") ? true : false;
        std::cout<<"m_bAllowCriticalLevel= "<<m_bAllowCriticalLevel<<std::endl;


return 0;
}


void CKadInterface::StartKad()
{
        if (!Kademlia::CKademlia::IsRunning() ) {
		Kademlia::CKademlia::Start(m_iSockFd);
        } 
}

void CKadInterface::StopKad()
{
        // Stop Kad if it's running
        if (Kademlia::CKademlia::IsRunning()) {
                Kademlia::CKademlia::Stop();
        }
}

void CKadInterface::BootstrapKad(uint32_t ip, uint16_t port){

	m_uBootIP = ip;
	m_uBootPort = port;

	m_bBootstrapped = true;
	BootstrapKad();

}

void CKadInterface::BootstrapKad()
{
        if (!Kademlia::CKademlia::IsRunning()) {
                Kademlia::CKademlia::Start();
        }
	//IP and port are in host byte order.
	//just before ::sendto(), we will make it network byte order
        Kademlia::CKademlia::Bootstrap(m_uBootIP, m_uBootPort, true);
}


/*Should we use same socket for sending and receiving?*/
/*Right now, same sock for sending and receiving*/
void CKadInterface::StartNetworkThread(){

int iRet;

	iRet = pthread_create(&m_Thread, NULL, CKadInterface::ThreadCallback, (void *)this);
	if (iRet){
		 cout<<"ERROR; return code from pthread_create() is "<<iRet<<endl;
		 exit(-1);
	}
}

void* CKadInterface::ThreadCallback(void *arg){
int iFd;

	CKadInterface& kadInt = *(CKadInterface*)arg;

	iFd = kadInt.GetSockFd();
	/*
	5 sock fd: 1 for kad, 1 for daemon tcp listen, 1 for new tcp conn for daemon, 1 for clientTcpService for file transfer.
	1 for socket pair.. to notify file tranfer complete from file tranfer thread to network thread.
	*/

	/*
	Adding another fd: for monitoring configuration file 
	*/
   	struct pollfd pollFd[6];

	pollFd[0].fd = iFd;
	pollFd[0].events = POLLIN;

	pollFd[1].fd = kadInt.GetDaemonSockFd();
	pollFd[1].events = POLLIN;

	pollFd[2].fd = -1;
	pollFd[2].events = POLLIN;

	pollFd[3].fd = kadInt.GetClientTcpSockFd();
	pollFd[3].events = POLLIN;

	pollFd[4].fd = kadInt.GetSocketPair()[0];
	pollFd[4].events = POLLIN;

	pollFd[5].fd = kadInt.GetFileWatchFd();
	pollFd[5].events = POLLIN;

	while (1){ 
		int iStatus = poll(pollFd, 6, POLL_TIMEOUT );

		if( iStatus < 0){ 
			cout<<"poll failed in main loop"<<iStatus<<endl;; 
		}

		if( 0 == iStatus){
		
			kadInt.OnTimer();
		}
		else if(pollFd[0].revents & POLLIN){
			kadInt.CheckEvent(pollFd[0].fd); 
		}
		else if(pollFd[1].revents & POLLIN){
			AddLogLineM(false, wxT("TCP request arrived"));

			struct sockaddr_in FromAddr;
			socklen_t FromAddrLen=sizeof(FromAddr);

			int iClientSock = accept(pollFd[1].fd, (struct sockaddr *) &FromAddr, &FromAddrLen);
			if(iClientSock<0){
				std::cout<<"Error in accept\n";
			}
			else{
				uint32_t uIP = ntohl(FromAddr.sin_addr.s_addr);
				uint16_t uPort = ntohs(FromAddr.sin_port);
				DebugSend("Request from Daemon", uIP, uPort);
			}

			pollFd[2].fd = iClientSock;
			kadInt.SetPeerSockFd(iClientSock);

		}
		else if(pollFd[2].revents & POLLIN){
			int iRet = kadInt.CheckDaemonEvent(pollFd[2].fd);//kadInt.OnCommandReceived(pollFd[2].fd);
			if(0 == iRet){
				/*Shutdown by peer*/
				pollFd[2].fd = -1;
				kadInt.SetPeerSockFd(-1);
			}	
		}
		else if(pollFd[3].revents & POLLIN){
			AddLogLineM(false, wxT("req arrived in client tcp sock"));

			struct sockaddr_in FromAddr;
                        socklen_t FromAddrLen=sizeof(FromAddr);

                        int iClientSock = accept(pollFd[3].fd, (struct sockaddr *) &FromAddr, &FromAddrLen);
                        if(iClientSock<0){
                                std::cout<<"Error in accept\n";
                        }
                        else{
                                uint32_t uIP = ntohl(FromAddr.sin_addr.s_addr);
                                uint16_t uPort = ntohs(FromAddr.sin_port);
                                DebugSend("Request from peer for a TcpService", uIP, uPort);
				std::cout<<"Accepted File Transfer Connection From : "<<GetDotFormatIP(uIP)<<":"<<uPort<<std::endl;
				std::cout<<"my accepted file transfer sock="<<iClientSock<<std::endl;

				bool bRet = kadInt.IncReceiverThreadCount();
				if(!bRet){
					std::cout<<"Receiver Thread count reached to max; Rejecting the connection\n";
					close(iClientSock);
				}
				else{
					//activate file tranfer receiver thread
					CFileTransfer* filetrans = new CFileTransfer(iClientSock,kadInt.GetSharedDir(), kadInt.GetDownloadDir());
					filetrans->SetSocketPair(kadInt.GetSocketPair()[1]);
					filetrans->SetGossipIDCheck(kadInt.m_bDoQueryBeforeTransfer);
					if(kadInt.m_bDoQueryBeforeTransfer){
						filetrans->SetGossipList(kadInt.m_vGossipList);
					}
					filetrans->HandShake(true);
					kadInt.m_filetrans.push_back(filetrans);

				}

                        }

		}
		else if(pollFd[4].revents & POLLIN){
			AddLogLineM(false, wxT("in sock pair"));
			int iRet = kadInt.CheckFileTransferStatus(pollFd[4].fd);
			if(iRet == -1){
				break;
			}

			
		}
                else if(pollFd[5].revents & POLLIN){
                        AddLogLineM(false, wxT("File Modify event occured"));
                        int iRet = kadInt.ReadModifiedConfigFile(pollFd[5].fd);
                        if(iRet == -1){
				std::cout<<"Error in rereading config file\n";
                               // break;
                        }

                }
		else{
			std::cout<<"Error in pollfd: "<<iStatus<<std::endl;
			break;
		}
	}

        int iStatus = 0;
        pthread_exit((void*)&iStatus);


return NULL;
}

int CKadInterface::CheckFileTransferStatus(int iFd){

char szBuffer[MAXCHAR];

	bzero(szBuffer,MAXCHAR);

#if 0
        int iReceived = recv(iFd, (void*)szBuffer, MAXCHAR, 0 );
#else
        sockaddr_in sFrom;
        memset((void*)&sFrom, 0, sizeof(sFrom));

        int fromlen = sizeof(sFrom);

	//this is a socketpair recvfrom..why are we getting the address?
        int iReceived = recvfrom(iFd, (void*)szBuffer, MAXCHAR, 0,
			(sockaddr*)&sFrom, (socklen_t*)&fromlen );

        uint32_t uAddr = ntohl(sFrom.sin_addr.s_addr);
        uint16_t uPort = ntohs(sFrom.sin_port);

#endif
	const char* pBuf = szBuffer;

        if(0==iReceived){
                AddLogLineM(false, wxT("sock pair is down: "));
        }

        if(iReceived<0) {
                AddLogLineM(false, wxT("Failed to refceived data from sock pair!"));
        }
        else{
		int iRead = 0;
		while(iRead<iReceived){

			const char* pBuf = szBuffer;
			pBuf+=iRead;

			EventMessage msg;
			std::string sBuf(pBuf,iReceived-iRead);
			iRead += msg.deserialize(sBuf);

#if 0
			if(msg.GetFileType() == FILE_GOSSIP){
				m_iReceiverThreadCount--;
				if(m_iReceiverThreadCount<0)
					m_iReceiverThreadCount=0;
			}
#endif

			if(msg.GetStatus() == EVT_SHUTDOWN){
					AddLogLineM(false, wxT("Shutting down Networking Thread!"));
					return -1;
			}

			switch(msg.GetReasonCode()){

			case FILE_SEND_REPORT:


				if(msg.GetFileType() == FILE_GOSSIP){
					bool bRet = ( msg.GetStatus() == EVT_ERROR ) ? false: true;
					HandleSenderReport(msg.GetFileName(), bRet);
				}
				if(msg.GetFileType() == FILE_OTHER){
							if(EVT_SUCCESS == msg.GetStatus()){
								int iSearchID = msg.GetSearchID();
								Kademlia::CSearchManager::StopSearch(iSearchID, false);
								NotifyApp(msg.GetFilePath(), msg.GetFileType(), msg.GetRelatedModule());
								//now, we should publish this certificate..
								PublishFile(msg.GetFilePath());
							}
							else if(EVT_ERROR == msg.GetStatus()){
								int iSearchID = msg.GetSearchID();
								AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("Error received in file tranfer..searchID= (%i)"),iSearchID));
								RetrySource(iSearchID);
							}
							else{
							}
				}
				break;

			case FILE_RECV_REPORT:

				std::cout<<"# of receiver thereads:"<<m_iReceiverThreadCount<<std::endl;

				m_iReceiverThreadCount--;
				if(m_iReceiverThreadCount<0)
					m_iReceiverThreadCount=0;

				if(msg.GetFileType() == FILE_GOSSIP){
					if(msg.GetStatus() == EVT_SUCCESS){
						AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("Success received in file tranfer..searchID= (%i)"),msg.GetSearchID()));
						ProcessTcpGossipPacket(msg.GetFilePath(), uAddr, uPort, msg.GetRelatedModule());
					}
					else{
						AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("Failure in file tranfer..searchID= (%i)"),msg.GetSearchID()));
						//now, we should retry...to download again..
					}
				}
				break;
			default:
				std::cout<<"Unknown Reason code:"<<std::endl;
				break;
				
			}

#if 0
			switch(msg.GetStatus()){
				case  EVT_SEND_REPORT:
						{
							HandleSenderReport();
							break;
						}

				case  EVT_SUCCESS:
						{
							int iSearchID = msg.GetSearchID();
							Kademlia::CSearchManager::StopSearch(iSearchID, false);

							switch(msg.GetFileType()){

							case FILE_GOSSIP:
								ProcessTcpGossipPacket(msg.GetFilePath(), uAddr, uPort, msg.GetRelatedModule());
								break;

							case FILE_OTHER:
								NotifyApp(msg.GetFilePath(), msg.GetFileType(), msg.GetRelatedModule());

								//check if this is pollution attack response
								//Kademlia::CKademlia::GetValidatePublish()->CheckValidation(msg.GetFilePath());	
								break;

							default:
								std::cout<<"Unknown file type:"<<std::endl;
								break;
							}


							AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("Success received in file tranfer..searchID= (%i)"),iSearchID));

							//we should delete the map with the search id
							break;
						}

				case  EVT_ERROR:
						{
							int iSearchID = msg.GetSearchID();
							AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("Error received in file tranfer..searchID= (%i)"),iSearchID));
							RetrySource(iSearchID);
							break;
						}
				case EVT_SHUTDOWN:
						{
							AddLogLineM(false, wxT("Shutting down Networking Thread!"));
							return -1;
						}
				default:
                                        AddLogLineM(false, wxT("Unknown notification from file transfer thread..!"));
                                        break;

			}
#endif

		}

	}

return 0;
}

void CKadInterface::NotifyApp(const std::string& sFileName, uint8_t filetype, uint8_t relatedModule){

	std::cout<<"in notify appp in kad"<<std::endl;
}
void CKadInterface::ProcessTcpGossipPacket(const std::string& sFileName, uint32 ip, uint16 port, uint8_t module){

	std::cout<<"in process gossip packet in kad"<<std::endl;
}


void CKadInterface::CheckEvent(int iFd){

byte buf[MAXCHAR];
int iSize;
int iReceived;
struct sockaddr_in FromAddr;
socklen_t FromAddrLen=sizeof(FromAddr);
uint32	uIP;
uint16  uPort;


	bzero(buf,MAXCHAR);

	iReceived = recvfrom(iFd, buf, MAXCHAR, 0, (struct sockaddr *) &FromAddr, &FromAddrLen);

	if(iReceived<0) {
		cout<<"Failed to refceived data!"<<endl;
	}
	else{
//		cout<<"received: "<<iReceived<<endl;
//		AddDebugLogLineM(false, logClientUDP, wxT("Packet received from: ")  + Uint32_16toStringIP_Port(ntohl(FromAddr.sin_addr.s_addr), ntohs(FromAddr.sin_port)) );


	}
	//received IP and Port are in network byte order..
	//so..we keep them in host byte order..
	//before sending in ::sendto(), we will convert it to network byte order
	uIP = ntohl(FromAddr.sin_addr.s_addr);
	uPort = ntohs(FromAddr.sin_port);

//	DebugSend("Dont't know", uIP, uPort);
	OnPacketReceived(uIP, uPort, (byte*)buf, iReceived );

}

int CKadInterface::CheckDaemonEvent(int iFd){

return OnCommandReceived(iFd);
}

void CKadInterface::OnPacketReceived(uint32 ip, uint16 port, byte* buffer, size_t length)
{
        wxCHECK_RET(length >= 2, wxT("Invalid packet."));

        uint8_t *decryptedBuffer=buffer;;
        uint32_t receiverVerifyKey;
        uint32_t senderVerifyKey;
        int packetLen = length;
        uint8_t protocol = decryptedBuffer[0];
        uint8_t opcode   = decryptedBuffer[1];
	//printf("In OnPacketReceived: opcode=%02x\n",opcode);
#if 0
	GossipPacket gPack;
#endif


        if (packetLen >= 1) {

		 try {
                        switch (protocol) {
                                case OP_EMULEPROT:
                                        break;

				/*Kademlia Packet*/
                                case OP_KADEMLIAHEADER:
					//cout<<"Kademlia header arrived!"<<endl;
                                        if (packetLen >= 2) {
                                                Kademlia::CKademlia::ProcessPacket(decryptedBuffer, packetLen, ip, port, (Kademlia::CPrefs::GetUDPVerifyKey(ip) == receiverVerifyKey), Kademlia::CKadUDPKey(senderVerifyKey, GetPublicIP(false)));
                                        } else {
                                                throw wxString(wxT("Kad packet too short"));
                                        }

                                        break;

                                case OP_KADEMLIAPACKEDPROT:
						cout<<"Should not occur OP_KADEMLIAPACKEDPROT"<<endl;
                                        break;

#if 0
				/*Gossip Packet*/
				case GOSSIP_PROTO:
					cout<<"Gossip header arrived!: len="<<length<<endl;
					memcpy(&gPack,decryptedBuffer, length);
					ProcessGossipPacket(gPack, ip, port, length);
					break;
#endif

                                default:
                                        AddDebugLogLineM(false, logClientUDP, wxString::Format(wxT("Unknown opcode on received packet: 0x%x"), protocol));
                        }
                } catch (const wxString& e) {
                        AddDebugLogLineM(false, logClientUDP, wxT("Error while parsing UDP packet: ") + e);
		} catch (const CInvalidPacket& e) {
                        AddDebugLogLineM(false, logClientUDP, wxT("Invalid UDP packet encountered: ") + e.what());
                } catch (const CEOFException& e) {
                        AddDebugLogLineM(false, logClientUDP, wxT("Malformed packet encountered while parsing UDP packet: ") + e.what());
                }

	}

}



void CKadInterface::SetLogger(){
// Load debug-categories
#ifdef __DEBUG__
        int count = CLogger::GetDebugCategoryCount();

        for ( int i = 0; i < count; i++ ) {
                const CDebugCategory& cat = CLogger::GetDebugCategory( i );

                //bool enabled = false;
                bool enabled = true;
                //cfg->Read( wxT("/Debug/Cat_") + cat.GetName(), &enabled );

                CLogger::SetEnabled( cat.GetType(), enabled );
                //std::cout<<cat.GetName().mb_str()<<" is "<<CLogger::IsEnabled(cat.GetType())<<std::endl;;
        }
#endif
}

int CKadInterface::CreateSocket(){

struct sockaddr_in kAddr;


	m_iSockFd = ::socket(AF_INET, SOCK_DGRAM, 0);
	if (m_iSockFd == -1){
		cout<<"Could not create socket."<<endl;
		return -1;
	}
	reuse(m_iSockFd);

	bzero((char *) &kAddr, sizeof(kAddr));
	kAddr.sin_family = AF_INET;
	kAddr.sin_addr.s_addr = INADDR_ANY;
	kAddr.sin_port = htons(m_uClientPort);

	if (bind(m_iSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr)) < 0) {
		cout<<"ERROR on binding in kad sock"<<endl;
		return -1;
	}


return 0;
}

int CKadInterface::CreateDaemonSocket(){
struct sockaddr_in kAddr;
int iRet;

        m_iDaemonSockFd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (m_iDaemonSockFd == -1){
                cout<<"Could not create tcp socket."<<endl;
                return -1;
        }
	reuse(m_iDaemonSockFd);

        bzero((char *) &kAddr, sizeof(kAddr));
        kAddr.sin_family = AF_INET;
        kAddr.sin_addr.s_addr = INADDR_ANY;
        kAddr.sin_port = htons(m_uDaemonTcpPort);

        iRet = bind(m_iDaemonSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr));

	if(iRet<0){
                cout<<"ERROR on binding in tcp sock"<<endl;
                return -1;
        }

	/*Listen for maximum one client*/
	iRet = listen(m_iDaemonSockFd, 1) ;
	if(iRet < 0 ){
		std::cout<<"Error in listen\n";
	}

return 0;

}

int CKadInterface::CreateClientTcpSocket(){
struct sockaddr_in kAddr;
int iRet;

        m_iClientTcpSockFd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (m_iClientTcpSockFd == -1){
                cout<<"Could not create tcp socket."<<endl;
                return -1;
        }
        reuse(m_iClientTcpSockFd);

        bzero((char *) &kAddr, sizeof(kAddr));
        kAddr.sin_family = AF_INET;
        kAddr.sin_addr.s_addr = INADDR_ANY;
        kAddr.sin_port = htons(m_uClientTcpPort);

        iRet = bind(m_iClientTcpSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr));

        if(iRet<0){
                cout<<"ERROR on binding in client tcp sock"<<endl;
                return -1;
        }

        /*Listen for maximum one client*/
        iRet = listen(m_iClientTcpSockFd, 1) ;
        if(iRet < 0 ){
                std::cout<<"Error in listen\n";
        }

return 0;

}

void CKadInterface::DestroySocket()
{
       if(m_iSockFd != SOCKET_ERROR)
       {
              ::close(m_iSockFd);
              m_iSockFd = SOCKET_ERROR;
       }

       if(m_iDaemonSockFd != SOCKET_ERROR)
       {
              ::close(m_iDaemonSockFd);
              m_iDaemonSockFd = SOCKET_ERROR;
       }
       if(m_iClientTcpSockFd != SOCKET_ERROR)
       {
              ::close(m_iClientTcpSockFd);
              m_iClientTcpSockFd = SOCKET_ERROR;
       }
}




//helper functions
uint32 CKadInterface::GetPublicIP(bool ignorelocal) 
{
	if (Kademlia::CKademlia::IsConnected() && Kademlia::CKademlia::GetIPAddress() ) {
		return wxUINT32_SWAP_ALWAYS(Kademlia::CKademlia::GetIPAddress());
	} else {
		return  0 ;
	}
}


void CKadInterface::StrTrim(string& str)const{

	string::size_type pos = str.find_last_not_of(' ');
	if(pos != string::npos) {
		str.erase(pos + 1);
		pos = str.find_first_not_of(' ');
		if(pos != string::npos) 
			str.erase(0, pos);
	}
	else 
		str.erase(str.begin(), str.end());
}


void CKadInterface::OnTimer(){

	// Publish files to server if needed.
	 sharedfiles->Process();
	

	 if( Kademlia::CKademlia::IsRunning() ) {

		Kademlia::CKademlia::Process();

#ifdef SAZZAD
		if(Kademlia::CKademlia::GetPrefs()->HasLostConnection()) {
			StopKad();
			clientudp->Close();
			clientudp->Open();
			if (thePrefs::Reconnect()) {
				StartKad();
			}
		}
#endif
	}

#define MAX_LIMIT 20

//stopping sending contact list for now..
#if 1
	static int iContactCount=0;
	//if(iContactCount == MAX_LIMIT*3*10){ //10 mins
	if(iContactCount == MAX_LIMIT*3*20){ //20 mins
		std::cout<<"Sending all contacts\n";
		iContactCount=0;
		std::string sList;
		GetAllContacts(sList);
		std::cout<<"Contact List at:"<<GetGMTime()<<"# "<<sList<<endl;
		//don't send contacts right now..
		//NotifyAllContacts(sList);
	}
	else{
		iContactCount++;
	}
#endif

//adding # of contact check


	if(GetContactSize()==0){
		static int iCount=0;
		iCount++;
		if(iCount==MAX_LIMIT){
			iCount=0;
			NotifyZeroContact();
		}
		
	}

}

int CKadInterface::GetContactSize(){

        Kademlia::ContactList contacts;
        if( Kademlia::CKademlia::IsRunning() ) {
                Kademlia::CKademlia::GetAllContacts(contacts);
                Kademlia::ContactList::size_type numContacts = contacts.size();
                //numContacts = std::min<ContactList::size_type>(numContacts, CONTACT_FILE_LIMIT); // safety precaution, should not be above
		return numContacts;
	}

return 0;
}

void CKadInterface::GetAllContacts(std::string& sList){

	Kademlia::ContactList contacts;
	if( Kademlia::CKademlia::IsRunning() ) {
		Kademlia::CKademlia::GetAllContacts(contacts);
		Kademlia::ContactList::size_type numContacts = contacts.size();
		//numContacts = std::min<ContactList::size_type>(numContacts, CONTACT_FILE_LIMIT); // safety precaution, should not be above

		if (numContacts <= 0) {
			AddLogLineM(false, wxString::Format(wxPLURAL("Only %d Kad contact available, nodes.dat not written", "Only %d Kad contacts available, nodes.dat not written", numContacts), numContacts));
			return;
		}

		if(m_bAlreadyReformed){
			//we will report only the predefined graph
			for(int i=0;i<m_vContacts.size(); ++i){
				Kademlia::CContact *c = m_vContacts[i];
				char buf[20]={0};
				sprintf(buf,"%u : ",c->GetIPAddress());
				sList+=buf;
			}
			return;
		}
		for (Kademlia::ContactList::const_iterator it = contacts.begin(); it != contacts.end(); ++it) {

			Kademlia::CContact *c = *it;

			char buf[20]={0};
			sprintf(buf,"%u : ",c->GetIPAddress());
			sList+=buf; 
		}

	}

}

void CKadInterface::PublishFile(const std::string& sFile){

	if( !Kademlia::CKademlia::IsRunning() ) {
		return;
	}

	wxString src = wxString(sFile.c_str(), wxConvUTF8);

	size_t pos = sFile.find_last_of('/');

	std::string sFileName(sFile);

	if(pos != std::string::npos){
		sFileName = sFile.substr(pos+1);
	}

	std::string sDest = m_sSharedDir+"/"+sFileName;
	std::cout<<"dest file: "<<sDest<<std::endl;
	wxString dest = wxString(sDest.c_str(), wxConvUTF8);
	
	wxCopyFile(src, dest);

        sharedfiles->Reload();
	
}

void CKadInterface::FindNodeByID(const uint8_t* nodeID){
	//Kademlia::CDerivedKadClientSearcher requester;
	//Kademlia::CKademlia::FindIPByNodeID(*(dynamic_cast<CKadClientSearcher*>(requester)), nodeID);
	//Kademlia::CKademlia::FindIPByNodeID(requester, nodeID);
	//SearchKeyKad("config");
	//SearchKeyKad("ee2d0def-678e-4bde-8452-2038c3a3649c.pem");
	//SearchKeyKad("14c2fa65-8f38-4d5f-bb2e-6b0da3d48cd9.pem");
}

string CKadInterface::GetString(const vector<string>& vConf, const string& sLine){

        vector<string>::const_iterator vIter;
        string sTemp;
        int iPos=-1;

        for(vIter=vConf.begin(); vIter!=vConf.end(); vIter++){
                sTemp = *vIter;
                iPos = sTemp.find(sLine);
                if(iPos == string::npos)
                        continue;

                iPos = sTemp.find("=");
                return sTemp.substr(iPos+1);
        }

        //AddDebugLogLineM(false, logClientUDP, wxChar(sLine.c_str()) );
        //cout<<sLine<<endl;
        return "";

}

void CKadInterface::CleanUp(){
}

void CKadInterface::SetSharedDir(const std::string& sPath){

#if 0
	/*First make it absolute*/
	wxString wxStr(sPath.c_str(), wxConvUTF8);
        wxFileName wxFile(wxStr);
        if(!wxFile.IsAbsolute()){
                wxFile.MakeAbsolute();
                m_sSharedDir = wxFile.GetLongPath().mb_str();
        }
	else
		m_sSharedDir = sPath;
#endif
	m_sSharedDir = MakeAbsPath(sPath); 
	std::cout<<"Shared path="<<m_sSharedDir<<std::endl;
	sharedfiles->SetC2SharedPath(m_sSharedDir);
}

void CKadInterface::SetDownloadDir(const std::string& sPath){

	m_sDownloadDir = MakeAbsPath(sPath); 
}

std::string CKadInterface::MakeAbsPath(const std::string& sPath){

        std::string sAbsPath;

        wxString wxStr(sPath.c_str(), wxConvUTF8);
        wxFileName wxFile(wxStr);
        if(!wxFile.IsAbsolute()){
                wxFile.MakeAbsolute();
                sAbsPath = wxFile.GetLongPath().mb_str();
        }
        else{
                sAbsPath = sPath;
        }

return sAbsPath;
}




void CKadInterface::SearchKeyKad(const std::string& sSearchStr){

	if( !Kademlia::CKademlia::IsRunning() ) {
		return;
	}
	if(!m_bBootstrapped){
		std::cout<<"Not bootstrapped so far"<<std::endl;
		return;
	}
        static uint32 m_nSearchID = 0;
        m_nSearchID++;


	uint32 real_id = m_nSearchID;
	SearchType search_type = KadSearch;
        CSearchList::CSearchParams params;
	params.searchString = wxString(sSearchStr.c_str(), wxConvUTF8);
        params.typeText = wxEmptyString;

	//! The filename extension. May be empty.
	//SAZZAD: need to be changed with ".pem" for certificates
        params.extension = wxEmptyString;

	//! The smallest filesize in bytes to accept, zero for any.
        params.minSize = 0 ;

	//! The largest filesize in bytes to accept, zero for any.
	params.maxSize = 1500;


        wxString error = searchlist->StartNewSearch(&real_id, search_type, params);
        if (!error.IsEmpty()) {
                // Search failed / Remote in progress
		std::cout<<"Search Failed"<<std::endl;
	}

	AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Start a new search for '%s'")) % params.searchString);
	AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("searchID= (%i)"),m_nSearchID));

}

void CKadInterface::SearchSrcKad(CSearchFile* toadd){

                //if( GetMaxSourcePerFileUDP() > GetSourceCount()){
                if( 1 ){

                        //Once we can handle lowID users in Kad, we remove the second IsConnected
#ifdef SAZZAD
                        if (theApp->downloadqueue->DoKademliaFileRequest() && (Kademlia::CKademlia::GetTotalFile() < KADEMLIATOTALFILE) && (dwCurTick > m_LastSearchTimeKad) &&  Kademlia::CKademlia::IsConnected() && theApp->IsConnected() && !IsStopped()){
                                //Kademlia
#else
#define ADD_BY_SAZZAD 5
//need to see the impact of this condition

                        //if ( (Kademlia::CKademlia::GetTotalFile() < KADEMLIATOTALFILE) && Kademlia::CKademlia::IsConnected() ){
                        if ( (Kademlia::CKademlia::GetTotalFile() < KADEMLIATOTALFILE+ADD_BY_SAZZAD) && Kademlia::CKademlia::IsConnected() ){
#endif
                                //theApp->downloadqueue->SetLastKademliaFileRequest();
				CPartFile* partfile = new CPartFile(toadd);

#if 0
				wxString file = partfile->GetFileName().GetPrintableString();
				std::string sFileName(file.mb_str());
				cout<<"in search: file="<<sFileName<<endl;
				cout<<"size="<<partfile->GetFileSize()<<endl;
#endif


                                Kademlia::CUInt128 kadFileID(toadd->GetFileHash().GetHash());

                                Kademlia::CSearch* pSearch = Kademlia::CSearchManager::PrepareLookup(Kademlia::CSearch::FILE, true, kadFileID);
                                AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Preparing a Kad Search for '%s'")) % toadd->GetFileName());

				//check if we need to get this file or not

                                if (pSearch) {
                                        AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Kad lookup started for '%s'")) % toadd->GetFileName());

                                        partfile->SetKadFileSearchID(pSearch->GetSearchID());
					//SAZZAD: need to add mutex..
					//wxMutexLocker lock(m_mutex);
					m_filelist.push_back( partfile );
                                }

                        }
                } else {
                }


}


CPartFile* CKadInterface::GetFileByID(const CMD4Hash& filehash) const
{
        //wxMutexLocker lock( m_mutex );//SAZZAD, need to add..

        for ( uint16 i = 0; i < m_filelist.size(); ++i ) {
                if ( filehash == m_filelist[i]->GetFileHash()) {
                        return m_filelist[ i ];
                }
        }

        return NULL;
}

CPartFile* CKadInterface::GetFileByKadFileSearchID(uint32 id) const
{
        //wxMutexLocker lock( m_mutex ); //SAZZAD, need to add later..

        for ( uint16 i = 0; i < m_filelist.size(); ++i ) {
                if ( id == m_filelist[i]->GetKadFileSearchID()) {
                        return m_filelist[ i ];
                }
        }

        return NULL;
}

void CKadInterface::AddSource(uint32_t searchID, uint32_t ip, uint16_t port){


	AddrPortPair p(ip,port);
	std::map<int, std::vector<AddrPortPair> >::iterator mIter;
	mIter = m_SourceMap.find(searchID);
	if(mIter == m_SourceMap.end()){
		std::vector<AddrPortPair> v;
		v.push_back(p);
		m_SourceMap.insert(std::pair<int, std::vector<AddrPortPair> >(searchID, v));
		AddDebugLogLineM(false, logKadSearch, wxT("Inserting Src in m_Srcmap"));
	}
	else{
		std::vector<AddrPortPair>& vSrc = mIter->second;
		std::vector<AddrPortPair>::iterator vIter;

		vIter = find(vSrc.begin(), vSrc.end(), p);
		if(vIter == vSrc.end() ){
			vSrc.push_back(p);
			AddDebugLogLineM(false, logKadSearch, wxT("Inserting Src in m_Srcmap"));
		}
		
	}

}

void CKadInterface::RetrySource(int iSearchID){

	if(iSearchID == -1)
		return;

	std::map<int, std::vector<AddrPortPair> >::iterator mIter;	
	mIter = m_SourceMap.find(iSearchID);
	if(mIter==m_SourceMap.end()){
                AddDebugLogLineM(false, logKadSearch, wxT("Source list is empty"));
		return;
	}
	std::vector<AddrPortPair>& vSrc = mIter->second;
#if 0
	AddrPortPair p = vSrc.back();
	vSrc.pop_back();
#endif
	int iSockFd=-1;
	while(1){
		if(vSrc.empty()){
			std::cout<<"Source list is empty"<<std::endl;
			//remove the search id..
			CFileTransfer* pFile = CheckFileTransByID(iSearchID);
			if(pFile){
				pFile->SetSearchID(-1);
			}
			return;
		}

		std::vector<AddrPortPair>::iterator vIter = vSrc.begin();
		AddrPortPair p = *vIter;
		vSrc.erase(vIter);

		//connect to this port..
		iSockFd = CFileTransfer::ConnectPeer(p.first, p.second);
		if(iSockFd==-1){
			std::cout<<"Failed to create tcp service socket: "<<GetDotFormatIP(p.first)<<":"<<p.second<<std::endl;
			//return;
			continue;
		}
		else{
			std::cout<<"Connected Successfully to: "<<GetDotFormatIP(p.first)<<":"<<p.second<<std::endl;
			break;
		}
	}
	
        CFileTransfer* pFile = CheckFileTransByID(iSearchID);
	if(!pFile){
                AddDebugLogLineM(false, logKadSearch, wxT("This should not happen..file ptr is empty!"));
		return;
	}

	AddDebugLogLineM(false, logKadSearch, wxT("Retrying file transfer.."));
	//pFile->ReceiveFile(iSockFd);
	pFile->HandShake(false, false, iSockFd);


}

void CKadInterface::KademliaSearchFile(uint32_t searchID, const Kademlia::CUInt128* pcontactID, const Kademlia::CUInt128* pbuddyID, uint8_t type, uint32_t ip, uint16_t tcp, uint16_t udp, uint32_t buddyip, uint16_t buddyport, uint8_t byCryptOptions)
{
        AddDebugLogLineM(false, logKadSearch, wxString::Format(wxT("Search result sources (type %i)"),type));

        //Safety measure to make sure we are looking for these sources
        CPartFile* temp = GetFileByKadFileSearchID(searchID);
        if( !temp ) {
                AddDebugLogLineM(false, logKadSearch, wxT("This is not the file we're looking for..."));
                return;
        }
	printf("IP=%x\n",ip);
	printf("tcp=%d\n",tcp);
	printf("udp=%d\n",udp);

#if 0
	wxString file = temp->GetFileName().GetPrintableString();
	std::string sFileName(file.mb_str());
	cout<<"file="<<sFileName<<endl;
	cout<<"size="<<temp->GetFileSize();
#endif

	AddDebugLogLineM(false, logSearch,
                                CFormat(wxT("File %u: %s"))
                                        % temp->GetFileSize()
                                        % temp->GetFileName());

	std::cout<<"going to open file transfer for search id="<<searchID<<std::endl;
	if(NULL!=CheckFileTransByID(searchID)){
		std::cout<<"Already file transfer going on for this search id"<<std::endl;
		AddSource(searchID, ip, tcp);
		return;
	}

#if 0
	if(NULL!=CheckFileTransByHash(temp->GetFileHash())){
		std::cout<<"Already file transfer going on for this file hash"<<std::endl;
		return;
	}
#endif

	int iSockFd = CFileTransfer::ConnectPeer(ip, tcp);
	if(iSockFd==-1){
		std::cout<<"Failed to create tcp service socket"<<std::endl;
		return;
	}
	std::cout<<"Connected Successfully to: "<<GetDotFormatIP(ip)<<":"<<tcp<<std::endl;
	std::cout<<"my file transfer port="<<iSockFd<<std::endl;


	wxString file = temp->GetFileName().GetPrintableString();
	std::string sFileName(file.mb_str());
	int iSize = temp->GetFileSize();

	CFileTransfer* filetrans = new CFileTransfer(iSockFd,GetSharedDir(), GetDownloadDir(), searchID);
	filetrans->SetFileHash(temp->GetFileHash());
	filetrans->SetSocketPair(GetSocketPair()[1]);
	//filetrans->ReceiveFile(sFileName, iSize);
	filetrans->HandShake(false, false, sFileName, iSize);
	m_filetrans.push_back(filetrans);

}


CFileTransfer* CKadInterface::CheckFileTransByID(const int iSearchID) const
{
        //wxMutexLocker lock( m_mutex );//SAZZAD, need to add..

        for ( uint16 i = 0; i < m_filetrans.size(); ++i ) {
                if ( iSearchID == m_filetrans[i]->GetSearchID()) {
                        return m_filetrans[ i ];
                }
        }

        return NULL;
}
CFileTransfer* CKadInterface::CheckFileTransByHash(const CMD4Hash& fHash) const
{
        //wxMutexLocker lock( m_mutex );//SAZZAD, need to add..

        for ( uint16 i = 0; i < m_filetrans.size(); ++i ) {
                if ( fHash == m_filetrans[i]->GetFileHash()) {
                        return m_filetrans[ i ];
                }
        }

        return NULL;
}

int CKadInterface::InitSocketPair(){

        int rc = socketpair( AF_UNIX, SOCK_STREAM, 0, m_iSockPairFd );
        if ( rc < 0 ) {
                printf("Cannot open socketpair");
		return -1;
        }

return 0;
}

void CKadInterface::MakeDirs(){

	//make m_sSharedDir, m_sDownloadDir on startup

        if(!wxDirExists(wxString(m_sSharedDir.c_str(), wxConvUTF8))){
		if(!wxMkdir(wxString(m_sSharedDir.c_str(), wxConvUTF8)) ){
			std::cout<<"Falied to create Dir: "<<m_sSharedDir<<std::endl;
		}
        }
        if(!wxDirExists(wxString(m_sDownloadDir.c_str(), wxConvUTF8))){
		if(!wxMkdir(wxString(m_sDownloadDir.c_str(), wxConvUTF8)) ){
			std::cout<<"Falied to create Dir: "<<m_sDownloadDir<<std::endl;
		}
        }
        if(!wxDirExists(wxString(".tmp", wxConvUTF8))){
		if(!wxMkdir(wxString(".tmp", wxConvUTF8)) ){
			std::cout<<"Falied to create Dir: .tmp"<<std::endl;
		}
        }

}

void CKadInterface::ShutdownNetworkThread(){

        EventMessage msg;
	msg.SetStatus(EVT_SHUTDOWN);

        std::string sBuf;
        msg.serialize(sBuf);

        int iStatus = ::send(m_iSockPairFd[1], sBuf.c_str(), sBuf.size(), 0);
        if(iStatus<0){
                AddLogLineM(false, wxT("Failed sending shutdown command in sockpair"));
        }
}

int CKadInterface::ReadModifiedConfigFile(int iFd){


#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

int length, i = 0;
char buffer[EVENT_BUF_LEN];

	length = read( m_iFileWatchFd, buffer, EVENT_BUF_LEN );

	/*checking for error*/
	if ( length < 0 ) {
		std::cout<<"Error in read\n" ;
		return -1;
	}

	int iStatus = ParseConfig();
	if(iStatus != 0 )
		return -1;
	std::cout<<"after reread..max gossip="<<m_iMaxGossip<<std::endl;

#if 0
	/*actually read return the list of change events happens. Here, read the change event one by one and process it accordingly.*/
	while ( i < length ){
		struct inotify_event *event = ( struct inotify_event * ) &buffer[ i ];     
		if ( event->len ) {
			if ( event->mask & IN_MODIFY ) {
				std::cout<<"Got IN_MODIFY signal\n";
			}
		}
	i += EVENT_SIZE + event->len;
	}
#endif


return 0;
}

int CKadInterface::GetFileWatchFd(){

	m_iFileWatchFd = inotify_init();

	if ( m_iFileWatchFd < 0 ) {
		std::cout<<"inotify_init Error" <<std::endl;
		return -1;
	}

	m_iFileWatchWd= inotify_add_watch( m_iFileWatchFd, m_sConfigFileName.c_str(), IN_MODIFY );
	if(m_iFileWatchWd<0){
		close( m_iFileWatchFd );
		return -1;
	}

return m_iFileWatchFd;
}

bool CKadInterface::IncReceiverThreadCount(){

	//if(m_iReceiverThreadCount>MAX_RECEIVER_THREAD_ALLOWED){
	if(m_iReceiverThreadCount>m_iMaxReceiverThread){
		return false;
	}

	m_iReceiverThreadCount++;

return true;
}

void CKadInterface::ReformTopology(){

	if(m_bAlreadyReformed){
		std::cout<<"Already Reformed..returning"<<std::endl;
		return;
	}
	m_bAlreadyReformed = true;

        Kademlia::ContactList contacts;
        if( Kademlia::CKademlia::IsRunning() ) {
                Kademlia::CKademlia::GetAllContacts(contacts);
                Kademlia::ContactList::size_type numContacts = contacts.size();
                //numContacts = std::min<ContactList::size_type>(numContacts, CONTACT_FILE_LIMIT); // safety precaution, should not be above

                if (numContacts <= 0) {
                        AddLogLineM(false, wxString::Format(wxPLURAL("Only %d Kad contact available, nodes.dat not written", "Only %d Kad contacts available, nodes.dat not written", numContacts), numContacts));
                        return;
                }

		//first insert all contacts in a vector
		std::vector<Kademlia::CContact*> vContacts;

		  /* initialize random seed: */
		srand ( time(NULL) );

		//generate a number between 5 to 15;
		//int iDegree = rand()%3+2;
		//int iDegree = rand()%10+5;
		int iDegree = rand()%30+40;
		std::cout<<"Calculated Degree= "<<iDegree<<std::endl;


		for (Kademlia::ContactList::const_iterator it = contacts.begin(); it != contacts.end(); ++it) {
			Kademlia::CContact *c = *it;
			vContacts.push_back(c);
		}

		random_shuffle(vContacts.begin(), vContacts.end());

		//Kademlia::CKademlia::SetLockAddContact(true);

		for(int i=0;i<vContacts.size();i++){
			Kademlia::CContact *c = vContacts[i];

			//we will not delete frist iDegree contacts
			if(i<iDegree){
				m_vContacts.push_back(c);
				continue;
			}

		//	c->MarkAsDeleted();
		}

	}

}
