#include<iostream>
#include<signal.h>

#include "kademlia_interface.h"
#include "Logger.h"
#include "gossip.h"
#include "util.h"
//#include "SmartPtr.h"


void SetupSignalHandler();
void SignalHandler(int sig);

using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::string;


CGossip *kadObj = NULL;
bool bForeground = false;
bool bRun = true;
const char* log_file="comm_log.txt";
const char* pid_file="pid.txt";
const char version[]="1.2.8";

/**
* Main function of the Communication Component.
* Creates an instance of CGossip class.
* The process can be run in foreground or in background as a daemon.
* If -f argument is passed in the executable, the executable will run in foreground, else it will run in background.
* If run in background, a pid file is created.
* If run in background, all output will be written in log_file variable file name.
*/

int main(int argc, char* argv[]){


	//foreground?
	if(argv[1]){	
		if(strcmp(argv[1],"-v")==0){
			cout<<"Version: "<<version<<endl;;
			return 0;
		}
		else if(strcmp(argv[1],"-f")==0){
			cout<<"Running foreground\n";
			bForeground = true;
		}
		else{
			cout<<"Usage: ./p2p [-f] [-v]"<<endl;;
			return 0;
		}
	}

	kadObj = CGossip::getGossipInstance("config.txt");
	if(kadObj==NULL){
		cout<<"failed to create gossip object"<<endl;
		return -1;
	}

	/*Setup Signal Handler*/
	SetupSignalHandler();
        if( !bForeground )
        {
		cout<<"Creating Child\n";
		unlink(log_file);
                int ret = RunAsDaemon(log_file);
                if (ret < 0)
                {
                        cerr << "Could not background.  Exiting." << cerr;
                        exit(-2);
                }
                else if( ret == 0 ) //parent exits..
                {
			return 0;
                }

                // child continues
		if( write_pid(pid_file) == -1 ){
			cout<<"Can't write pid file\n";
		}

        }


	int iStatus = kadObj->Run();
	if(iStatus<0){
		cout<<"Unable to Run Kad"<<endl;
		bRun = false;
	}
	while(bRun){
		if( bForeground ){
			cout<<"Options:"<<endl;
			cout<<"1. Send Bootstrap Kad msg"<<endl;
			cout<<"2. GetAllContacts()"<<endl;
			cout<<"3. Find IP by Node ID"<<endl;
			cout<<"4. Send Command"<<endl;
			cout<<"5. Exit"<<endl;
			cout<<"Waiting for command:"<<endl;
		
			int iCommand;
			//cin>>iCommand;
			scanf("%d",&iCommand);

			switch(iCommand){

			case 1:
				kadObj->BootstrapKad();
				break;
			case 2:
				//kadObj->GetAllContacts();
				break;
			case 3:
				//kadObj->FindNodeByID(nodeID);
				break;
			case 4:
				//kadObj->SendCommand(sCommand);
				break;
			case 5:
				//kadObj->SendCommand(sCommand);
				delete kadObj;
				exit(0);
				break;
			default:
				cout<<"Invalide Command"<<endl;
				break;
			}
		}
		else{
			sleep(1);
		}
	}
	AddDebugLogLineM(false, logClientUDP, wxT("Shutting down communicator ") );
	delete kadObj;
	kadObj = NULL;
	unlink(pid_file);


return 0;
}


/**
* Signal handler setup.
* The moduel can handle signals as specified in this handler.
*/

void SetupSignalHandler(){

        // get rid of sigpipe
        signal(SIGPIPE, SIG_IGN);

        (void) signal(SIGINT,SignalHandler);
        (void) signal(SIGTERM,SignalHandler);
        (void) signal(SIGSEGV,SignalHandler);
        (void) signal(SIGILL,SignalHandler);
        (void) signal(SIGABRT,SignalHandler);
}

void RotateLog(){
	char sCmd[500]={0};
	sprintf(sCmd,"cat %s >> %s.1",log_file, log_file);
	system(sCmd);
}

/**
* If the module catches a signal, this handler is called.
* It makes a graceful shutdown of the module.
*/


void SignalHandler(int sig){

	AddDebugLogLineM(false, logClientUDP, wxT("In Signal Handler ") );
	std::cout<<"signal got: "<<sig<<std::endl;
	RotateLog();
	delete kadObj;
	unlink(pid_file);
	bRun = false;
        exit(sig);
}


