#ifndef UTIL_H
#define UTIL_H

#include<vector>

/** \file util.h
* Utility functions.
*/


/** \var typedef AddressPortPair
* A STD pair for ip address and port.
*/
typedef std::pair<uint32_t, uint16_t> AddressPortPair;

/**
* Returns the size of the sFile in bytes.
*/
int GetFileSize(const std::string& sFile);

/**
* Returns the dot formatted IP address from 4 byte ip.
*/
std::string GetDotFormatIP(uint32_t ip);

/**
* Makes socket fd reuseable.
*/
void reuse(int fd);

/**
* Makes socket fd non-blocking.
*/
void nonblock(int fd);

/**
* Disables naggles algorithm.
*/
void nodelay(int fd);
void ByteOrderTesting();

/**
* Runs current process as daemon.
* write all output to stdoutname file.
*/
int RunAsDaemon(const char* stdoutname);

/**
* Creates a pid file with the pid number of the daemon.
*/
int write_pid(const char * sFileName);

/**
* Retrieves IP and port of s socket sock.
*/
AddressPortPair GetSockName(uint32_t sock);

/**
* A special string tokenizer. Used for finding gossip message ID from a list.
*/
int StringTokenizer(const std::string sToParse, const char* cToken, std::vector<std::string>& v);


AddressPortPair GetPeerName(uint32_t sock);

std::string GetGMTime(void);


std::string itoa(int i);


#endif//UTIL_H
