#include<stdio.h>
#include<assert.h>
#include<pthread.h>
#include<sys/queue.h>
#include <time.h> 

#include <pgm/pgm.h>
#include "pgm_interface.h"


#define MAXSIZE		100
#define PAYLOAD		0x0A
#define PAYLOAD_SIZE	1024


static int              max_tpdu = 1500;
static int              max_rte = 400*1000;             /* very conservative rate, 2.5mb/s */
static int              sqns = 100;
static int              rs_k = 8;
static int              rs_n = 255;
//static bool             use_multicast_loop = TRUE;
static bool             use_multicast_loop = FALSE;
static bool             is_terminated = FALSE;

static         pthread_t pTID;
static 	filetransfer_callback g_cb;

static pgm_transport_t* send_transport = NULL;
static pgm_transport_t* recv_transport = NULL;
static int              terminate_pipe[2];


struct _packet{
	uint8_t m_8Type;
	char m_szTransactionID[16]; //we use this id 16 bytes fixed; it should be a boost uuid
        int m_iFileNameLen;
        char m_szFileName[MAXSIZE];
        int m_iFileSize;
        unsigned int m_uSeqNo;
        int m_iPayloadLen;
        char m_szPayload[PAYLOAD_SIZE];
};
typedef struct _packet packet;


struct _file_list{
	char m_szTransactionID[16];
	unsigned int m_iReceived;
	LIST_ENTRY(_file_list) entries;
};
typedef struct _file_list file_list;
LIST_HEAD(, _file_list) flist_head;


struct _packet_list{
	packet m_packet;
	LIST_ENTRY(_packet_list)entries;
};
typedef struct _packet_list packet_list;
LIST_HEAD(,_packet_list) plist_head;

//forward declaration
static int on_data (void*restrict, size_t, pgm_tsi_t*restrict);
static void close_receiver();
static void SetTransactionID(char* s, int len);
static void save_packet(const packet* p);
static void write_file(const packet* pkt);
static void fill_file_info(file_info* finfo, const packet* pkt);

bool init_pgm(){

        pgm_error_t* pgm_err = NULL;
        //setlocale (LC_ALL, "");

        if (!pgm_init (&pgm_err)) {
                fprintf (stderr, "Unable to start PGM engine: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
		return false;
        }

	LIST_INIT(&flist_head);
	LIST_INIT(&plist_head);

return true;
}


void close_pgm(){
	printf("close gpm\n");

	close_receiver();

        if (send_transport) {
                pgm_transport_destroy (send_transport, TRUE);
                send_transport = NULL;
        }
        if (recv_transport) {
                pgm_transport_destroy (recv_transport, TRUE);
                recv_transport = NULL;
        }

        pgm_shutdown();

	//delete before exit
	while (flist_head.lh_first != NULL){
		LIST_REMOVE(flist_head.lh_first, entries); 
	}
	while (plist_head.lh_first != NULL){
		LIST_REMOVE(plist_head.lh_first, entries); 
	}
}


bool create_send_transport ( const char* network,int udp_encap_port, int port, bool use_fec)
{
        struct pgm_transport_info_t* res = NULL;
        pgm_error_t* pgm_err = NULL;

	/* parse network parameter into transport address structure */
        char network_param[1024];
        sprintf (network_param, "%s", network);
        if (!pgm_if_get_transport_info (network, NULL, &res, &pgm_err)) {
                fprintf (stderr, "Parsing network parameter: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                return FALSE;
        }
	/* create global session identifier */
        if (!pgm_gsi_create_from_hostname (&res->ti_gsi, &pgm_err)) {
                fprintf (stderr, "Creating GSI: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                pgm_if_free_transport_info (res);
                return FALSE;
        }
        if (udp_encap_port) {
                res->ti_udp_encap_ucast_port = udp_encap_port;
                res->ti_udp_encap_mcast_port = udp_encap_port;
        }
        if (port)
                res->ti_dport = port;

/* Use RFC 2113 tagging for PGM Router Assist */
//      res->ti_ip_router_alert = 1;
        if (!pgm_transport_create (&send_transport, res, &pgm_err)) {
                fprintf (stderr, "Creating transport: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                pgm_if_free_transport_info (res);
                return FALSE;
        }
        pgm_if_free_transport_info (res);

	/* set PGM parameters */
        pgm_transport_set_send_only (send_transport, TRUE);
        pgm_transport_set_max_tpdu (send_transport, max_tpdu);
        pgm_transport_set_txw_sqns (send_transport, sqns);
        pgm_transport_set_txw_max_rte (send_transport, max_rte);
        pgm_transport_set_multicast_loop (send_transport, use_multicast_loop);
        pgm_transport_set_hops (send_transport, 16);
        pgm_transport_set_ambient_spm (send_transport, pgm_secs(30));
        unsigned spm_heartbeat[] = { pgm_msecs(100), pgm_msecs(100), pgm_msecs(100), pgm_msecs(100), pgm_msecs(1300), pgm_secs(7
), pgm_secs(16), pgm_secs(25), pgm_secs(30) };
        pgm_transport_set_heartbeat_spm (send_transport, spm_heartbeat, sizeof(spm_heartbeat)/sizeof(spm_heartbeat[0]));
        if (use_fec) {
                pgm_transport_set_fec (send_transport, 0, TRUE, TRUE, rs_n, rs_k);
        }

	/* assign transport to specified address */
        if (!pgm_transport_bind (send_transport, &pgm_err)) {
                fprintf (stderr, "Binding transport: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                pgm_transport_destroy (send_transport, FALSE);
                send_transport = NULL;
                return FALSE;
        }
        return TRUE;
}

bool create_recv_transport ( const char* network,int udp_encap_port, int port, bool use_fec){

        struct pgm_transport_info_t* res = NULL;
        pgm_error_t* pgm_err = NULL;

        puts ("Create recv transport.");

	/* parse network parameter into transport address structure */
        char network_param[1024];
        sprintf (network_param, "%s", network);
        if (!pgm_if_get_transport_info (network_param, NULL, &res, &pgm_err)) {
                fprintf (stderr, "Parsing network parameter: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                return FALSE;
        }
	/* create global session identifier */
        if (!pgm_gsi_create_from_hostname (&res->ti_gsi, &pgm_err)) {
                fprintf (stderr, "Creating GSI: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                pgm_if_free_transport_info (res);
                return FALSE;
        }
        if (udp_encap_port) {
                res->ti_udp_encap_ucast_port = udp_encap_port;
                res->ti_udp_encap_mcast_port = udp_encap_port;
        }
        if (port)
                res->ti_dport = port;
        if (!pgm_transport_create (&recv_transport, res, &pgm_err)) {
                fprintf (stderr, "Creating recv transport: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                pgm_if_free_transport_info (res);
                return FALSE;
        }
        pgm_if_free_transport_info (res);

	/* set PGM parameters */
        pgm_transport_set_nonblocking (recv_transport, TRUE);
        pgm_transport_set_recv_only (recv_transport, TRUE, FALSE);
        pgm_transport_set_max_tpdu (recv_transport, max_tpdu);
        pgm_transport_set_rxw_sqns (recv_transport, sqns);
        pgm_transport_set_multicast_loop (recv_transport, use_multicast_loop);
        pgm_transport_set_hops (recv_transport, 16);
        pgm_transport_set_peer_expiry (recv_transport, pgm_secs(300));
        pgm_transport_set_spmr_expiry (recv_transport, pgm_msecs(250));
        pgm_transport_set_nak_bo_ivl (recv_transport, pgm_msecs(50));
        pgm_transport_set_nak_rpt_ivl (recv_transport, pgm_secs(2));
        pgm_transport_set_nak_rdata_ivl (recv_transport, pgm_secs(2));
        pgm_transport_set_nak_data_retries (recv_transport, 50);
        pgm_transport_set_nak_ncf_retries (recv_transport, 50);
        if (use_fec) {
                pgm_transport_set_fec (recv_transport, 0, TRUE, TRUE, rs_n, rs_k);
        }

	/* assign transport to specified address */
        if (!pgm_transport_bind (recv_transport, &pgm_err)) {
                fprintf (stderr, "Binding recv transport: %s\n", pgm_err->message);
                pgm_error_free (pgm_err);
                pgm_transport_destroy (recv_transport, FALSE);
                recv_transport = NULL;
                return FALSE;
        }
        puts ("Create recv transport complete.");

return TRUE;
}
int send_packet(const char* szData, int iLen){

	if(!send_transport){
		fprintf (stderr, "pgm_send() failed: transport is not initialized.\n");
		return -1;
	}
	
	int bytes_written;
	const int status = pgm_send (send_transport, szData, iLen, &bytes_written);
	if (PGM_IO_STATUS_NORMAL != status) {
		fprintf (stderr, "pgm_send() failed.\n");
		return -1;
	}

return bytes_written;
}


void* receiver_loop(void* arg){

        int fds;
        fd_set readfds;
        puts ("Entering PGM message loop ... ");
        pgm_error_t* pgm_err = NULL;



       do {
                struct timeval tv;
                char buffer[4096];
                size_t len;
                pgm_tsi_t from;
                const int status = pgm_recvfrom (recv_transport,
                                                 buffer,
                                                 sizeof(buffer),
                                                 0,
                                                 &len,
                                                 &from,
                                                 &pgm_err);
                switch (status) {
                case PGM_IO_STATUS_NORMAL:
                        on_data (buffer, len, &from);
                        break;
                case PGM_IO_STATUS_TIMER_PENDING:
                        pgm_transport_get_timer_pending (recv_transport, &tv);
                        goto block;
                case PGM_IO_STATUS_RATE_LIMITED:
                        pgm_transport_get_rate_remaining (recv_transport, &tv);
                case PGM_IO_STATUS_WOULD_BLOCK:
/* select for next event */
block:
                        fds = terminate_pipe[0] + 1;
                        FD_ZERO(&readfds);
                        FD_SET(terminate_pipe[0], &readfds);
                        pgm_transport_select_info (recv_transport, &readfds, NULL, &fds);
                        fds = select (fds, &readfds, NULL, NULL, PGM_IO_STATUS_WOULD_BLOCK == status ? NULL : &tv);
                        break;

                default:
                        if (pgm_err) {
                                fprintf (stderr, "%s\n", pgm_err->message);
                                pgm_error_free (pgm_err);
                                pgm_err = NULL;
                        }
                        if (PGM_IO_STATUS_ERROR == status)
                                break;
                }
        } while (!is_terminated);

        puts ("Message loop terminated, cleaning up.");

	/* cleanup */
        close (terminate_pipe[0]);
        close (terminate_pipe[1]);

return NULL;
}


static void close_receiver()
{
	printf("closeing receiver thread\n");
        is_terminated = TRUE;
        const char one = '1';
        const size_t writelen = write (terminate_pipe[1], &one, sizeof(one));
        //assert (sizeof(one) == writelen);

	//thread join
	int iRet = pthread_join(pTID, NULL);
	if(iRet!=0){
		perror("pthread_join returned error\n");
			
	}
	printf("returnig closeing receiver thread\n");
}

static int on_data (
        void*      restrict data,
        size_t              len,
        pgm_tsi_t* restrict from
        )
{
/* protect against non-null terminated strings */
        char buf[1500], tsi[PGM_TSISTRLEN];

	const uint8_t* pData = (uint8_t*)data;
	uint8_t pType = pData[0];

	if(pType == PAYLOAD){
		//printf("payload arrived\n");
		packet p;
		memcpy((void*)&p,data,len);

		p.m_iFileNameLen = ntohl(p.m_iFileNameLen);
		p.m_iFileSize = ntohl(p.m_iFileSize);
		p.m_uSeqNo = ntohl(p.m_uSeqNo);
		p.m_iPayloadLen = ntohl(p.m_iPayloadLen);

#if 0
		printf("seq=%u\n",p.m_uSeqNo);
		printf("TID=%s\n",p.m_szTransactionID);
		printf("Filename=%s\n",p.m_szFileName);
		printf("FileSize=%d\n",p.m_iFileSize);
		printf("payload len=%d\n",p.m_iPayloadLen);
		printf("payload=%s\n",p.m_szPayload);
#endif

		save_packet(&p);
	}
	else{
		snprintf (buf, sizeof(buf), "%s", (char*)data);
	        pgm_tsi_print_r (from, tsi, sizeof(tsi));
		//printf ("\"%s\" (%zu bytes from %s)\n", buf, len, tsi);
	}

        return 0;
}

int start_receiver_thread(){


	printf("statrting the thread\n");
        int e = pipe (terminate_pipe);

        int iRet = pthread_create(&pTID,NULL,receiver_loop, NULL);
        if (iRet){
                 perror("error in creating receiver thread:");
                 return -1;
        }


return 0;
}

//sender thread will autometically close down after down..
void* sender_thread(void* arg){
	file_info* finfo = (file_info*)arg;

        FILE* pFile = fopen(finfo->m_szFileLoc, "rb");
        if( pFile == NULL){
                fprintf (stderr, "Unable to open file for read: %s\n",finfo->m_szFileLoc);
		pthread_exit(0);
                return NULL;
        }

	//construct the packet

	unsigned int iSeq = 0;
	while(1){
		packet p;
		bzero(&p,sizeof(packet));

		p.m_8Type = PAYLOAD;
		memcpy(p.m_szTransactionID, finfo->m_szTransactionID, 16);
		p.m_iFileNameLen = htonl(finfo->m_iFileNameLen);

		memcpy(p.m_szFileName,finfo->m_szFileName,finfo->m_iFileNameLen);

		p.m_iFileSize = htonl(finfo->m_iFileSize);

		p.m_uSeqNo = htonl(iSeq++);

                int iReadSize=fread(p.m_szPayload,1,PAYLOAD_SIZE,pFile);
                if(iReadSize<=0){
                        break;
                }
		p.m_iPayloadLen = htonl(iReadSize);

		int iRet = send_packet((char*)&p, sizeof(packet));
		if(iRet<0){
			return -1;
		}

	}
	fclose(pFile);
        pthread_exit(0);
}
int send_file(const char* szFileName, int iFileLen, const char* szFileLoc, int iFileLocLen, int iFileSize){

	//check the file existance
	if(iFileSize==0){
                fprintf (stderr, "File Size is 0\n");
		return -1;
	}
	
	pthread_t pSenderThread;
        pthread_attr_t threadAttr;
        if( pthread_attr_init(&threadAttr) != 0 )
                fprintf (stderr, "pthread attr initit error\n");

        if( pthread_attr_setdetachstate(&threadAttr, PTHREAD_CREATE_DETACHED) != 0 )
                fprintf (stderr, "Error in pthread_attr_setdetachstate\n");


	static file_info finfo;
	bzero(&finfo,sizeof(file_info));

	finfo.m_iFileNameLen = iFileLen;
	memcpy(finfo.m_szFileName, szFileName, iFileLen);

	finfo.m_iFileLocLen = iFileLocLen;
	memcpy(finfo.m_szFileLoc, szFileLoc, iFileLocLen);

	finfo.m_iFileSize = iFileSize;
	SetTransactionID(finfo.m_szTransactionID,16);

        int iRet = pthread_create(&pSenderThread,&threadAttr,sender_thread, (void*)&finfo);
        if (iRet){
                perror("error in creating sender multicast thread:");
                return -1;
        }

       if( pthread_attr_destroy(&threadAttr) != 0 )
                fprintf (stderr, "Error in pthread_attr_destroy\n");

return 0;
}

/*This should be a boost uuid*/
static void SetTransactionID(char* s, int len){

	bzero(s,len);
	unsigned int t;
	time(&t);
	sprintf(s,"%u",t);
	//printf("generated t=%u\n",t);
}

static void save_packet(const packet* pkt){

	//insert the packet in the plist;
	packet_list* pl = (packet_list*)malloc(sizeof(packet_list));
	pl->m_packet = *pkt;
	//memcpy(&pl->m_packet, pkt, sizeof(packet));
	LIST_INSERT_HEAD(&plist_head, pl , entries); 

	//printf("seq inserted=%d\n",pl->m_packet.m_uSeqNo);

	//one packet is the whole packet
	if(pkt->m_iFileSize == pkt->m_iPayloadLen){
		//save the file and fire the eventmessage
		write_file(pkt);
		return;
	}
	//traverse the flist
	bool bFound = false;
	file_list* p = NULL;
	for( p= flist_head.lh_first; p!=NULL; p=p->entries.le_next){
		if(strcmp( p->m_szTransactionID , pkt->m_szTransactionID) == 0 ){
			bFound = true;
			//found the entry;
			p->m_iReceived += pkt->m_iPayloadLen;

			if(p->m_iReceived == pkt->m_iFileSize ){
				//file transfer complelet, save file to the disk and notify the policy
				//delete the entry
				LIST_REMOVE(p,entries);
				write_file(pkt);
				//notify the file transfer..
				file_info* finfo = (file_info*)malloc(sizeof(file_info));
				fill_file_info(finfo,pkt);
				g_cb(finfo);

			}
			break;
		}
	}
	if(!bFound){
		//first time, pkt arrived, save in the list
		file_list* l = (file_list*)malloc(sizeof(file_list));
		memcpy(l->m_szTransactionID, pkt->m_szTransactionID, 16);
		l->m_iReceived = pkt->m_iPayloadLen;
		LIST_INSERT_HEAD(&flist_head, l , entries); 
	}

}

static void write_file(const packet* pkt){

        FILE* pFile = fopen(pkt->m_szFileName, "wb");
        if( pFile == NULL){
                fprintf (stderr, "Error in writing file\n");
                return;
        }

	//traverse packet_list
	int iSeq = 0;
	int iReceived = 0;

	while(1){
		bool bFound = false;

		packet_list* p = NULL;
		for( p= plist_head.lh_first; p!=NULL; p=p->entries.le_next){
			if(p->m_packet.m_uSeqNo == iSeq ){
				bFound = true;
				iSeq++;
				break;
			}
		}

		if(bFound){
			fwrite(p->m_packet.m_szPayload,1,p->m_packet.m_iPayloadLen,pFile);
			iReceived+=p->m_packet.m_iPayloadLen;
			//delete it from the list
			LIST_REMOVE(p,entries);

			if(iReceived == p->m_packet.m_iFileSize){
				//complete
				fclose(pFile);
				//fireevent
				fprintf (stderr, "File written successfully\n");
				break;
			}
		}
		else{
			fprintf (stderr, "Seq no not found in packet_list: %d\n",iSeq);
			break;
		}

	}

}

void set_filetransfer_callback(filetransfer_callback cb){

	g_cb = cb;
}

static void fill_file_info(file_info* finfo, const packet* pkt){

        finfo->m_iFileNameLen = pkt->m_iFileNameLen;

        bzero(finfo->m_szFileName, MAXSIZE);
        memcpy(finfo->m_szFileName, pkt->m_szFileName, pkt->m_iFileNameLen);

        finfo->m_iFileSize = pkt->m_iFileSize;
}
