#ifndef _PGM_INTERFACE_H
#define _PGM_INTERFACE_H

/** \file pgm_interface.h
* This is an interface between OpenPGM library and CGossip class.    
*/

/**\def MAXSIZE
* Maximum size of a file name in bytes.
*/
#define MAXSIZE         100



/**\struct
* Multicast file transfer packet structure.
*/
struct fileinfo{
        int m_iFileNameLen;
        char m_szFileName[MAXSIZE];
        int m_iFileLocLen;
        char m_szFileLoc[MAXSIZE];
        int m_iFileSize;
        char m_szTransactionID[16];
};
typedef struct fileinfo file_info;



/**
* Initialize PGM component in OpenPGM library.
*/
bool init_pgm();

/**
* Closes PGM component in OpenPGM library.
*/
void close_pgm();

/**
* Creates sender transport. 
* We use different transport for send and receive.
*/
bool create_send_transport(const char* network,int udp_encap_port, int port, bool use_fec);

/**
* Creates receive transport. 
*/
bool create_recv_transport(const char* network,int udp_encap_port, int port, bool use_fec);

/**
* Sends a multicast packet. Only for testing.
*/
int send_packet(const char* szData, int iLen);

/**
* Sends a file. 
*/
int send_file(const char* szFileName, int iFileLen, const char* szFileLoc, int iFileLocLen, int iFileSize);


/**
* Starts receiver thread.
*/
int start_receiver_thread();

/**
* Stops PGM.
*/
void close_pgm();

/**
* Callback function to notify file tranfer status when it is finished. 
* It calls C++ function for using EventMessage to notify kademlia_interface network thread.
*/
typedef void (* filetransfer_callback)(file_info*) ;

/**
* Set file transfer status callback function.
*/
void set_filetransfer_callback(filetransfer_callback cb);




#endif // _PGM_INTERFACE_H

