
#include <boost/uuid/uuid_io.hpp>
#include <arpa/inet.h>
#include<vector>
#include<fstream>
#include<sstream>



#include "gossip.h"
#include "kademlia/routing/Maps.h"
#include "kademlia/routing/Contact.h"
#include "protocol/Protocols.h"
#include "EventMessage.h"
#include "FileTransfer.h"
#include "util.h"
#include "BootStrapClient.h"

#include "kademlia/net/ValidatePublish.h"



//transaction loggin window length 1 hour
#define MAX_LOG_DIFF	3600 
//we will not log more than this limit
#define MAX_LOG_LIMIT	25

bool CGossip::bInstance = false;
CGossip* CGossip::instance = NULL;

CGossip* CGossip::getGossipInstance(std::string sFile){

	if(!bInstance){
		bInstance = true;
		instance = new CGossip(sFile);
		return instance;
	}
	else{
		return instance;
	}
}

CGossip::CGossip(std::string sFileName):
	CKadInterface(sFileName),
	m_pBootStrapClient(0){

        m_pBootStrapClient = new CBootStrapClient();

/*should we delete the tmp files in the startup? TODO*/
}

CGossip::~CGossip(){
	bInstance = false;

        if ( !m_gfiletrans.empty() ) {

		int i=0;
                for ( i = 0; i < m_gfiletrans.size(); i++ ) {
			delete m_gfiletrans[i];
                }

        }
 
        delete m_pBootStrapClient;
        m_pBootStrapClient = NULL;

	printf("Exiting gossip destructor\n");
}


/*receives command from the receiver thread; from the command terminal*/
int CGossip::SendCommand(const std::string& sFileName, bool bTcp, uint8_t module){

	//Multicast
	if(module == MODULE_GOSSIP_AND_MULTICAST || module == MODULE_MULTICAST){
		if(GetMulticastInstance()){
			std::cout<<"Sending Multicast file\n";

			wxString wxStr(sFileName.c_str(), wxConvUTF8);
			std::string sName  = std::string( wxFileNameFromPath(wxStr).mb_str() );

			GetMulticastInstance()->SendFile(sName, sFileName);
		}				
	}

	if(!IsGossipEnable()){
		std::cout<<"Gossip Module is not enabled\n";
		return 0;
	}


	if(module != MODULE_GOSSIP_AND_MULTICAST &&  module != MODULE_GOSSIP){
		std::cout<<"Command doesn't want Gossipping \n";
		return 0;
	}

	//gossip
	if(bTcp){
		boost::uuids::uuid Tag;
		int iRet = GetTransactionID(sFileName, Tag);
		if(iRet==0){
			SendTcpCommand(sFileName, Tag, 0, 0);

			//now logging to the TransactionLog: in sender
			DoTransactionLog(Tag, sFileName);
			
		}
	}
	else{
		int iSize = 0;
		GossipPacket gPack;
		gPack.m_iProtocol=GOSSIP_PROTO;
		gPack.m_Tag = boost::uuids::random_generator()();

		//test
		FILE* fp = fopen(sFileName.c_str(),"r");
		if(!fp)
			return -1;

		char szBuffer[4096]={0};
		while (1){
			bzero(szBuffer, 4096);

			int iReadSize=fread(szBuffer,1,4096,fp);
			iSize = iReadSize;
			if(iReadSize<=0){
				break;
			}
		break;
		}
		//printf("iSize=%d\n",iSize);
		memcpy((void*)&gPack.m_cPack,(void*)szBuffer,iSize);
		int len = sizeof(gPack.m_iProtocol) + boost::uuids::uuid::static_size() + iSize;
		SendCommand(gPack, len, 0, 0);
	}



return 0;
}

/*This is a TCP Gossip version; similar to SendCommand, but in TCP, we don't huse Gossip Header anymore*/
int CGossip::SendTcpCommand(const std::string& sFileToSend, const boost::uuids::uuid& Tag, uint32_t srcIP, uint16_t srcPort){

        Kademlia::ContactList contacts;
        if( Kademlia::CKademlia::IsRunning() ) {
                Kademlia::CKademlia::GetAllContacts(contacts);
                Kademlia::ContactList::size_type numContacts = contacts.size();

                if (numContacts <= 0) {
                        AddLogLineM(false, wxString::Format(wxPLURAL("Only %d Kad contact available, nodes.dat not written", "Only %d Kad contacts available, nodes.dat not written", numContacts), numContacts));
                }
                else{
                        for (Kademlia::ContactList::const_iterator it = contacts.begin(); it != contacts.end(); ++it) {

                                Kademlia::CContact *c = *it;

                                if(c->GetIPAddress() == srcIP && c->GetUDPPort() == srcPort ){
                                        std::cout<<"src ip and port matched\n";
                                        continue;
                                }
                                char buf[14];
                                sprintf(buf,"%x:%x",c->GetIPAddress(),c->GetUDPPort());
                                std::string sAddr(buf,13);
                                //search
                                MapFwHist::const_iterator iter = mFwHist.find(Tag);
                                if(iter == mFwHist.end()){
                                        std::cout<<"Sending for first time\n";
                                        std::vector<std::string> vAddr;
                                        vAddr.push_back(sAddr);
                                        //insert
                                        std::pair<MapFwHist::iterator, bool> ret = mFwHist.insert( std::pair<boost::uuids::uuid, std::vector<std::string> >(Tag, vAddr) );
                                        if(ret.second == false){
                                                std::cout<<"Unable to insert entry in mFwHist\n";
                                        }
                                       SendTcpGossipPacket(sFileToSend, c->GetIPAddress(), c->GetTCPPort());
                                }
                                else{

                                        //now..search in vector
                                        const std::vector<std::string>& vAddr = iter->second;
                                        if(MAX_GOSSIP_SEND == vAddr.size()){
                                                std::cout<<"Already gossiped max\n";
                                                break;
                                        }
                                        if(vAddr.end() == find(vAddr.begin(), vAddr.end(), sAddr) ){

					        SendTcpGossipPacket(sFileToSend, c->GetIPAddress(), c->GetTCPPort());

                                                std::vector<std::string> vUpdateAddr(vAddr);
                                                vUpdateAddr.push_back(sAddr);
                                                mFwHist[Tag] = vUpdateAddr;
                                        }
                                        else{
                                                std::cout<<"Already Sent \n";
                                                continue;
                                        }
                                }

//                                break;

                        }
                }

	}

return 0;
}


/*Sazzad: We send the command to MAX_GOSSIP clients at most: we check the dest is not matched with the src packet
We also check packet is sent to a dest once*/
int CGossip::SendCommand(const GossipPacket& gPack, int len, uint32_t srcIP, uint16_t srcPort){

        Kademlia::ContactList contacts;
        if( Kademlia::CKademlia::IsRunning() ) {
                Kademlia::CKademlia::GetAllContacts(contacts);
                Kademlia::ContactList::size_type numContacts = contacts.size();

                if (numContacts <= 0) {
                        AddLogLineM(false, wxString::Format(wxPLURAL("Only %d Kad contact available, nodes.dat not written", "Only %d Kad contacts available, nodes.dat not written", numContacts), numContacts));
                }
                else{
                        for (Kademlia::ContactList::const_iterator it = contacts.begin(); it != contacts.end(); ++it) {

                                Kademlia::CContact *c = *it;

                                if(c->GetIPAddress() == srcIP && c->GetUDPPort() == srcPort ){
					std::cout<<"src ip and port matched\n";
                                        continue;
				}

                                char buf[14];
                                sprintf(buf,"%x:%x",c->GetIPAddress(),c->GetUDPPort());
                                std::string sAddr(buf,13);
                                //search
                                MapFwHist::const_iterator iter = mFwHist.find(gPack.m_Tag);
                                if(iter == mFwHist.end()){
                                        std::cout<<"Sending for first time\n";
					std::vector<std::string> vAddr;
					vAddr.push_back(sAddr);
                                        //insert
                                        std::pair<MapFwHist::iterator, bool> ret = mFwHist.insert( std::pair<boost::uuids::uuid, std::vector<std::string> >(gPack.m_Tag, vAddr) );
					if(ret.second == false){
						std::cout<<"Unable to insert entry in mFwHist\n";
					}
                                        SendGossipPacket(gPack, len, c->GetIPAddress(), c->GetUDPPort());
                                }
                                else{

					//now..search in vector
					const std::vector<std::string>& vAddr = iter->second;
					if(MAX_GOSSIP_SEND == vAddr.size()){
						std::cout<<"Already gossiped max\n";
						break;
					}
					if(vAddr.end() == find(vAddr.begin(), vAddr.end(), sAddr) ){

						SendGossipPacket(gPack, len, c->GetIPAddress(), c->GetUDPPort());

						std::vector<std::string> vUpdateAddr(vAddr);
						vUpdateAddr.push_back(sAddr);
						mFwHist[gPack.m_Tag] = vUpdateAddr;
					}
					else{
						std::cout<<"Already Sent \n";
						continue;
					}
				}
				
//                                break;

                        }
                }

        }

return 0;
}

int CGossip::SendTcpGossipPacket(const std::string& sFileName, uint32_t destinationHost, uint16_t destinationPort){

        int iSockFd = CFileTransfer::ConnectPeer(destinationHost, destinationPort);
        if(iSockFd==-1){
                std::cout<<"Failed to create tcp service socket"<<std::endl;
                return -1;
        }
        std::cout<<"Connected Successfully to: "<<GetDotFormatIP(destinationHost)<<":"<<destinationPort<<std::endl;
        std::cout<<"my file transfer sock="<<iSockFd<<std::endl;


	int iFileSize = GetFileSize(sFileName);

        CFileTransfer* filetrans = new CFileTransfer(iSockFd,GetSharedDir(), GetDownloadDir());
        filetrans->HandShake(false, true, sFileName, iFileSize);
        m_gfiletrans.push_back(filetrans);

return 0;
}

int CGossip::SendGossipPacket(const GossipPacket& packet, int pktLen, uint32_t destinationHost, uint16_t destinationPort){

        struct sockaddr_in destAddr;
        socklen_t addrLen = sizeof(destAddr);
        bzero(&destAddr, sizeof(destAddr));
        destAddr.sin_family = AF_INET;
        destAddr.sin_addr.s_addr = htonl(destinationHost);
        destAddr.sin_port = htons(destinationPort);

        int iSendSize = sendto(GetSockFd(), (void*)&packet, pktLen, 0, (sockaddr*)&destAddr, addrLen );
        if( iSendSize == -1 ){
                std::cout<<"Gossip Send failed"<<std::endl;
        }
        std::cout << "packet sent : " << iSendSize << " bytes" <<"to: "<<GetDotFormatIP(destinationHost)<<":"<<destinationPort<<std::endl;;

return iSendSize;
}

void CGossip::ProcessTcpGossipPacket(const std::string& sFileName, uint32 ip, uint16 port, uint8_t module){

	boost::uuids::uuid Tag;
	int iRet = GetTransactionID(sFileName, Tag);
        std::cout<<"Transaction ID=" << Tag;

	char buf[14];
	sprintf(buf,"%x:%x",ip,port);
	std::string sAddr(buf,13);

        MapFwHist::const_iterator iter = mRecvHist.find(Tag);
        if(iter == mRecvHist.end()){
                std::cout<<"Arrived for the first time: "<<std::endl;
#if 0
                /*Notify Upper layer*/
		NotifyApp(sFileName, FILE_GOSSIP);
#endif

		std::vector<std::string> vAddr;
		vAddr.push_back(sAddr);
		//insert
		std::pair<MapFwHist::iterator, bool> ret = mRecvHist.insert( std::pair<boost::uuids::uuid, std::vector<std::string> >(Tag, vAddr) );
		if(ret.second == false){
			std::cout<<"Unable to insert entry in mRecvHist\n";
		}
        }
	//since, we only gossip after varification from policy component, we should always send the received packet to the policy component.
	//policy component should decide when whould it execute the command or not.
#if 1
	/*Notify Upper layer*/
	NotifyApp(sFileName, FILE_GOSSIP, module);
#endif


	//since we received it from ip, port, we will never sent it to that addr. So, insert the ip, port in the mFwHist so that the addr will never be picked up
	iter = mFwHist.find(Tag);
	if(iter == mFwHist.end()){
		std::vector<std::string> vAddr;
		vAddr.push_back(sAddr);
		std::pair<MapFwHist::iterator, bool> ret = mFwHist.insert( std::pair<boost::uuids::uuid, std::vector<std::string> >(Tag, vAddr) );
		if(ret.second == false){
			std::cout<<"Unable to insert entry in mFwHist\n";
		}
	}
	else{
		std::vector<std::string> vAddr;
		vAddr.push_back(sAddr);
	}

	//now logging to the TransactionLog: in receiver
	DoTransactionLog(Tag, sFileName);

	//for now, stop forwarding..
        //SendTcpCommand(sFileName, Tag, ip, port);
}

void CGossip::ProcessGossipPacket(const GossipPacket& gPack, uint32 ip, uint16 port, size_t length){

#if 0
        const char* s = (char*)gPack.m_cPack.m_szCommand;
        printf("command=%s\n",s);
#endif
        printf("ip=%x\n",ip);
        printf("port=%d\n",port);
        std::cout<<"Gossip ID=" << gPack.m_Tag <<std::endl;

	MapFwHist::const_iterator iter = mRecvHist.find(gPack.m_Tag);
	if(iter == mRecvHist.end()){
		std::cout<<"Arrived for the first time: "<<std::endl;
		//should fire the event to the user
		/*Notify Upper layer*/

        	int iSize = length - (sizeof(gPack.m_iProtocol) + boost::uuids::uuid::static_size() ); 
{
//test
static int iFileNo = 1;
char szFileName[100]={0};
sprintf(szFileName,"cmd__r_%d",iFileNo);

std::string sFileName(szFileName, strlen(szFileName));
std::ofstream myfile(sFileName.c_str());
if (!myfile.is_open()){
        std::cout<<"Failed to open command file\n";
        return;

}
std::string s((char*)gPack.m_cPack, iSize);
myfile<<s;
myfile.close();



#if 0
		CommandPacket cPack;
		memcpy((void*)&cPack, gPack.m_cPack, iSize);
		NotifyCommand(cPack, iSize);
#else
        EventMessage msg;
	msg.SetReasonCode(FILE_RECV_REQ);
	msg.SetFileType(FILE_GOSSIP);
	std::string sPath= MakeAbsPath(sFileName);
        msg.SetFileName(sPath);
        //msg.SetFilePath(m_sFileName);

        std::string sBuf;
        msg.serialize(sBuf);
	NotifyCommand(sBuf);

#endif
}
		//NotifyCommand(gPack.m_cPack, iSize);


		char buf[14];
		sprintf(buf,"%x:%x",ip,port);
		std::string sAddr(buf,13);

		std::vector<std::string> vAddr;
		vAddr.push_back(sAddr);
		//insert
		std::pair<MapFwHist::iterator, bool> ret = mRecvHist.insert( std::pair<boost::uuids::uuid, std::vector<std::string> >(gPack.m_Tag, vAddr) );
		if(ret.second == false){
			std::cout<<"Unable to insert entry in mRecvHist\n";
		}
	}


        SendCommand(gPack, length, ip, port);
}


//int CGossip::NotifyCommand(const CommandPacket& cPack, size_t len){
int CGossip::NotifyCommand(const std::string& sBuf){

	int iRet=-1;
	int iFd = GetPeerSockFd();
	if(iFd>0){
		iRet = ::send(iFd, (void*)sBuf.c_str(), sBuf.size(), 0);
		if(iRet < 0)
			std::cout<<"Send command failed\n";
	}
return iRet;
}

void CGossip::OnPacketReceived(uint32 ip, uint16 port, byte* buffer, size_t length)
{
        wxCHECK_RET(length >= 2, wxT("Invalid packet."));

        uint8_t *decryptedBuffer=buffer;;
        uint32_t receiverVerifyKey;
        uint32_t senderVerifyKey;
        int packetLen = length;
        uint8_t protocol = decryptedBuffer[0];
        uint8_t opcode   = decryptedBuffer[1];
        //printf("In OnPacketReceived in Gossip: opcode=%02x\n",opcode);
        GossipPacket gPack;


        if (packetLen >= 1) {

                 try {
                        switch (protocol) {
                                case OP_EMULEPROT:
                                        break;

                                /*Kademlia Packet*/
                                case OP_KADEMLIAHEADER:
                                        //std::cout<<"Kademlia header arrived!"<<std::endl;
                                        if (packetLen >= 2) {
                                                Kademlia::CKademlia::ProcessPacket(decryptedBuffer, packetLen, ip, port, (Kademlia::CPrefs::GetUDPVerifyKey(ip) == receiverVerifyKey), Kademlia::CKadUDPKey(senderVerifyKey, GetPublicIP(false)));
                                        } else {
                                                throw wxString(wxT("Kad packet too short"));
                                        }

                                        break;

                                case OP_KADEMLIAPACKEDPROT:
                                                std::cout<<"Should not occur OP_KADEMLIAPACKEDPROT"<<std::endl;
                                        break;

                                /*Gossip Packet*/
                                case GOSSIP_PROTO:
                                        std::cout<<"Gossip header arrived!: len="<<length<<std::endl;
                                        memcpy(&gPack,decryptedBuffer, length);
                                        ProcessGossipPacket(gPack, ip, port, length);
                                        break;

                                default:
                                        AddDebugLogLineM(false, logClientUDP, wxString::Format(wxT("Unknown opcode on received packet: 0x%x"), protocol));
                        }
                } catch (const wxString& e) {
                        AddDebugLogLineM(false, logClientUDP, wxT("Error while parsing UDP packet: ") + e);
                } catch (const CInvalidPacket& e) {
                        AddDebugLogLineM(false, logClientUDP, wxT("Invalid UDP packet encountered: ") + e.what());
                } catch (const CEOFException& e) {
                        AddDebugLogLineM(false, logClientUDP, wxT("Malformed packet encountered while parsing UDP packet: ") + e.what());
                }

        }

}


int CGossip::OnCommandReceived(int iFd){

char szBuf[MAXCHAR];

        int iReceived = recv(iFd, (void*)szBuf, MAXCHAR, 0 );
        if(0== iReceived ){
                return 0;
        }
	std::cout<<"TCP data arrived: "<<iReceived<<std::endl;

	int iRead = 0;
	while(iRead<iReceived){

		const char* pBuf = szBuf;
		pBuf+=iRead;

		EventMessage msg;
		std::string sBuf(pBuf,iReceived-iRead);
		iRead += msg.deserialize(sBuf);

		std::cout<<"iRead="<<iRead<<std::endl;


		switch(msg.GetReasonCode()){

			case FILE_GOSSIP_REQ:
			{
				const std::string& sFileName = msg.GetFileName();
				int iTransport = msg.GetStatus();
				switch(iTransport){
					case USE_TCP:
							SendCommand(sFileName, true, msg.GetRelatedModule());
							break;
					case USE_UDP:
							SendCommand(sFileName, false, msg.GetRelatedModule());
							break;
					default:
							std::cout<<"Unknown transport type:"<<iTransport<<std::endl;
							break;
				}
				break;
			}
		
			case FILE_PUBLISH_REQ:
			{
				const std::string& sFileName = msg.GetFileName();
				std::cout<<"Publish:File name: "<<sFileName<<std::endl;
				PublishFile(sFileName);
				break;
			}

			case FILE_FIND_REQ:
			{
				const std::string& sFileName = msg.GetFileName();
				std::cout<<"FindFIle: File name: "<<sFileName<<std::endl;
				SearchKeyKad(sFileName);

				break;
			}
			case FILE_BOOTSTRAP_SRV_REQ:
			{
				//we encode ip and port in these two fields;
				std::cout<<"Bootstrap Srv: ip="<<msg.GetFileSize()<<" port="<<msg.GetSearchID()<<std::endl;
				int iRet = GetKadBootstrapInfo(msg.GetFileSize(), msg.GetSearchID());
				if(iRet==0 && m_pBootStrapClient){
					std::string sPath = GetDownloadDir()+"/"+m_pBootStrapClient->GetEventMessage()->GetFileName();
					NotifyApp(sPath,
						 FILE_BOOTSTRAP,
						 MODULE_UNKNOWN);
				}
				break;
			}
			case FILE_BOOTSTRAP_KAD_REQ:
			{
				//we encode ip and port in these two fields;
				std::cout<<"Bootstrap KAD: ip="<<msg.GetFileSize()<<" port="<<msg.GetSearchID()<<std::endl;
				BootstrapKad(msg.GetFileSize(), msg.GetSearchID());
				break;
			}

			case FILE_REPORT_CERT_REQ:
			{
				std::cout<<"File report cert req: File name: "<<msg.GetFileName()<<std::endl;
				Kademlia::CKademlia::GetValidatePublish()->CheckValidation(msg.GetFileName());
				break;
			}

			default:
				std::cout<<"Unknown packet received from gossipcomm"<<std::endl;
				iReceived = -1; //to make the loop break;
				break;
		}

	}//while

return iReceived;
}

//about file transfer complete
void CGossip::NotifyApp(const std::string& sFileName, uint8_t filetype, uint8_t relatedModule){

	if(sFileName.empty()){
		AddLogLineM(false, wxT("From file trafer sock pair: File Name is Empty "));
		return;
	}

	AddLogLineM(false, wxT("From file trafer sock pair: received: ")+wxString(sFileName.c_str(), wxConvUTF8));

        std::string sPath = sFileName;

	std::string sFileToNotify = MakeAbsPath( CopyFileToTmpDir(sFileName) ); 

        EventMessage msg;
        msg.SetReasonCode(FILE_RECV_REQ);
        msg.SetFileType(filetype);
        msg.SetFileName(sFileToNotify);
        msg.SetRelatedModule(relatedModule);
        //msg.SetFileName(sPath);
        //msg.SetFilePath(m_sFileName);

        std::string sBuf;
        msg.serialize(sBuf);
        NotifyCommand(sBuf);

}

std::string CGossip::CopyFileToTmpDir(const std::string & sFile){

        wxString src = wxString(sFile.c_str(), wxConvUTF8);

        size_t pos = sFile.find_last_of('/');

        std::string sFileName(sFile);

        if(pos != std::string::npos){
                sFileName = sFile.substr(pos+1);
        }

        std::string sDest = std::string(".tmp")+"/"+sFileName;
        //std::cout<<"dest file: "<<sDest<<std::endl;

        wxString dest = wxString(sDest.c_str(), wxConvUTF8);

        if(!wxCopyFile(src, dest)){
                std::cout<<"copy failed: "<<sFile<<" To: "<<sDest<<std::endl;
        }

return sDest;
}


int CGossip::GetTransactionID(const std::string& sFile, boost::uuids::uuid& id){

	std::string sOut;
        std::ifstream ifFile(sFile.c_str());
        if (!ifFile.is_open()){
                std::cout<<"Unable to open file\n";
                return -1;
        }

        while (!ifFile.eof() ){
                std::string sLine;
                getline (ifFile,sLine);
                int iPos = sLine.find("Subject: ");
                if(iPos != std::string::npos){
                        sOut = sLine.substr(iPos+9);
                        break;
                }
        }
        ifFile.close();

	std::stringstream ss;
        ss<<sOut;
        ss >> id;

return 0;
}


int CGossip::GetKadBootstrapInfo(int iServerIP, int iPort){

        if(!m_pBootStrapClient)
                return -1;

        int iRet  = m_pBootStrapClient->Download( iServerIP, iPort, GetDownloadDir());
        if(iRet == 0){
                AddLogLineM(false, wxT("Download Successfully Completed"));
                //std::cout<<"file name="<<m_pBootStrapClient->GetEventMessage()->GetFileName()<<std::endl;
                //notify policy_component
        }
        else{
                AddLogLineM(false, wxT("Download was not Successfull"));
        }

return iRet;
}

int CGossip::DoTransactionLog(const  boost::uuids::uuid&Tag, const std::string& sFileName){

	time_t tCurrent = 0;
	time(&tCurrent);

	//first insert it to the queue
	std::string sSavedFile = BackupFileToDownloadDir(sFileName, Tag);

	if(!sSavedFile.empty()){


		//new elements in the back, old elements in the front
		//check duplicate
		bool bDuplicate(false);
		std::deque<TransactionLog>::const_iterator qIter;
		for(qIter = m_qTransactionLog.begin(); qIter != m_qTransactionLog.end(); qIter++){
			const TransactionLog& l = *qIter;
			if(Tag == l.Tag){
				bDuplicate = true;
			}
		}
	
		if(!bDuplicate ){

			TransactionLog log(tCurrent,Tag, sSavedFile);
			std::cout<<"inserting new transaction: "<<Tag<<std::endl;
			m_qTransactionLog.push_back(log);
			if(m_qTransactionLog.size() >= MAX_LOG_LIMIT){
				m_qTransactionLog.pop_front();
			}
		}
	}

	//now, remove outdated elements
	while(!m_qTransactionLog.empty()){
		const TransactionLog& l = m_qTransactionLog.front();
		if((tCurrent-l.t)>MAX_LOG_DIFF){
			std::cout<<"removing old transaction: "<<l.t<<std::endl;
			m_qTransactionLog.pop_front();
		}
		else{
			break;
		}
	}

return 0;
}
int CGossip::GetTransactionLog(std::string& transList)const{
	time_t tCurrent = 0;
	time(&tCurrent);
	transList.clear();

	std::deque<TransactionLog>::const_reverse_iterator rIter;

	for(rIter = m_qTransactionLog.rbegin(); rIter < m_qTransactionLog.rend(); rIter++){

		const TransactionLog& l = *rIter;
		if((tCurrent-l.t)<=MAX_LOG_DIFF){
			std::stringstream ss;
			ss<<l.Tag;
			std::string sTransaction;
			ss>>sTransaction;

			transList+=sTransaction;
			transList+=";";
		}
		else{
			break;
		}
	}

return 0;
}

int CGossip::FindMissingGossip(const std::string& sTransList, std::string& sMissingList)const{

	sMissingList.clear();

	std::vector<std::string> vList;
	StringTokenizer(sTransList, ";", vList);

	for(int i=0;i<vList.size();i++){

		const std::string& sTrans = vList[i];
		if(sTrans.empty())
			continue;
		std::cout<<sTrans<<std::endl;

		boost::uuids::uuid Tag;
		std::stringstream ss;
		ss<<sTrans;
		ss >> Tag;

		MapFwHist::const_iterator iter = mFwHist.find(Tag);
		if(iter == mFwHist.end()){
			sMissingList+=sTrans;
			sMissingList+=";";
		}
	}
	std::cout<<"From PING: my missing list: "<<sMissingList<<std::endl;

return 0;
}

int CGossip::SendMissingGossip(const std::string& sMissingList, uint32_t ip, uint16_t port){


        std::vector<std::string> vList;
        StringTokenizer(sMissingList, ";", vList);

        for(int i=0;i<vList.size();i++){

                const std::string& sTrans = vList[i];
                if(sTrans.empty())
                        continue;
                std::cout<<"in PONG: missing list: "<<sTrans<<std::endl;

                boost::uuids::uuid Tag;
                std::stringstream ss;
                ss<<sTrans;
                ss >> Tag;

		//now, send this file to that neighbor
		SendMissingGossip(Tag, ip, port);
	}
return 0;
}

int CGossip::SendMissingGossip(const boost::uuids::uuid& Tag, uint32_t ip, uint16_t port){

        std::deque<TransactionLog>::const_iterator qIter;
        for(qIter = m_qTransactionLog.begin(); qIter != m_qTransactionLog.end(); qIter++){
                const TransactionLog& l = *qIter;
                if(Tag == l.Tag){
			const std::string& sFileToSend = l.sFileName;

			if(GetFileSize(sFileToSend) <= 0){
				std::cout<<"Error in File size ="<<sFileToSend<<std::endl;
				continue;
			}
			wxString wxStr(sFileToSend.c_str(), wxConvUTF8);
			if(!wxFileExists(wxStr)){
				std::cout<<"File doesn't exist="<<sFileToSend<<std::endl;
				continue;
			}


			SendTcpGossipPacket(sFileToSend, ip, port);
			std::cout<<"Gossip file sent from hash retrive to :"<<GetDotFormatIP(ip)<<" :port="<<port<<std::endl;
                }
        }


}

std::string CGossip::BackupFileToDownloadDir(const std::string & sFile, const  boost::uuids::uuid& Tag){

	//sanity check
	boost::uuids::uuid Tag2;
	int iRet = GetTransactionID(sFile, Tag2);
	if(iRet==0){
		if(Tag!=Tag2){
			std::cout<<"Tag doesn't match for file: "<<sFile<<std::endl;
			return std::string();
		}
	}
	else{
		return std::string();
	}

	std::stringstream ss;
	ss<<Tag;
	std::string sTransaction;
	ss>>sTransaction;
	std::string sDest = GetDownloadDir()+"/"+sTransaction;


        wxString src = wxString(sFile.c_str(), wxConvUTF8);
        wxString dest = wxString(sDest.c_str(), wxConvUTF8);

        if(!wxCopyFile(src, dest)){
                std::cout<<"copy failed: "<<sFile<<std::endl;
		return std::string();
        }

return sDest;
}

