#ifndef _VALIDATE_INDEX_H
#define _VALIDATE_INDEX_H

#include <queue>

#include "../utils/UInt128.h"
#include "../kademlia/Entry.h"

class CFileTransfer;
typedef std::pair<uint32_t, uint16_t> AddrPortPair;

namespace Kademlia{

struct StorePublishSrcInfo{
	CUInt128 target;
	CEntry*  entry;
	bool 	bValidated;

	StorePublishSrcInfo(CUInt128 t, CEntry* e):
		target(t),
		entry(e),
		bValidated(false){
	}
};
typedef StorePublishSrcInfo StoreSrcInfo;

struct StorePublishKeyInfo{
        CUInt128 file;
        CKeyEntry*  entry;
        bool    bValidated;
        std::string sFileName;

        StorePublishKeyInfo(CUInt128 t, CKeyEntry* e, const std::string& sName ):
                file(t),
                entry(e),
                bValidated(false),
                sFileName(sName){
        }
};
typedef StorePublishKeyInfo StoreKeyInfo;

class CValidatePublish{

public:
	CValidatePublish();
	virtual ~CValidatePublish();

	void StartValidatePublishSrc(const CUInt128& file, const CUInt128& target,  CEntry* entry);
	void StartValidatePublishKey(const CUInt128& file, const CUInt128& target,  CKeyEntry* entry, const std::vector<std::string>& vFileName);
	void CheckValidation(const std::string sFile);

private:
	void AcceptPublishRequest(const CUInt128& file);
	int StartFileTransfer(uint32_t ip, uint16_t tcp, const std::string& file, int iFileSize, const CUInt128&);
	std::string GetFileName(const CUInt128& target, int iFileSize);
	bool IsSelfReceived(uint32_t remote_ip, uint16_t remote_port);
	bool IsAlreadyValidated(const CUInt128& file);
	void DoPendingTransfer(const CUInt128& file, const std::string& sFileName, int iFileSize);







private:

	std::map<CUInt128,StoreSrcInfo>			m_PublishSrcMap;
	std::map<CUInt128,std::vector<StoreKeyInfo> >	m_PublishKeyMap;

        typedef std::deque<CFileTransfer*> FileTransfer;
        FileTransfer m_filetrans;
	std::map<CUInt128, AddrPortPair>		m_AddrPortPair;


};

}//namespace kademlia

#endif// _VALIDATE_INDEX_H

