#include <vector>

#include "ValidatePublish.h"
#include "../kademlia/Indexed.h"
#include "../kademlia/Kademlia.h"
#include "../../Gossip/FileTransfer.h"
#include "../../Gossip/util.h"

#include "../../Gossip/gossip.h"
#include "MD4Hash.h"


////////////////////////////////////////
using namespace Kademlia;
////////////////////////////////////////



CValidatePublish::CValidatePublish(){
}

CValidatePublish::~CValidatePublish(){

        if ( !m_filetrans.empty() ) {
                for ( unsigned int i = 0; i < m_filetrans.size(); i++ ) {
                          delete m_filetrans[i];
                }
        }


}

void CValidatePublish::StartValidatePublishSrc(const CUInt128& file, const CUInt128& target,  Kademlia::CEntry* entry){
//file is the file hash
	
	StoreSrcInfo info(target, entry);
	std::pair<std::map<CUInt128,StoreSrcInfo>::iterator, bool> bRet  = m_PublishSrcMap.insert(std::pair<CUInt128, StoreSrcInfo>(file, info));
	if(bRet.second==false){
		std::cout<<"Inserrting in the PublishSrcMap fails\n"<<std::endl;
	}

	//get the file name, file size, dest ip, dest port
	int iFileSize = entry->m_uSize;
	std::string sFileName = GetFileName(file, iFileSize);
	uint16_t tcpPort = entry->m_uTCPport ;
	uint32_t ip = entry->m_uIP;
	std::cout<<"iFileSize="<<iFileSize<<std::endl;
	std::cout<<"iFileName="<<sFileName<<std::endl;
	std::cout<<"TcpPort="<<tcpPort<<std::endl;
	std::cout<<"ip="<<ip<<std::endl;
	std::cout<<"My ip="<<CKademlia::GetPrefs()->GetIPAddress()<<std::endl;
	std::cout<<"My tcpPort="<<thePrefs::GetPort()<<std::endl;

	if(IsSelfReceived(ip, tcpPort)){
		AcceptPublishRequest(file);
	}
	else{
		//Start File Transfer
		StartFileTransfer(ip, tcpPort, sFileName, iFileSize, file);
	}

}


void CValidatePublish::StartValidatePublishKey(const CUInt128& file, const CUInt128& target,  Kademlia::CKeyEntry* entry, const std::vector<std::string>& vFileName){
//target is the file hash

	if(IsAlreadyValidated(target)){

		CIndexed *indexed = CKademlia::GetIndexed();

		uint8_t load = 0;
		if (!indexed->AddKeyword(file, target, entry, load)) {
			std::cout<<"Addkeyword failed"<<std::endl;
			delete entry;
			entry = NULL;
		}
		return ;
	}

	std::string sFileName;
	std::cout<<"File name vector size = "<<vFileName.size()<<std::endl;
	if(!vFileName.empty()){
		sFileName = vFileName[0];
	}
	int iFileSize = entry->m_uSize;

	StoreKeyInfo info(file, entry, sFileName);
	std::map<CUInt128,std::vector<StoreKeyInfo> >::iterator mIter;

	mIter = m_PublishKeyMap.find(target);

	if(mIter == m_PublishKeyMap.end()){
		
		std::vector<StoreKeyInfo> vInfo;
		vInfo.push_back(info);	

		m_PublishKeyMap.insert(std::pair<CUInt128, std::vector<StoreKeyInfo> >(target, vInfo));
	}
	else{
		std::vector<StoreKeyInfo>& vInfo = mIter->second;
		vInfo.push_back(info);	
		m_PublishKeyMap.insert(std::pair<CUInt128, std::vector<StoreKeyInfo> >(target, vInfo));
	}

	//Start any pending transfer..
	DoPendingTransfer(target, sFileName, iFileSize);
}

std::string CValidatePublish::GetFileName(const CUInt128& target, int iFileSize){

	std::map<CUInt128,std::vector<StoreKeyInfo> >::iterator mIter;
	mIter = m_PublishKeyMap.find(target);
	if(mIter == m_PublishKeyMap.end()){
		return std::string();
	}
	const std::vector<StoreKeyInfo>& vInfo = mIter->second;

	//file name and file size in all the StoreInfo should be same, NO?
	const StoreKeyInfo& info = vInfo[0];
	if(info.entry){
		if(info.entry->m_uSize != iFileSize){
			std::cout<<"File Size doesn't match!!!"<<std::endl;
			return std::string();
		}
		return info.sFileName;
	}

return std::string();
}

void CValidatePublish::AcceptPublishRequest(const CUInt128& file){


	CIndexed *indexed = CKademlia::GetIndexed();

	//accept all keywords publish
	std::map<CUInt128,std::vector<StoreKeyInfo> >::iterator mIter;

	mIter = m_PublishKeyMap.find(file);

	if(mIter != m_PublishKeyMap.end()){

		std::vector<StoreKeyInfo>& vInfo =  mIter->second;
		std::vector<StoreKeyInfo>::iterator vIter;
		for(vIter = vInfo.begin(); vIter!= vInfo.end(); vIter++){
			StoreKeyInfo& info = *vIter;

			if(info.entry == NULL)
				continue;

			uint8_t load = 0;
			std::cout<<"Validated publish key: "<<file.ToHexString()<<std::endl;
			if (!indexed->AddKeyword(info.file, file, info.entry, load)) {
				std::cout<<"Addkeyword failed"<<std::endl;
					//We already indexed the maximum number of keywords.
					//We do not index anymore but we still send a success..
					//Reason: Because if a VERY busy node tells the publisher it failed,
					//this busy node will spread to all the surrounding nodes causing popular
					//keywords to be stored on MANY nodes..
					//So, once we are full, we will periodically clean our list until we can
					//begin storing again..
						delete info.entry;
						info.entry = NULL;
			}

			//removing from the vector
			//vInfo.erase(vIter);
		}

		//remove from the map
		//m_PublishKeyMap.erase(mIter);
	}

	//accept all src publish
	std::map<CUInt128,StoreSrcInfo>::iterator mSrcIter;
	mSrcIter = m_PublishSrcMap.find(file);

	if(mSrcIter != m_PublishSrcMap.end()){
		StoreSrcInfo& info = mSrcIter->second;

		uint8_t load=0;
		if (info.entry && indexed->AddSources(file, info.target, info.entry, load)) {
			//flag = true;
			std::cout<<"Validated publish src: "<<file.ToHexString()<<std::endl;
			info.bValidated = true;
	
		} else {
			std::cout<<"AddSources failed"<<std::endl;
			delete info.entry;
			info.entry = NULL;
		}
	}
	//m_PublishSrcMap.erase(mSrcIter);
	std::cout<<"Validation Completed! "<<std::endl;

}

/*
*/
int CValidatePublish::StartFileTransfer(uint32_t ip, uint16_t tcp, const std::string& sFileName, int iFileSize, const CUInt128& hash){

	if(sFileName.empty()){
		AddrPortPair p(ip,tcp);
		m_AddrPortPair.insert(std::pair<CUInt128,AddrPortPair>(hash, p));
		return -1;
	}

        int iSockFd = CFileTransfer::ConnectPeer(ip, tcp);
        if(iSockFd==-1){
                std::cout<<"Failed to create tcp service socket"<<std::endl;
                return -1;
        }
        std::cout<<"Connected Successfully to: "<<GetDotFormatIP(ip)<<":"<<tcp<<std::endl;
        std::cout<<"my file transfer port="<<iSockFd<<std::endl;

        CFileTransfer* filetrans = new CFileTransfer(iSockFd, CGossip::getGossipInstance("")->GetSharedDir(), CGossip::getGossipInstance("")->GetDownloadDir());
        filetrans->SetFileHash(CMD4Hash(hash));
        filetrans->SetSocketPair(CGossip::getGossipInstance("")->GetSocketPair()[1]);
	filetrans->HandShake(false, false, sFileName, iFileSize);
	m_filetrans.push_back(filetrans);

return 0;

}
bool CValidatePublish::IsSelfReceived(uint32_t remote_ip, uint16_t remote_port){

	uint32_t local_ip = CKademlia::GetPrefs()->GetIPAddress();
	uint16_t local_port = thePrefs::GetPort();

	if(local_ip == remote_ip && local_port == remote_port){
		std::cout<<"Self received"<<std::endl;
		return true;
	}

return false;
}

void CValidatePublish::CheckValidation(const std::string sFilePath){

	if(CKademlia::GetUDPListener()->m_bValidateIndex == false)
		return;

	//get only the name part of the file, not the path
        wxString wxStr(sFilePath.c_str(), wxConvUTF8);
        std::string sName  = std::string( wxFileNameFromPath(wxStr).mb_str() );
	std::cout<<"File Name="<<sName<<std::endl;

	FileTransfer::iterator qIter;
	CUInt128 file;
	for(qIter=m_filetrans.begin(); qIter!=m_filetrans.end(); qIter++){

		CFileTransfer* f = *qIter;
		if(sName == f->GetFileName()){
			std::cout<<"File Name Found in the map ="<<sName<<std::endl;
			const unsigned char* hash = f->GetFileHash().GetHash();
			file = CUInt128(static_cast<const uint8_t*>(hash));
			AcceptPublishRequest(file);
			break;
		}
	}

}

void CValidatePublish::DoPendingTransfer(const CUInt128& file, const std::string& sFileName, int iFileSize){

	std::map<CUInt128, AddrPortPair>::iterator mIter;
	mIter = m_AddrPortPair.find(file);
	if(mIter!=m_AddrPortPair.end()){
		AddrPortPair& p = mIter->second;
		std::cout<<"File Hash Found in the map for DoPendingTransfer = "<<sFileName<<std::endl;
		StartFileTransfer(p.first, p.second, sFileName, iFileSize, mIter->first);
	}
}

bool CValidatePublish::IsAlreadyValidated(const CUInt128& file){

	std::map<CUInt128,StoreSrcInfo>::iterator mSrcIter;
	mSrcIter = m_PublishSrcMap.find(file);

	if(mSrcIter != m_PublishSrcMap.end()){
		StoreSrcInfo& info = mSrcIter->second;
		if(info.bValidated){
			std::cout<<"Publisher is already validated!!\n";
			return true;
		}
	}

return false;
}
