#include<iostream>
#include<vector>
#include<string>
#include<fstream>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <error.h>
#include <netinet/tcp.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>

using namespace std;

#include "log.h"
#include "debug.h"
#include "LogServer.h"
#include "util.h"

#define MAXCHAR 1500
#define SOCKET_ERROR -1
#define POLL_TIMEOUT (4*1000)


CLogServer::CLogServer(const std::string& sConfigFile):

	m_sConfigFile(sConfigFile),
	m_iDaemonSockFd(SOCKET_ERROR),
	m_uServerIP(0),
        m_uServerPort(0),
	m_sLogFile(),
	m_bRun(true){
}

CLogServer::~CLogServer(){

	m_bRun = false;
	Stop();
	DestroySocket();
}

int CLogServer::Start(){

	if(-1 == ParseConfig()){
		return -1;
	}

	LOG_OPEN(m_sLogFile);

	if(-1 == CreateDaemonSocket()){
		return -1;
	}

	Run();

return 0;
}

int CLogServer::Stop()const{

	DB_AND_LOG("Closing Server..");
	//StopPollLoop();	
	LOG_CLOSE();
return 0;
}

int CLogServer::Run(){


	struct pollfd pollFd[1];

	pollFd[0].fd = m_iDaemonSockFd;
	pollFd[0].events = POLLIN;

	while (m_bRun){
		int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
		if( iStatus < 0){
			cout<<"poll failed in main loop"<<iStatus<<endl;;
		}
		else if( 0 == iStatus){
			//DB("Timeout in poll");
		}
		else if(pollFd[0].revents & POLLIN){
			HandleConnection(pollFd[0].fd);
		} 
		else{
			DB_AND_LOG("error in poll");
			break;
		}
	    
	}//while loop

return 0;
}


int CLogServer::HandleConnection(int iFd){

	struct sockaddr_in FromAddr;
	socklen_t FromAddrLen=sizeof(FromAddr);
	char szBuf[1500]={0};
	int iSent = -1;

	int iReceived = recvfrom(iFd, szBuf, 1500, 0, (struct sockaddr *) &FromAddr, &FromAddrLen);
	if(iReceived>0){
		uint32_t uIP = ntohl(FromAddr.sin_addr.s_addr);
		uint16_t uPort = ntohs(FromAddr.sin_port);
		std::string sLog(szBuf, iReceived);
		DB_AND_LOG("Log from Node: "<< GetDotFormatIP(uIP)<<":"<< uPort<<"#"<<sLog);

		iSent = sendto(iFd, "ACK", 3, 0, (struct sockaddr*)&FromAddr, FromAddrLen);
		if(iSent<0){
			DB_AND_LOG("Error in sendning reply to :"<<GetDotFormatIP(uIP)<<":"<< uPort);
		}


	}	
return iSent;
}

int CLogServer::CreateDaemonSocket(){
	struct sockaddr_in kAddr;
	int iRet;

	m_iDaemonSockFd = ::socket(AF_INET, SOCK_DGRAM, 0);
	if (m_iDaemonSockFd == -1){
		DB_AND_LOG("Could not create tcp socket.");
		return -1;
	}

	//reuse(m_iDaemonSockFd);

	bzero((char *) &kAddr, sizeof(kAddr));
	kAddr.sin_family = AF_INET;
	kAddr.sin_addr.s_addr = htonl(m_uServerIP); 
	kAddr.sin_port = htons(m_uServerPort);

printf("ip=%x\n", kAddr.sin_addr.s_addr);
printf("port=%d\n", kAddr.sin_port);

	iRet = bind(m_iDaemonSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr));
	if(iRet<0){
		DB_AND_LOG("ERROR on binding in tcp sock");
		return -1;
	}

return 0;
}


int CLogServer::ParseConfig(){

	std::vector<std::string> vConf;
	char szBuf[MAXCHAR];
	std::string sTemp;

	std::ifstream fileHandler(m_sConfigFile.c_str());
	if(!fileHandler.is_open()){
		std::cerr<<"Unable to open Input file"<<m_sConfigFile<<std::endl;
		return -1;
	}

	do{
		bzero(szBuf,MAXCHAR);
		fileHandler.getline(szBuf,MAXCHAR);
		vConf.push_back(szBuf);
	}while(!fileHandler.eof());
	fileHandler.close();


	//Now, parse config file
        sTemp = GetString(vConf, "server_ip") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_uServerIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

	sTemp = GetString(vConf, "server_port") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_uServerPort = atoi(sTemp.c_str());

	sTemp = GetString(vConf, "log_file") ;
	if(sTemp.empty())
		return -1;
	StrTrim(sTemp);
	m_sLogFile = sTemp;

return 0;
}

void CLogServer::StrTrim(std::string& str){

	std::string::size_type pos = str.find_last_not_of(' ');
	if(pos != std::string::npos) {
		str.erase(pos + 1);
		pos = str.find_first_not_of(' ');
		if(pos != std::string::npos)
			str.erase(0, pos);
		}
		else
			str.erase(str.begin(), str.end());
}

void CLogServer::DestroySocket()
{
	if(m_iDaemonSockFd != SOCKET_ERROR){
		::close(m_iDaemonSockFd);
		m_iDaemonSockFd = SOCKET_ERROR;
	}

}


std::string CLogServer::GetString(const std::vector<std::string>& vConf, const std::string& sLine)const{

	std::vector<std::string>::const_iterator vIter;
	std::string sTemp;
	int iPos=-1;

	for(vIter=vConf.begin(); vIter!=vConf.end(); vIter++){
		sTemp = *vIter;
		iPos = sTemp.find(sLine);
		if(iPos == std::string::npos)
			continue;

		iPos = sTemp.find("=");
		return sTemp.substr(iPos+1);
	}

return "";
}

