#ifndef LOG_SERVER_H
#define LOG_SERVER_H

#include<vector>




class CLogServer{


public:
	/**
	* Constructor of the class. 
	* It takes configuration file name as its argument.
	* An example of such configuration file is:
        * \verbinclude server_config.txt
	*/
	CLogServer(const std::string& sConfigFile);
	
	/**
	* Destructor of the class.
	*/
	virtual ~CLogServer();

	/**
	* This method is used to start the server.
	*/
	int Start();

	/**
	* This method is used to stop the server.
	*/
	int Stop()const;

private:

	/**
	* In this method, the server waits in a poll for any incomming connection from clients.
	*/
	int Run();

	/**
	* Once server receives a connection request and accepts it, this method is used to handle that connection.
	* It sends a SMIME signed file to the client listing the possible ip:port of nodes for bootstrapping.
	*/
	int HandleConnection(int iFd);

	int CreateDaemonSocket();

	int InitSocketPair();

	/**
	* Parses the configuration file. 
	* Returns -1 if an misconfiguration found in the configuration file.
	*/
	int ParseConfig();
	void StrTrim(std::string& str);
	void DestroySocket();
	std::string GetString(const std::vector<std::string> & vConf, const std::string& sLine) const;


private:
	std::string m_sConfigFile;
	int m_iDaemonSockFd;
	uint32_t m_uServerIP;
	uint16_t m_uServerPort;
	std::string m_sLogFile;

	bool m_bRun;

};

#endif//LOG_SERVER_H

