#include<iostream>
#include<vector>
#include<string>
#include<fstream>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <error.h>
#include <netinet/tcp.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>

using namespace std;

#define MAXCHAR 1500
#define SOCKET_ERROR -1
#define POLL_TIMEOUT (100)
#define MAX_SEND	5

void DestroySocket();
int Run();
int ReceiveAck(int);
int SendLog(int, const std::string&);
int CreateDaemonSocket();

uint32_t	m_uServerIP=0;
uint16_t	m_uServerPort=0;
int 	m_iDaemonSockFd;

int main(){

	if(-1 == CreateDaemonSocket()){
		return -1;
	}
	m_uServerIP=ntohl(inet_addr("127.0.0.1"));
	m_uServerPort=75690;

	Run();

	DestroySocket();
return 0;
}


int Run(){


	std::string sLog("From client 1");
	int iSent = SendLog(m_iDaemonSockFd, sLog);
	if(iSent<0)
		return -1;

	struct pollfd pollFd[1];

	pollFd[0].fd = m_iDaemonSockFd;
	pollFd[0].events = POLLIN;

	int iCount=0;
	while (iCount<MAX_SEND){
		int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
		if( iStatus < 0){
			cout<<"poll failed in main loop"<<iStatus<<endl;;
			return -1;
		}
		else if( 0 == iStatus){
			printf("Timeout in poll\n");
			iCount++;
			SendLog(m_iDaemonSockFd, sLog);
		}
		else if(pollFd[0].revents & POLLIN){
			ReceiveAck(pollFd[0].fd);
			break;
		} 
		else{
			printf("error in poll");
			break;
		}
	    
	}//while loop

return 0;
}


int SendLog(int iFd, const std::string& sLog){

	struct sockaddr_in FromAddr;
	socklen_t FromAddrLen=sizeof(FromAddr);

	bzero((char *) &FromAddr, sizeof(FromAddr));
	FromAddr.sin_family = AF_INET;
	FromAddr.sin_addr.s_addr = htonl(m_uServerIP); 
	FromAddr.sin_port = htons(m_uServerPort);

	int iSent = -1;
	iSent = sendto(iFd, sLog.c_str(), sLog.size(), 0, (struct sockaddr*)&FromAddr, FromAddrLen);
	if(iSent<0){
		std::cout<<"Error in sendning log to :"<<(m_uServerIP)<<":"<< m_uServerPort<<std::endl;;
	}	

return iSent;
}

int ReceiveAck(int iFd){

	struct sockaddr_in FromAddr;
	socklen_t FromAddrLen=sizeof(FromAddr);

	char szBuf[1500]={0};
	int iReceived = recvfrom(iFd, szBuf, 1500, 0, (struct sockaddr *) &FromAddr, &FromAddrLen);
	if(iReceived>0){
		uint32_t uIP = ntohl(FromAddr.sin_addr.s_addr);
		uint16_t uPort = ntohs(FromAddr.sin_port);
		std::string sLog(szBuf, iReceived);
		std::cout<<"Ack received : "<< (uIP)<<":"<< uPort<<"#"<<sLog<<std::endl;

	}
return 0;
}

int CreateDaemonSocket(){
	struct sockaddr_in FromAddr;
	int iRet;

	m_iDaemonSockFd = ::socket(AF_INET, SOCK_DGRAM, 0);
	if (m_iDaemonSockFd == -1){
		printf("Could not create tcp socket.");
		return -1;
	}

	//reuse(m_iDaemonSockFd);

	bzero((char *) &FromAddr, sizeof(FromAddr));
	FromAddr.sin_family = AF_INET;
	FromAddr.sin_addr.s_addr = htonl(INADDR_ANY); 
	FromAddr.sin_port = htons(0);

	iRet = bind(m_iDaemonSockFd, (struct sockaddr *) &FromAddr, sizeof(FromAddr));
	if(iRet<0){
		printf("ERROR on binding in tcp sock");
		return -1;
	}

return 0;
}


void DestroySocket()
{
	if(m_iDaemonSockFd != SOCKET_ERROR){
		::close(m_iDaemonSockFd);
		m_iDaemonSockFd = SOCKET_ERROR;
	}

}

