
#include<iostream>
#include<fstream>
#include<sstream>


#include<signal.h>
#include <poll.h>


#include "gossip_comm.h"
#include "util.h"

#define POLL_TIMEOUT 1*1000 //1 sec
#define LOG_TIMEOUT	60 //per minute

void SetupSignalHandler();
void SignalHandler(int sig);
bool IsCommandReady();
bool CreateReadyFile(const char* fName);

using std::cout;
using std::cin;
using std::cerr;
using std::endl;
using std::string;


CGossipComm *Obj = NULL;
bool bForeground = false;
bool bRun = true;
const char* log_file="policy_log.txt";
const char* pid_file="pid.txt";
const char* cmd_file="ready_file.txt";
const char* version = "1.1.5";


/**
* Main function of policy component module.
* It creates an object of CGossipComm class with the configuration file name as the argument.
* The argv argument takes -f option to run the module in the foreground. 
* If -f argument is not given, then the module will run in the background as a daemon mode.
* If the module is run as background, then a pid.txt file is generated having the pid (process ID) in the file.
* It also enables signal handler to catch signals and make a graceful shutdown.
*/

int main(int argc, char* argv[]){


        //foreground?
        if(argv[1]){

		if(strcmp(argv[1],"-v")==0){
                        cout<<"Version: "<<version<<endl;;
                        return 0;
                }
                else if(strcmp(argv[1],"-f")==0){
                        cout<<"Running foreground\n";
                        bForeground = true;
                }
		else{
                        cout<<"Usage: ./bin [-f or -v] "<<endl;;
                        return 0;
		}
        }

	Obj = new CGossipComm("policy_config.txt");
	CreateReadyFile(cmd_file);

	/*Setup Signal Handler*/
	SetupSignalHandler();
        if( !bForeground )
        {
                cout<<"Creating Child\n";
		unlink(log_file);
                int ret = RunAsDaemon(log_file);
                if (ret < 0)
                {
                        cerr << "Could not background.  Exiting." << cerr;
                        exit(-2);
                }
                else if( ret == 0 ) //parent exits..
                {
                        return 0;
                }

                // child continues
                if( write_pid(pid_file) == -1 ){
                        cout<<"Can't write pid file\n";
                }

        }

	int iStatus = Obj->Run();
	if(iStatus<0){
		cout<<"Unable to Run Terminal"<<endl;
		if(!bForeground){
			unlink(pid_file);
		}
	}
	else{

		int iTimeoutCount = 0;
		while(1){
			if( bForeground ){

				cout<<"Options:"<<endl;
				//cout<<"1. Show All Contacts"<<endl;
				cout<<"1. Send Command"<<endl;
				cout<<"2. Load Subscription"<<endl;
				cout<<"3. Send Log"<<endl;
				cout<<"Waiting for command:"<<endl;
			
				int iCommand;
				//cin>>iCommand;
				scanf("%d",&iCommand);

				switch(iCommand){

				case 2:
					//Obj->GetAllContacts();
					Obj->LoadSubscription();
					break;
				case 1:
					Obj->SendCommand();
					break;
				case 3:
					Obj->Logging("KEEP ALIVE");
					break;
				default:
					cout<<"Invalide Command"<<endl;
					break;
				}
			}
			else{
				//poll and then check for command file
				int iStatus = poll(NULL,0,POLL_TIMEOUT);
				if(iStatus == 0){
					bool bRet = IsCommandReady();
					if(bRet){
						Obj->SendCommand();
					}
				}
				if(iTimeoutCount == LOG_TIMEOUT){
					iTimeoutCount = 0;	
					//Obj->Logging("KEEP ALIVE");
				}

				iTimeoutCount++;	
			}
		}
	}

	delete Obj;

return 0;
}

/**
* If the module is run as a daemon, it will take command from a file called "ready_file.txt". 
* If the user wants to send a command, he has to put a 1 in this file.
* The daemon checks the file in each 1 second interval.
* If there is a 1 in that file, then the module will send the command as specified in configuration file.
* After sending command, the module puts a 0 in the file.
*/

bool IsCommandReady(){


	bool bRet = false;
        std::ifstream iFile(cmd_file);

        if(!iFile.is_open())
        {
                std::cout<<"Unable to open Input file"<<cmd_file<<std::endl;
                return false;
        }
	int iVal = -1;
	iFile>>iVal;
	if(iVal==1){
		bRet = true;
	}
	iFile.close();


	if(bRet){ //reset
			if(!CreateReadyFile(cmd_file))
				return false;
	}
		

return bRet;
}


/**
* Creates a ready_file.txt and puts a 0 initially.
* This file is required if the module is run as daemon.
*/
bool CreateReadyFile(const char* fName){

	std::ofstream oFile(fName);
	if(!oFile.is_open())
	{
		std::cout<<"Unable to open Output file"<<fName<<std::endl;
		return false;
	}
	int iVal=0;
	oFile<<iVal;
	oFile.close();

}

/**
* Signal handler setup.
* The moduel can handle signals as specified in this handler.
*/
void SetupSignalHandler(){

        (void) signal(SIGINT,SignalHandler);
        (void) signal(SIGTERM,SignalHandler);
//        (void) signal(SIGSEGV,SignalHandler);
        (void) signal(SIGILL,SignalHandler);
        (void) signal(SIGABRT,SignalHandler);
}

/**
* If the module catches a signal, this handler is called.
* It makes a graceful shutdown of the module.
*/

void SignalHandler(int sig){

	printf("in signal handler=%d\n",sig);
	delete Obj;
        unlink(pid_file);
        exit(sig);
}
