#include<iostream>
#include<fstream>
#include<vector>
#include<sstream>
#include<set>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <error.h>
#include <netinet/tcp.h>

#include<wx/string.h>
#include<wx/filefn.h>
#include <wx/filename.h>

#include <boost/uuid/uuid_io.hpp>

#include "gossip_comm.h"
#include "digiSig.h"
#include "serializable.h"
#include "debug.h"
#include "EventMessage.h"
#include "log.h"
#include "util.h"

#define POLL_TIMEOUT	100*1000 //sec
#define MAXCHAR	1500

#define OUTPUT_FILE	"output_command.sh"

#define MAX_CA_CHAIN	10

#define MAX_SEND        5 //for logserver comm
#define LOG_POLL_TIMEOUT 100
#define SOCKET_ERROR	-1



using std::cerr;
using std::endl;


CGossipComm::CGossipComm(std::string sFileName):
	m_sConfigFile(sFileName),
	m_bSignEnable(true),
	m_sLogFileName(),
	m_bInit(false),
	m_bBootstrapEnable(false),
	m_bValidateCert(true),
	m_iLogSockFd(SOCKET_ERROR),
	m_iLogServerIP(0),
	m_iLogServerPort(0),
	m_bLoggerInit(false),
	m_iBootIndex(0),
	m_bFloodingEnable(false),
	m_bNonCooperationEnable(false),
	m_sCricLevel("-1"){
}

CGossipComm::~CGossipComm(){

	if(!m_bInit)
		return;

	::close(m_iSockFd);

	StopLogger();


	ShutdownNetworkThread();
        int iRet = pthread_join(m_Thread, NULL);
        if(iRet!=0){
                std::cout<<"pthread_jon returns error\n";
        }

        close(m_iSockPairFd[0]);
        close(m_iSockPairFd[1]);

	LOG("log file is closing");
	LOG_CLOSE();

}


int CGossipComm::Run(){


        int iStatus = ParseConfig();
        if(iStatus != 0 ){
		std::cout<<"Error in parsing config file\n";
                return -1;
	}

	MakeDirs();


	iStatus = LoadPolicyProfile();
        if(iStatus != 0 ){
		std::cout<<"Error in loading policy profile file\n";
                return -1;
	}

        iStatus = CreateSocket();
        if(iStatus != 0 ){
		std::cout<<"Error in Creating Sockets\n";
                return -1;
	}

        iStatus = InitSocketPair();
        if(iStatus != 0 ){
		std::cout<<"Error in Creating Socket Pair\n";
                return -1;
	}

        /*Start Network Thread*/
        StartNetworkThread();

	//Logger startup
	if(StartLogger()!=0){
		std::cout<<"Error in Creating Logger\n";
                return -1;
	}

	LOG_OPEN(m_sLogFileName);


#if 0 //with multiple entities, we don't need it
	/*We copy the certificate file to keydir..*/
	CopyFileToKeyDir(m_sCertificate);

	/*Publish Certificate*/
	PublishCert(m_sCertificate);
#endif
	//
	PublishCACerts();

	/*We copy the trusted ca file to keydir..*/
	CopyFileToKeyDir(m_sTrustedCAPath);

	m_digisig.GetAllIssuer(m_sTrustedCAPath, m_issuer_set);

	m_bInit = true;

	if(m_bBootstrapEnable)
		FindBootStrapNode();

return 0;
}


void CGossipComm::StartNetworkThread(){

        int iRet = pthread_create(&m_Thread, NULL, CGossipComm::ThreadCallback, (void *)this);
        if (iRet){
		 DB_AND_LOG("ERROR in StartNetworkThread; return code from pthread_create() is "<<iRet);
                 exit(-1);
        }
}

void* CGossipComm::ThreadCallback(void *arg){
int iFd;

        CGossipComm& GossipInt = *(CGossipComm*)arg;

        iFd = GossipInt.GetSockFd();
        struct pollfd pollFd[2];

        pollFd[0].fd = iFd;
        pollFd[0].events = POLLIN;

        pollFd[1].fd = GossipInt.GetSocketPair()[0];
        pollFd[1].events = POLLIN;

	unsigned int iKeepaliveCount=0;
        while (1){
                int iStatus = poll(pollFd, 2, POLL_TIMEOUT );

                if( iStatus < 0){
                        DB_AND_LOG("poll failed: "<<iStatus);
                }

		if( 0 == iStatus){
			//timeout
			if(iKeepaliveCount==3){
				GossipInt.Logging("KEEPALIVE");
				iKeepaliveCount=0;
			}
			else{
				iKeepaliveCount++;
			}
                }
                else if(pollFd[0].revents & POLLIN){
			try{
				GossipInt.CheckEvent(pollFd[0].fd);
			}
			catch(std::exception& e){
				DB_AND_LOG("exception caught in CheckEvent.."<<e.what());
			}
                }
                else if(pollFd[1].revents & POLLIN){
			DB_AND_LOG("shutting down network thread");
			break;
                }
                else{
                        LOG("Error in pollfd: "<<iStatus);
                        break;
                }

        }


return NULL;
}

int CGossipComm::ParseConfig(){

        std::vector<std::string> vConf;
        char szBuf[MAXCHAR];
        std::string sTemp;


        std::ifstream fileHandler(m_sConfigFile.c_str());

        if(!fileHandler.is_open())
        {
                std::cout<<"Unable to open Input file"<<m_sConfigFile<<std::endl;
                return -1;
        }

        do{
                bzero(szBuf,MAXCHAR);
                fileHandler.getline(szBuf,MAXCHAR);
                vConf.push_back(szBuf);
        }while(!fileHandler.eof());
        fileHandler.close();

	std::cout<<"Parsing started\n";

        //Now, parse config file
        sTemp = GetString(vConf, "gossip_ip") ;
        if(sTemp.empty()){
		std::cout<<"gossip_ip is not found\n";
                return -1;
	}
        StrTrim(sTemp);
	if(sTemp == "localhost")
                sTemp="127.0.0.1";
        m_uGossipIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

        sTemp = GetString(vConf, "gossip_port") ;
        if(sTemp.empty()) {
		std::cout<<"gossip_port is not found\n";
                return -1;
	}
        m_uGossipPort = atoi(sTemp.c_str());


	//client
        sTemp = GetString(vConf, "client_ip") ;
        if(sTemp.empty()){
		std::cout<<"client_ip is not found\n";
                return -1;
	}
        StrTrim(sTemp);
        if(sTemp == "localhost")
                sTemp="127.0.0.1";
        m_uClientIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

        sTemp = GetString(vConf, "client_port") ;
        if(sTemp.empty()) {
		std::cout<<"client_port is not found\n";
                return -1;
	}
        m_uClientPort = atoi(sTemp.c_str());

#if 0
	sTemp = GetString(vConf, "client_private_key") ;
        if(sTemp.empty())
                return -1;
        StrTrim(sTemp);
        m_sPrivateKey = sTemp;
#endif

	sTemp = GetString(vConf, "test_certificate") ;
        StrTrim(sTemp);
        m_sTestCertificate = sTemp;

#if 0
        sTemp = GetString(vConf, "digital_sign") ;
        if(sTemp.empty()){
		std::cout<<"client_ip is not found");
                return -1;
	}
        StrTrim(sTemp);
	if(sTemp == "yes")
		m_bSignEnable = true;
	else
		m_bSignEnable = false;
#endif

        sTemp = GetString(vConf, "client_name") ;
        if(sTemp.empty()){
		std::cout<<"client name is not found\n";
                return -1;
	}
        StrTrim(sTemp);

	m_sClientName = sTemp;
	std::stringstream ss;
	ss<<sTemp;
	ss >> m_ClientName;

        sTemp = GetString(vConf, "certificate_cache_dir") ;
        if(sTemp.empty()){
		std::cout<<"certificate_cache_dir is not found\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sKeyDir = sTemp;

#if 0
        sTemp = GetString(vConf, "client_certificate") ;
        if(sTemp.empty())
               return -1;
        StrTrim(sTemp);
        m_sCertificate = MakeAbsPath(sTemp);
	//MakeAbsPath(sTemp);
#endif

	//find all certificates of CA need to publish ..if any..
	while(1){

		sTemp = GetString(vConf, "cacert_") ;
		if(sTemp.empty())
			break;
		StrTrim(sTemp);
		m_vCAs.push_back(MakeAbsPath(sTemp));
	}

        sTemp = GetString(vConf, "trusted_ca") ;
        if(sTemp.empty()){
		std::cout<<"truested_ca is not found\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sTrustedCAPath = MakeAbsPath(sTemp);


        sTemp = GetString(vConf, "trusted_ca_name") ;
        if(sTemp.empty()){
		std::cout<<"truested_ca_name is not found\n";
               return -1;
	}
        StrTrim(sTemp);
	m_sTrustedCAName = sTemp;

        sTemp = GetString(vConf, "tcp_gossip") ;
        if(sTemp.empty()){
		std::cout<<"tcp_gossip is not found\n";
               return -1;
	}
        StrTrim(sTemp);
	if(sTemp == "yes")
		m_iTransportProtocol = USE_TCP;
	else
		m_iTransportProtocol = USE_UDP;

        sTemp = GetString(vConf, "command_store") ;
        if(sTemp.empty()){
		std::cout<<"command_store is not found\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sCommandRepository = sTemp;

        sTemp = GetString(vConf, "log_file") ;
        if(sTemp.empty()){
		std::cout<<"log_file is not found\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sLogFileName = sTemp;

        sTemp = GetString(vConf, "subscription_file") ;
        if(sTemp.empty()){
		std::cout<<"subscription_file is not found\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sSubscriptionFileName = sTemp;
	if(false == LoadSubscription() ){
		std::cout<<"LoadSubscription failed\n";
		return -1;
	}

        sTemp = GetString(vConf, "command_file_name") ;
        if(sTemp.empty()){
		std::cout<<"command_file_name is not present\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sCommandFileName = sTemp;

        //client
        sTemp = GetString(vConf, "bootstrap_server_ip") ;
        if(sTemp.empty()){
		std::cout<<"bootstrap_server_ip is not present\n";
                return -1;
	}
        StrTrim(sTemp);
        m_iBootServerIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

        sTemp = GetString(vConf, "bootstrap_server_port") ;
        if(sTemp.empty()){
		std::cout<<"bootstrap_server_port is not present\n";
                return -1;
	}
        m_iBootServerPort = atoi(sTemp.c_str());

        sTemp = GetString(vConf, "bootstrap_enable") ;
        if(sTemp.empty()){
		std::cout<<"bootstrap_enable is not present\n";
                return -1;
	}
        StrTrim(sTemp);
        if(sTemp == "yes")
                m_bBootstrapEnable = true;
        else
                m_bBootstrapEnable = false;


	//default user id
        sTemp = GetString(vConf, "default_user") ;
        if(sTemp.empty()){
		std::cout<<"default_user is not present\n";
               return -1;
	}
        StrTrim(sTemp);
        m_sDefaultUserId = sTemp;
	DB("default_id="<<m_sDefaultUserId);
        //find all user ids
        while(1){
		std::string sLeft;
                sTemp = GetTwoSidedString(vConf, "user_id_",sLeft) ;
                if(sTemp.empty())
                        break;
                StrTrim(sTemp);
		m_UserId.insert(std::pair<std::string, std::string>(sLeft,sTemp));
        }

        sTemp = GetString(vConf, "client_certificate_dir") ;
        if(sTemp.empty()){
		std::cout<<"client_certificate_dir is not present\n";
                return -1;
	}
        StrTrim(sTemp);
        m_sClientCertDir = MakeAbsPath(sTemp);

        sTemp = GetString(vConf, "policy_profile") ;
        if(sTemp.empty()){
		std::cout<<"policy_profile is not present\n";
		return -1;
	}
        StrTrim(sTemp);
        m_sPolicyProfile = MakeAbsPath(sTemp);

        //Logserver
        sTemp = GetString(vConf, "log_server_ip") ;
        if(sTemp.empty()){
                std::cout<<"log_server_ip is not present\n";
                return -1;
        }
        StrTrim(sTemp);
        m_iLogServerIP = ntohl( inet_addr(sTemp.c_str()) ); //keep in host byte order

        sTemp = GetString(vConf, "log_server_port") ;
        if(sTemp.empty()){
                std::cout<<"log_server_port is not present\n";
                return -1;
        }
        m_iLogServerPort = atoi(sTemp.c_str());


        sTemp = GetString(vConf, "flooding_attack") ;
        if(sTemp.empty()){
                std::cout<<"flooding_attack is not present\n";
        }
	StrTrim(sTemp);
        if(sTemp == "yes")
                m_bFloodingEnable = true;
        else
                m_bFloodingEnable = false;


        sTemp = GetString(vConf, "non-co-operation_attack") ;
        if(sTemp.empty()){
                std::cout<<"non-co-operation_attack is not present\n";
        }
        StrTrim(sTemp);
        if(sTemp == "yes")
                m_bNonCooperationEnable = true;
        else
                m_bNonCooperationEnable = false;

       sTemp = GetString(vConf, "critical_level") ;
       if(sTemp.empty()){
                std::cout<<"critical_level is not present\n";
                //return -1;
        }
	else{
		m_sCricLevel = sTemp;
	}



	DB("Parse COmpleted");

return 0;
}

void CGossipComm::CheckEvent(int iFd){

int iReceived;
char szBuf[MAXCHAR];
bool bRet;
std::string sFileName;


	bzero(szBuf,MAXCHAR);
        iReceived = recv(iFd, (void*)szBuf, MAXCHAR, 0 );
	
	if(0==iReceived){
		DB_AND_LOG("Peer is down: shutting down the terminal: "<<iReceived);
		//add clean up codes	//should we send signals?
		raise(SIGINT);
		return ;
	}
	DB_AND_LOG("Received: "<<iReceived);

        if(iReceived<0) {
                DB_AND_LOG("Failed to refceived data!: "<<iReceived);
		return;
        }
        else{
               int iRead = 0;
               while(iRead<iReceived){

                        const char* pBuf = szBuf;
                        pBuf+=iRead;

                        EventMessage msg;
                        std::string sBuf(pBuf,iReceived-iRead);
                        iRead += msg.deserialize(sBuf);
                        switch(msg.GetFileType()){

                                case  FILE_OTHER:
				{
					DB_AND_LOG("Notification arrived");
					//need to read file for getting the command..

					DB_AND_LOG("File completed:"<<msg.GetFileName());
					std::cout<<"File received:"<<msg.GetFileName()<<std::endl;
					sFileName = CopyFileToKeyDir(msg.GetFileName());
					m_recv_file_types.insert(std::pair<std::string, uint8_t>(sFileName, FILE_OTHER));

					if(m_bValidateCert){
						//m_vCert.push_back(sFileName);
						m_vCert.push_back(msg.GetFileName());
					}

					bRet = CollectCerts(sFileName);
					break;
				}
				case  FILE_GOSSIP:
			        {

					std::string sSubject;
					m_digisig.ParseMIME(msg.GetFileName(), SUBJECT, sSubject);
					std::string sLog="BEFORE_VALIDATION: gossip id= "+sSubject;
					Logging(sLog);

					if(msg.GetSearchID() == LOG_ONLY_TRUE){
						DB("Only logging..return");
						return;
					}

					struct stat fStatFrom;
					stat( msg.GetFileName().c_str(), &fStatFrom );
					if(fStatFrom.st_size == 0){
						DB_AND_LOG("0 size file arrived!!!");
						//break;
						return;
					}
					std::cout<<"Gossip File received:"<<msg.GetFileName()<<std::endl;
					DB_AND_LOG("command File name:"<<msg.GetFileName());
					sFileName = CopyFileToCommandDir(msg.GetFileName());

					//there is a risk that, if two file name is same? gossip
					m_recv_file_types.insert(std::pair<std::string, uint8_t>(sFileName, FILE_GOSSIP));


					if(msg.GetRelatedModule() == MODULE_MULTICAST){
						DB_AND_LOG("Notification arrived from multicast module");
						std::cout<<"Notification arrived from multicast module\n";
						std::pair<RcvdFileMap::iterator, bool> bCheckRet = m_recv_module_types.insert(std::pair<std::string, uint8_t>(MakeAbsPath(sFileName), MODULE_MULTICAST));
						if(bCheckRet.second==false){
							DB_AND_LOG("Insertion failed in m_recv_module_types Map");
						}
					}
					else if(msg.GetRelatedModule() == MODULE_GOSSIP){
						DB_AND_LOG("Notification arrived from gossip module");
						std::cout<<"Notification arrived from gossip module\n";
						std::pair<RcvdFileMap::iterator, bool> bCheckRet = m_recv_module_types.insert(std::pair<std::string, uint8_t>(MakeAbsPath(sFileName), MODULE_GOSSIP));
						if(bCheckRet.second==false){
							DB_AND_LOG("Insertion failed in m_recv_module_types Map");
						}
					}
					else{
						DB_AND_LOG("Notification arrived from Unknown module");
						return;
					}


					bRet = VerifySignature(sFileName);
					break;
				}
				case FILE_BOOTSTRAP:
				{
					DB_AND_LOG("Bootstrap file received: "<<msg.GetFileName());
					sFileName = CopyFileToCommandDir(msg.GetFileName());
					DB("filename before insearting in the map: "<<sFileName);
					m_recv_file_types.insert(std::pair<std::string, uint8_t>(sFileName, FILE_BOOTSTRAP));
					bRet = VerifySignature(sFileName);
					//BootStrapKadReq(0,0);
					break;
				}
				case FILE_ZERO_CONTACT:
				{
					DB_AND_LOG("Zero Size Contacts!!!");
					if(m_bBootstrapEnable && m_vBootAddrPort.size() != 0){
						AddrPortPair p(m_vBootAddrPort[GetRandomIndex()]);
						BootStrapKadReq(p.first,p.second);
					}
					
					//break;
					return;
				}
				case FILE_ALL_CONTACTS:
                                {
                                        DB_AND_LOG("All Size Contacts!!!");
					//logging
					std::string sContact;
					if(msg.GetReasonString().size()>1200)
						sContact = std::string(msg.GetReasonString(), 1200);
					else
						sContact = msg.GetReasonString();

					std::string sLog="All Contacts = "+sContact;
					try{
						Logging(sLog);
					}catch(std::exception& e){
						//i found terminate called after throwing an instance of 'std::length_error'
						  //what():  basic_string::_S_create

						DB_AND_LOG("Caught exception duging sendng log "<<e.what());
					}

                                        //break;
                                        return;
                                }

				default:
					std::cout<<"Something else arrived!"<<std::endl;
					LOG("Unknown event arrived");
					break;
			}
		}
        }

	if(bRet){
		std::cout<<"Sig OK\n";
		LOG("Sig OK");
	}
	else{
		std::cout<<"Sig NOT OK\n";
		LOG("Sig NOT OK");
	}

}

int CGossipComm::SendCommand(){

	return SendCommand(m_sCommandFileName);
}

int CGossipComm::SendCommand(const std::string& sCommandFile){
	
	if(sCommandFile.empty())
		return -1;

CDigiSig obj;
static unsigned int iFileNo = 0;
char szFileName[100]={0};
	iFileNo++;

	/*We will follow the convension of the new command file with signature and smime header; 
	we use the same fle name as the command file name and append a . and a number*/

	sprintf(szFileName,"%u",iFileNo);
	std::string sNewFileName(sCommandFile);
	sNewFileName+=".";
	sNewFileName+=std::string(szFileName, strlen(szFileName));

	/*Now, generate the smime file*/
	/*We put a random uuid as an unique transaction in subject field*/
        boost::uuids::uuid Tag = boost::uuids::random_generator()();
        std::stringstream ss;
	ss<<Tag;
	std::string sSub;
	ss>>sSub;
	DB("subject="<<sSub);


	bool bSendCert = (m_iTransportProtocol == USE_TCP) ? false : true;

	//int iRet = obj.SignSMIME(m_sCertificate, m_sPrivateKey, sCommandFile, sNewFileName, "All", m_sClientName, sSub, bSendCert);
	std::string& sFrom = m_UserId[m_sDefaultUserId];
	const std::string& sCert = m_sClientCertDir+"/"+sFrom+".pem";
	const std::string& sPriv = m_sClientCertDir+"/privkey_"+sFrom+".pem";
//test
#if 0 //only to test the spoofing runlevel..
	sFrom[sFrom.size()-1]='7';
#endif
	DB("default id = "<<sFrom);
	DB("cert = "<<sCert);
	DB("priv = "<<sPriv);

	if(!m_bFloodingEnable){
		int iRet = obj.SignSMIME(sCert, sPriv, sCommandFile, sNewFileName, "All", sFrom, sSub,m_sCricLevel, bSendCert);
		if(iRet!=0){
			std::cout<<"signature failed\n";
			LOG("Signature failed");
			return -1;
		}

		SendToGossip(MakeAbsPath(sNewFileName), MODULE_GOSSIP_AND_MULTICAST);
	}
	else{
		SendToGossip(MakeAbsPath(sCommandFile), MODULE_GOSSIP_AND_MULTICAST);
	}
	DB_AND_LOG("Gossip Command Sent: "<<sSub);

return 0;
}

int CGossipComm::SendToGossip(const std::string& sFileName, uint8_t module){

	EventMessage msg;
        msg.SetReasonCode(FILE_GOSSIP_REQ);
        msg.SetFileType(FILE_GOSSIP);
        msg.SetFileName(sFileName);
        //we pass tcp or udp gossiop using this field
        msg.SetStatus(m_iTransportProtocol);
        msg.SetRelatedModule(module);

        std::string sBuf;
        msg.serialize(sBuf);

        int iRet = ::send(m_iSockFd, (void*)sBuf.c_str(), sBuf.size(), 0);
        if(iRet < 0)
                DB_AND_LOG("Send command failed\n");


return 0;
}

int CGossipComm::PublishCert(const std::string& sCert){

        int iRet=-1;
        int iLen = -1;

        EventMessage msg;
        msg.SetReasonCode(FILE_PUBLISH_REQ);
        msg.SetFileType(FILE_OTHER);
        msg.SetFileName(sCert);

        std::string sBuf;
        msg.serialize(sBuf);

        iRet = ::send(m_iSockFd, (void*)sBuf.c_str(), sBuf.size(), MSG_DONTWAIT);

        if(iRet < 0)
                DB_AND_LOG("Send command failed\n");

	DB("send: "<<iRet<<std::endl);

return 0;
}

int CGossipComm::FindFile(const std::string& sFileName){

        int iRet=-1;

	EventMessage msg;
	msg.SetReasonCode(FILE_FIND_REQ);
	msg.SetFileType(FILE_OTHER);
        msg.SetFileName(sFileName);

        std::string sBuf;
        msg.serialize(sBuf);

        iRet = ::send(m_iSockFd, (void*)sBuf.c_str(), sBuf.size(), 0);

        if(iRet < 0){
                std::cout<<"Send command failed\n";
		LOG("sending command failed");
	}


return 0;
}


int CGossipComm::ReportFeedback(const std::string& sFileName, bool bSuccess){

        int iRet=-1;

        EventMessage msg;
	if(bSuccess)
		msg.SetStatus(EVT_SUCCESS);
	else
		msg.SetStatus(EVT_ERROR);

        msg.SetReasonCode(FILE_REPORT_CERT_REQ);
        msg.SetFileType(FILE_OTHER);
        msg.SetFileName(sFileName);

        std::string sBuf;
        msg.serialize(sBuf);

        iRet = ::send(m_iSockFd, (void*)sBuf.c_str(), sBuf.size(), 0);

        if(iRet < 0){
                std::cout<<"Send command failed\n";
                LOG("sending command failed");
        }


return 0;
}


bool CGossipComm::VerifySignature(const std::string& sRcvdFile){

CDigiSig obj;
bool bRet = false;

        if(m_bSignEnable){

		//get the from field 
		std::string sFrom;
		obj.ParseMIME(sRcvdFile, FROM, sFrom);
		DB("sFrom="<<sFrom);

		if(sFrom.empty()){
                        DB_AND_LOG("Failed to parse From ");
			return false;
		}

#if 0
		if(sFrom == m_sClientName){
                        DB_AND_LOG("Self received ");
                        return true;
                }
#endif
		if(CheckSelfReceived(sFrom)){
                        DB_AND_LOG("Self received ");
                        return true;
		}

                std::string sFileName = sFrom + std::string(".pem");
                std::string sPath = GetPath(sFileName);//m_sKeyDir + std::string("/") + sFileName;

                //should we check the map?
                m_packet.insert(std::pair<std::string, std::string>(sFileName, sRcvdFile));
		bool bAttachedCert = (m_iTransportProtocol == USE_TCP) ? true:false;

		if(bAttachedCert){
			m_digisig.GetSenderCertificate(sRcvdFile, sPath);

			//is sender trying to fake its runlevel?
			bRet = CheckSenderIntigrity(sRcvdFile,sPath);
			if(!bRet){
				DB_AND_LOG("Sender is trying to spoof!!!");
				return false;
			}
		}

                if( !FileExistsInKeyDir(sFileName)){
                        DB("File doesn't exist:"<<sFileName<<std::endl);
                        FindFile(sFileName);

                        //request a file search...
                        bRet = false;
                }
                else{
                        //file exists..but is it signed by our m_sTrustedCAName?
                        bRet = CollectCerts(sFileName);
                }

        }

return bRet;
}


bool CGossipComm::IsSignedByHigherCA(const std::string& sFileToVerify){

	std::string sPath = GetPath(sFileToVerify);
	std::string sIssuer = GetCAName(sPath);
	if( m_sTrustedCAName == sIssuer ){
		return false;
	}
	//find the issuer in the set
	std::set<std::string>::const_iterator sIter;
	sIter = m_issuer_set.find(sIssuer);
	if(sIter != m_issuer_set.end()){
		DB_AND_LOG("Cert is verified by higher CA"<<std::endl);
		return true;
	}
	else{
		return false;
	}

return false;
}

bool CGossipComm::CollectCerts(std::string& sFileName){

bool bRet = false;
	//make path
	std::string sPath = GetPath(sFileName);//m_sKeyDir + std::string("/") + sFileName;

	//file exists..but is it signed by our m_sTrustedCAName?
	//we first check it against the chain, becase it may happen that, the packet is certified by the higher CA where we trust the lower CA in CA chain. If we trust a lower CA, that means, we alreay have the chain of certificates, and hence we should check whether the certificate can be verified by the chain or not.


	if(IsSignedByHigherCA(sFileName)){
		bRet = ReverseVerification(false);
		return bRet;
	}

	for(int i=0;i<MAX_CA_CHAIN; i++){

		std::string sIssuer = GetCAName(sPath);
		DB("Issuer="<<sIssuer<<std::endl);
		//we should delete the file if it is not a proper certificate file
		if(sIssuer.empty()){
			DB_AND_LOG("File is not an appropriate certificate; deleting :"<<sFileName);
			DeleteFileFromCache(sPath);
			return false;
		}

		std::string sNewFileName = sIssuer + std::string(".pem");

		//should we check the return value?
		DB("NewFileName="<<sNewFileName<<std::endl);

		/*Need to fix later: it would make problem when another command will come in the middle of one command validation..*/

		std::pair< std::map<std::string, std::string>::iterator, bool> mRet;
		mRet = m_cert.insert(std::pair<std::string, std::string>(sNewFileName,sFileName));
		if(mRet.second == false){
			DB_AND_LOG("CA Name alreay exists in the map, erasing and insearting the new value\n");
			std::string& s = mRet.first->second;
			DB("old:"<<s);
			s = sFileName;
			DB("New:"<<s);
		}
		

		DB("size of map "<<m_cert.size()<<std::endl);

		if( m_sTrustedCAName == sIssuer ){
			DB_AND_LOG("File is signed by the trusted CA\n");
			//start reverse verification
			//bRet = obj.Verify(sPath, sCommand, cPack.m_szSign, cPack.m_iSignLen);
			bRet = ReverseVerification();
			break;
		}
		else{
			//certificate is not signed by our trusted CA
			if(FileExistsInKeyDir(sNewFileName) ){
				//m_cert.insert(std::pair<std::string, std::string>(sNewFileName,sFileName));
				sPath = GetPath(sNewFileName);//m_sKeyDir + std::string("/") + sNewFileName;
				sFileName = sNewFileName;
				continue;
			}
			else{

				//m_cert.insert(std::pair<std::string, std::string>(sNewFileName,sFileName));
				//start search for the unknow CA
				if( sNewFileName == sFileName ){
					DB_AND_LOG("self certificate loop\n");
					continue;
				}
				
				FindFile(sNewFileName);
				bRet = false;
				break;
			}
		}
	}//for loop

return bRet;
}

bool CGossipComm::VerifyCA(){

bool bRet = false;

	std::map<std::string, std::string>::const_iterator mIter;

	std::string sIssuer = m_sTrustedCAName+std::string(".pem");

	std::vector<std::string> vCA;
	std::string sFileToVerify;

	for(int i=0;i<MAX_CA_CHAIN;i++){

		mIter = m_cert.find(sIssuer);
		if(mIter == m_cert.end()){
			DB("certificate name not found in map!!\n");	
			break;
		}
		//put all CA files in vector
		vCA.push_back(mIter->first);
		sFileToVerify = mIter->second;
	
		//verify

		const std::string& sSubordinate = mIter->second;
		DB_AND_LOG("Going to verify "<<sSubordinate<<" by "<<sIssuer<<std::endl);
		sIssuer = sSubordinate;
	}

	//concatenate and verify the chain..
	//std::string sAllFile=GetPath("ca_chain.pem");
	std::string sAllFile=m_sTrustedCAPath;
	wxString wxFile1(sAllFile.c_str(),wxConvUTF8);
	if(!wxFileExists(wxFile1)){
		FILE* fp = fopen(sAllFile.c_str(), "w");
		if(!fp){
			std::cout<<"CA chain file can't be written"<<std::endl;
			LOG("CA chain file can't be written");
			return false;
		}
		fclose(fp);
	}

	for(int i=0;i<vCA.size(); i++){
		const std::string& sFile2 = GetPath(vCA[i]);
		DB("CA: "<<sFile2<<std::endl);
		wxString wxFile2(sFile2.c_str(), wxConvUTF8);

		int iFind = vCA[i].find(".pem");
		if(iFind != std::string::npos){
			std::string str = vCA[i].substr(0, iFind);
			//std::cout<<"without pem:"<<str<<std::endl;
			if(m_issuer_set.find(str) != m_issuer_set.end() ){
				DB("already in the file");
				continue;
			}
		}

		if(m_ca_in_file.find(sFile2) != m_ca_in_file.end() ){
			//already written in file
			continue;
		}
		//the int is dummy, we are using map for fast lookup
		m_ca_in_file.insert(std::pair<std::string, int>(sFile2, 0));
		
		
		//if not already written, then write
		if(!wxConcatFiles(wxFile1, wxFile2, wxFile1)){
			DB_AND_LOG("file concatenation failed\n");
			return false;		
		}
	}


return true;
}

bool CGossipComm::ReverseVerification(bool bVerifyCA){

bool bRet = false;
	DB_AND_LOG("Start ReverseVerification\n");	

	if(bVerifyCA){
		bRet = VerifyCA();
		if(!bRet){
			DB_AND_LOG("Chain verification failed\n");	
			return false;
		}
	}

	/*
	Two cases can happen: Verification for a command or verification for a certificate file 

	*/

	//first case
	std::map<std::string, std::string>::iterator mPktIter;
	std::map<std::string, std::string>::const_iterator mCertIter;

	std::vector<std::string> vKeyToErase;

	for(mPktIter = m_packet.begin(); mPktIter!=m_packet.end(); mPktIter++){
		const std::string& sFile = mPktIter->first;

		std::string sPath = GetPath(sFile);
		if(sPath.empty()){
			DB_AND_LOG("sPath is empty for file: ");
			continue;
		}
		std::string sIssuer = GetCAName(sPath);
		sIssuer+=std::string(".pem");

		DB("from pkt: Issuer="<<sIssuer<<std::endl);
		DB("size of map "<<m_cert.size()<<std::endl);

		mCertIter = m_cert.find(sIssuer);
		if(mCertIter == m_cert.end() ){
			DB("file issuer is not present in certificate chain???\n");
		//	return false;
		}

		//all cert in chain are already verified...

		const std::string& sFileToVerify = mPktIter->second;
		//std::string sPath = GetPath(sFile);//m_sKeyDir + std::string("/") + sIssuer;

		std::string sOutFile(OUTPUT_FILE);//need to make it configurable..
		std::string sAllFile=m_sTrustedCAPath;

		bRet = m_digisig.VerifySMIME(sFileToVerify, sAllFile, sOutFile, sPath );

		//if verified, we should remove it from the map, isn't it?
		if(bRet==1){

			uint8_t file_type = CheckFileType(sFileToVerify);
			
			if(file_type == FILE_GOSSIP){
				GossipAndExecute(MakeAbsPath(sFileToVerify), sOutFile);
			}
			else if(file_type == FILE_BOOTSTRAP){
				DB_AND_LOG("bootstrap file arrived!");
				GetKadBootIPandPort(sOutFile);
				//AddrPortPair p(m_vBootAddrPort[0]);
				AddrPortPair p(m_vBootAddrPort[GetRandomIndex()]);
				BootStrapKadReq(p.first,p.second);
			}
			DB_AND_LOG("Pending pkt is verified, erasig it from map\n");
			//break;
			DB_AND_LOG("Going to erase file from the map: "<<sFile);
			//m_packet.erase(sFile);
			//m_packet.erase(mPktIter);
			vKeyToErase.push_back(sFile);
			DB_AND_LOG("After erase file from the map: "<<sFile);
		}
		else{
			DB_AND_LOG("Pending pkt is NOT verified, erasig it from map\n");
			//m_packet.erase(sFile);
			//m_packet.erase(mPktIter);
			vKeyToErase.push_back(sFile);
		}

	}
	for(int i=0;i<vKeyToErase.size();++i){
		mPktIter=m_packet.find(vKeyToErase[i]);
		if(mPktIter!=m_packet.end()){
			m_packet.erase(mPktIter);
		}
	}


	//this is for index pollusion attack robustness..i remember....causing a crash in planetlab...i guess..lets not use it right now..

#if 0
	//second case
	if(m_bValidateCert){
		std::deque<std::string>::iterator vIter;
		DB_AND_LOG("Size of m_vCert=: "<<m_vCert.size());
		for(vIter=m_vCert.begin(); vIter!=m_vCert.end(); vIter++){
			std::string& sFile = *vIter;
			if(sFile.empty())
				continue;
			DB_AND_LOG("Going to verify certificate: "<<sFile);
			bool bResult = m_digisig.VerifyByCA(m_sTrustedCAPath,sFile );
			if(bResult){
				DB_AND_LOG("certificate verification succesfull");
				ReportFeedback(sFile, true);
				sFile.clear();
				//m_vCert.erase(vIter);
			}
			else{
				//we should delete the file from the cache if the file fails to pass
				DB_AND_LOG("Deleting the file from cache since it didn't pass");
				DeleteFileFromCache(sFile);
			}
		}
	}
#endif

return bRet;
}

void CGossipComm::GossipAndExecute(const std::string& sFileName, const std::string& sFileToExecute){

	LOG("Gossip again: "<<sFileName);
	//instead of hardcode module, we should derive it from the received module. If the command received from gossip module, we should do both gossip and multicast, if the command received from multicast module, we should do only gossip.

	if(!m_bNonCooperationEnable){
		uint8_t module_type = CheckModuleType(sFileName);
	
		switch(module_type){
			case  MODULE_MULTICAST:
					SendToGossip(MakeAbsPath(sFileName), MODULE_GOSSIP);
					break;
			case MODULE_GOSSIP:
					SendToGossip(MakeAbsPath(sFileName), MODULE_GOSSIP);
					//SendToGossip(MakeAbsPath(sFileName), MODULE_GOSSIP_AND_MULTICAST);
					break;
			default:
				DB_AND_LOG("Couldn't find module name for the file name in CHeckModuleType: "<<sFileName);
				break;
		}
	}

	//Execute the command
	std::string sFrom;
	m_digisig.ParseMIME(sFileName, FROM, sFrom);
	int iRunLevel = ExtractRunLevel(sFrom);
	std::map<int,std::string>::const_iterator mIter;
	mIter = m_ProfileMap.find(iRunLevel);
	if(mIter==m_ProfileMap.end()){
		DB_AND_LOG("Runlevel not found in the map: "<<iRunLevel);
		return;
	}
	const std::string& sProfile = mIter->second;
	DB("Profile="<<sProfile);

	std::string sMessageID;
	if(IsSubscribed(sFrom) && IsFirstTimeReceived(sFileName, sMessageID)){
		LOG("Command from subscribed service: "<<sFrom);
		LOG("Executing the file: "<<sFileToExecute);
		//Execution
		std::string sCommand("sh ");
		sCommand+=sFileToExecute;
//added to show parameters:
		sCommand+=" ";
		sCommand+=sProfile;
//

#if 0//we don't execute it for planetlab experiments..
		system(sCommand.c_str());
#endif

		//logging
		std::string sLog="Gossip FROM= "+sFrom+";"+"ID= "+sMessageID;
		Logging(sLog);
	}

}

bool CGossipComm::IsFirstTimeReceived(const std::string& sFile, std::string& sSubject){
	//std::string sSubject;
	m_digisig.ParseMIME(sFile, SUBJECT, sSubject);

	if(sSubject.empty()){
		LOG("Why subject is empty??: "<<sFile);
		return false;
	}

	if(m_MessageIDMap.find(sSubject) == m_MessageIDMap.end()){
		LOG("Message received for the first time: "<<sSubject);
		m_MessageIDMap.insert(std::pair<std::string, int>(sSubject,0));
		return true;
	}
	else
		LOG("Message received is already in the map: "<<sSubject);

return false;
}

bool CGossipComm::FileExistsInKeyDir(const std::string& sFileName){

	std::string sPath = GetPath(sFileName);//m_sKeyDir + std::string("/") + sFileName;
	wxString wxStr(sPath.c_str(), wxConvUTF8);

	if(!wxFileExists(wxStr))
		return false;

return true;
}


int CGossipComm::CreateSocket(){

struct sockaddr_in kAddr;

int iRet;

        m_iSockFd = ::socket(AF_INET, SOCK_STREAM, 0);
        if (m_iSockFd == -1){
                std::cout<<"Could not create socket."<<std::endl;
                return -1;
        }

	//reuse(m_iSockFd);
	//nonblock(m_iSockFd);
	nodelay(m_iSockFd);

        bzero((char *) &kAddr, sizeof(kAddr));
        kAddr.sin_family = AF_INET;
        kAddr.sin_addr.s_addr = INADDR_ANY;
        kAddr.sin_port = htons(m_uClientPort);

        iRet = bind(m_iSockFd, (struct sockaddr *) &kAddr, sizeof(kAddr));
        if (iRet < 0) {
                std::cout<<"ERROR on binding"<<std::endl;
                return -1;
        }

	//connect
	struct sockaddr_in destAddr;
        socklen_t addrLen = sizeof(destAddr);
        bzero(&destAddr, sizeof(destAddr));
        destAddr.sin_family = AF_INET;
        destAddr.sin_addr.s_addr = htonl(m_uGossipIP);
        destAddr.sin_port = htons(m_uGossipPort);

	iRet = ::connect(m_iSockFd, (struct sockaddr *) &destAddr, addrLen) ;
        if (iRet < 0) {
                std::cout<<"ERROR on connect"<<std::endl;
                return -1;
        }
	else
		DB_AND_LOG("connect successfully\n");


return 0;
}



std::string CGossipComm::GetString(std::vector<std::string>& vConf, const std::string& sLine){

        std::vector<std::string>::iterator vIter;
        std::string sTemp;
        int iPos=-1;

        for(vIter=vConf.begin(); vIter!=vConf.end(); vIter++){
                sTemp = *vIter;
                iPos = sTemp.find(sLine);
                if(iPos == std::string::npos)
                        continue;

		vConf.erase(vIter);

                iPos = sTemp.find("=");
                return sTemp.substr(iPos+1);
        }

        return "";

}

std::string CGossipComm::GetTwoSidedString(std::vector<std::string>& vConf, const std::string& sLine, std::string& sLeft){

        std::vector<std::string>::iterator vIter;
        std::string sTemp;
        int iPos=-1;

        for(vIter=vConf.begin(); vIter!=vConf.end(); vIter++){
                sTemp = *vIter;
                iPos = sTemp.find(sLine);
                if(iPos == std::string::npos)
                        continue;

                vConf.erase(vIter);

                iPos = sTemp.find("=");
		sLeft = sTemp.substr(0,iPos);
		StrTrim(sLeft);

                return sTemp.substr(iPos+1);
        }

        return "";

}

void CGossipComm::StrTrim(std::string& str)const{

        std::string::size_type pos = str.find_last_not_of(' ');
        if(pos != std::string::npos) {
                str.erase(pos + 1);
                pos = str.find_first_not_of(' ');
                if(pos != std::string::npos)
                        str.erase(0, pos);
        }
        else
                str.erase(str.begin(), str.end());
}


std::string CGossipComm::CopyFileToKeyDir(const std::string & sFile){

        wxString src = wxString(sFile.c_str(), wxConvUTF8);

        size_t pos = sFile.find_last_of('/');

        std::string sFileName(sFile);

        if(pos != std::string::npos){
                sFileName = sFile.substr(pos+1);
        }

        std::string sDest = m_sKeyDir+"/"+sFileName;
        //std::cout<<"dest file: "<<sDest<<std::endl;

        wxString dest = wxString(sDest.c_str(), wxConvUTF8);

        if(wxCopyFile(src, dest)){
		//std::cout<<"copy successfull"<<std::endl;
	}

return sFileName;
}

std::string CGossipComm::CopyFileToCommandDir(const std::string & sFile){

        wxString src = wxString(sFile.c_str(), wxConvUTF8);

        size_t pos = sFile.find_last_of('/');

        std::string sFileName(sFile);

        if(pos != std::string::npos){
                sFileName = sFile.substr(pos+1);
        }

        std::string sDest = m_sCommandRepository+"/"+sFileName;
        //std::cout<<"dest file: "<<sDest<<std::endl;

        wxString dest = wxString(sDest.c_str(), wxConvUTF8);

        if(!wxCopyFile(src, dest)){
                std::cout<<"copy failed: "<<sFile<<" To: "<<sDest<<std::endl;
        }

return sDest;
}



std::string CGossipComm::MakeAbsPath(const std::string& sPath){

        /*First make it absolute*/
	std::string sCertificate;

        wxString wxStr(sPath.c_str(), wxConvUTF8);
        wxFileName wxFile(wxStr);
        if(!wxFile.IsAbsolute()){
                wxFile.MakeAbsolute();
                sCertificate = wxFile.GetLongPath().mb_str();
        }
        else{
                sCertificate = sPath;
	}

	//std::cout<<"Abs cert="<<sCertificate<<std::endl;

return sCertificate;
}

void CGossipComm::PublishCACerts(){

	for(int i=0;i<m_vCAs.size(); i++){
		CopyFileToKeyDir(m_vCAs[i]);
		PublishCert(m_vCAs[i]);
	}

}

std::string CGossipComm::GetCAName(const std::string& sCert){

CDigiSig obj;
std::string sName;
	obj.GetIssuerName(sCert, sName);
return sName;
}

std::string CGossipComm::GetCertificateName(const std::string& sCert){

CDigiSig obj;
std::string sName;
        obj.GetSubjectName(sCert, sName);
return sName;
}


std::string CGossipComm::GetPath(const std::string& sFileName){
//once I got a crash here...both gossip and send contact data was mixed..
	std::string sPath;
	try{
		sPath+=m_sKeyDir;
		sPath+=std::string("/");
		sPath+=sFileName;
	}catch(std::exception & e){
		sPath.clear();
		DB_AND_LOG("Exception caught in GetPath: "<<e.what());
	}
	//return  m_sKeyDir + std::string("/") + sFileName;
	return sPath;
}

bool CGossipComm::IsSubscribed(const std::string& sFrom){

	if(m_bSubscribedAll)
		return true;

	if( m_vSubscribed.end() != find(m_vSubscribed.begin(), m_vSubscribed.end(), sFrom) ){

		DB_AND_LOG("Subscribed : "<<sFrom);
		return true;
	}

	DB_AND_LOG("Not subscribed : "<<sFrom);

return false;
}

bool CGossipComm::LoadSubscription(){

        char szBuf[MAXCHAR];
        std::ifstream fileHandler(m_sSubscriptionFileName.c_str());

	m_vSubscribed.clear();

        if(!fileHandler.is_open())
        {
                DB_AND_LOG("Unable to open Input file"<<m_sConfigFile);
                return false;
        }

        do{
                bzero(szBuf,MAXCHAR);
                fileHandler.getline(szBuf,MAXCHAR);
		if( fileHandler.gcount() == 0 )
                        continue;

		/*We don't want to have the delimeter included in the string*/
		std::string sTemp(szBuf, fileHandler.gcount()-1);
		DB_AND_LOG("subscribed: "<<sTemp);

                m_vSubscribed.push_back(sTemp);
        }while(!fileHandler.eof());
        fileHandler.close();

	/*Does it subscribed to all?*/
	if( m_vSubscribed.end() != find(m_vSubscribed.begin(), m_vSubscribed.end(), "all") ){
		DB_AND_LOG("subscribed to all");
		m_bSubscribedAll = true;
	}
	else
		m_bSubscribedAll = false;

return true;
}

int CGossipComm::InitSocketPair(){

        int rc = socketpair( AF_UNIX, SOCK_STREAM, 0, m_iSockPairFd );
        if ( rc < 0 ) {
                printf("Cannot open socketpair");
                return -1;
        }

return 0;
}

void CGossipComm::ShutdownNetworkThread(){

	DB("Sendign Shutdown Command");
        int iStatus = ::send(m_iSockPairFd[1], "STOP", 4, 0);
        if(iStatus<0){
                DB_AND_LOG("Failed sending shutdown command in sockpair");
        }
}

void CGossipComm::FindBootStrapNode(){

        EventMessage msg;
        msg.SetReasonCode(FILE_BOOTSTRAP_SRV_REQ);
	msg.SetFileSize(m_iBootServerIP);
	msg.SetSearchID(m_iBootServerPort);

        std::string sBuf;
        msg.serialize(sBuf);

        int iRet = ::send(m_iSockFd, (void*)sBuf.c_str(), sBuf.size(), 0);
        if(iRet < 0)
                DB_AND_LOG("Send command failed\n");

}

void CGossipComm::BootStrapKadReq(int ip, int port){

        EventMessage msg;
        msg.SetReasonCode(FILE_BOOTSTRAP_KAD_REQ);
        msg.SetFileSize(ip);
        msg.SetSearchID(port);

        std::string sBuf;
        msg.serialize(sBuf);

        int iRet = ::send(m_iSockFd, (void*)sBuf.c_str(), sBuf.size(), 0);
        if(iRet < 0)
                DB_AND_LOG("Send command failed\n");

}

uint8_t CGossipComm::CheckFileType(const std::string& sFile){

	std::map<std::string, uint8_t>::iterator mIter;
	mIter = m_recv_file_types.find(sFile);
	if(mIter == m_recv_file_types.end()){
		DB("in check file type (not found): "<<FILE_UNKNOWN);
		return FILE_UNKNOWN;
	}

	//we shoud delete the entry in the map; becasue, later we may have entry with same file name
	m_recv_file_types.erase(mIter);

return mIter->second;
}

uint8_t CGossipComm::CheckModuleType(const std::string& sFile){

        std::map<std::string, uint8_t>::iterator mIter;
        mIter = m_recv_module_types.find(sFile);
        if(mIter == m_recv_module_types.end()){
                DB("in check module type (not found): "<<MODULE_UNKNOWN);
                return MODULE_UNKNOWN;
        }

	//we shoud delete the entry in the map; becasue, later we may have entry with same file name
	//m_recv_module_types.erase(mIter);

return mIter->second;
}


int CGossipComm::GetKadBootIPandPort(const std::string& sFile){


        char szBuf[MAXCHAR];
        std::ifstream fileHandler(sFile.c_str());

        if(!fileHandler.is_open())
        {
                DB_AND_LOG("Unable to open Input file"<<sFile);
                return -1;
        }

        do{
                bzero(szBuf,MAXCHAR);
                fileHandler.getline(szBuf,MAXCHAR);
                if( fileHandler.gcount() == 0 )
                        continue;

                /*We don't want to have the delimeter included in the string*/
                std::string sTemp(szBuf, fileHandler.gcount()-1);

		int _ip, _port;
		_ip=_port=0;
		GetAddrPortFromString(sTemp,_ip,_port);
		AddrPortPair p(_ip,_port);
		m_vBootAddrPort.push_back(p);
		
                DB_AND_LOG("received boot ip =  "<<_ip<<" port="<<_port);
		//break;

        }while(!fileHandler.eof());
        fileHandler.close();


return 0;
}

void CGossipComm::GetAddrPortFromString(const std::string& sLine, int& ip, int& port)const{

	size_t pos = sLine.find_first_of(":");
	if(pos==std::string::npos){
		DB_AND_LOG("format is not okey in node.txt file");
		return;
	}
	std::string sIP = sLine.substr(0,pos);
	std::string sPort = sLine.substr(pos+1);
	DB("IP="<<sIP);
	DB("Port="<<sPort);

	ip = ntohl( inet_addr(sIP.c_str()) ); //keep in host byte order
	port = atoi(sPort.c_str());
	DB("IP="<<ip);
	DB("Port="<<port);

}
bool CGossipComm::CheckSelfReceived(const std::string& sFrom)const{

	std::map<std::string, std::string>::const_iterator mIter;
	for(mIter=m_UserId.begin(); mIter!=m_UserId.end(); mIter++){
		const std::string& s = mIter->second;
		if(s == sFrom){
			DB("Self loop");
			return true;
		}
	}

return false;
}
int CGossipComm::ExtractRunLevel(const std::string& sFrom)const{


	size_t pos = sFrom.find("+");
	if(pos==std::string::npos){
		DB_AND_LOG("Run level not found in the from field: "<<sFrom);
		return -1;
	}
	std::string sRunLevel = sFrom.substr(pos+1);

return atoi(sRunLevel.c_str());
}

bool CGossipComm::CheckSenderIntigrity(const std::string& sRcvdFile, const std::string& sCert){

	std::string sFrom;
	m_digisig.ParseMIME(sRcvdFile, FROM, sFrom);
	int iRunLevel1 = ExtractRunLevel(sFrom);
	DB("chkIntigrity: Run Level1 is "<<iRunLevel1);
	
	std::string sName = GetCertificateName(sCert);
	DB("chkIntigrity: cert Name:  "<<sName);
	int iRunLevel2 = ExtractRunLevel(sName);
	DB("chkIntigrity: Run Level2 is "<<iRunLevel2);
	if(iRunLevel1==iRunLevel2){
		DB_AND_LOG("chkIntigrity Succeed!!");
		return true;
	}
	else
		DB_AND_LOG("chkIntigrity Failed!!");

return false;
}

int CGossipComm::LoadPolicyProfile(){

        std::ifstream fileHandler(m_sPolicyProfile.c_str());

        if(!fileHandler.is_open())
        {
                DB_AND_LOG("Unable to open Input file"<<m_sPolicyProfile);
                return -1;
        }

        do{

		int iRunLevel=-1;
		std::string sProfile;
		
		fileHandler>>iRunLevel;
		fileHandler>>sProfile;
	
		if(iRunLevel!=-1){
			m_ProfileMap.insert(std::pair<int,std::string>(iRunLevel,sProfile));
		}

        }while(!fileHandler.eof());
        fileHandler.close();

return 0;
}

void CGossipComm::DeleteFileFromCache(const std::string& sPath){

	wxString _wxPath = wxString(sPath.c_str(), wxConvUTF8);
	wxString _wxFile = wxFileNameFromPath(_wxPath);
	std::string sFile = std::string(_wxFile.mb_str());
	std::string sKeyPath = GetPath(sFile);
	DB("file name="<<sKeyPath);
	wxString _wxKeyPath = wxString(sKeyPath.c_str(), wxConvUTF8);
	if(!wxRemoveFile(_wxKeyPath)){
		DB_AND_LOG("Failed to remove file: "<<sKeyPath);
	}
}

void CGossipComm::MakeDirs(){

        //make m_sCommandRepository, m_sKeyDir on startup

        if(!wxDirExists(wxString(m_sKeyDir.c_str(), wxConvUTF8))){
                if(!wxMkdir(wxString(m_sKeyDir.c_str(), wxConvUTF8)) ){
                        std::cout<<"Falied to create Dir: "<<m_sKeyDir<<std::endl;
                }
        }
        if(!wxDirExists(wxString(m_sCommandRepository.c_str(), wxConvUTF8))){
                if(!wxMkdir(wxString(m_sCommandRepository.c_str(), wxConvUTF8)) ){
                        std::cout<<"Falied to create Dir: "<<m_sCommandRepository<<std::endl;
                }
        }

}


//Communication with log server..
int CGossipComm::SendToLogServer(const std::string& sLog){

        int iSent = SendLog(sLog);
        if(iSent<0){
		DB("SendLog return error");
                return -1;
	}

        struct pollfd pollFd[1];

        pollFd[0].fd = m_iLogSockFd;
        pollFd[0].events = POLLIN;

        int iCount=0;
        while (iCount<MAX_SEND){
                int iStatus = poll(pollFd, 1, LOG_POLL_TIMEOUT );
                if( iStatus < 0){
                        std::cout<<"poll failed in SendToLogServer"<<iStatus<<std::endl;
                        return -1;
                }
                else if( 0 == iStatus){
                        LOG("Timeout in poll");
                        iCount++;
                        SendLog(sLog);
                }
                else if(pollFd[0].revents & POLLIN){
                        ReceiveAck();
                        break;
                }
                else{
                        LOG("error in poll in SendToLogServer");
                        break;
                }

        }//while loop

return 0;
}

int CGossipComm::SendLog( const std::string& sSendLog){

        struct sockaddr_in FromAddr;
        socklen_t FromAddrLen=sizeof(FromAddr);

        bzero((char *) &FromAddr, sizeof(FromAddr));
        FromAddr.sin_family = AF_INET;
        FromAddr.sin_addr.s_addr = htonl(m_iLogServerIP);
        FromAddr.sin_port = htons(m_iLogServerPort);

	std::string sLog(GetGMTime());
	sLog+="#";

	char szBuf[100]={0};
        int iRet = gethostname(szBuf, 100);
	if(iRet==0){
		int iLen =  strlen(szBuf);
		sLog += std::string(szBuf, iLen);	
	}
	sLog+="#";
	sLog+=m_sClientName;
	sLog+="#";
	sLog+=sSendLog;

        int iSent = -1;
        iSent = sendto(m_iLogSockFd, sLog.c_str(), sLog.size(), 0, (struct sockaddr*)&FromAddr, FromAddrLen);
        if(iSent<0){
                std::cout<<"Error in sendning log to :"<<(m_iLogServerIP)<<":"<< m_iLogServerPort<<std::endl;;
        }
	LOG("sent to log server="<<sSendLog);

return iSent;
}

int CGossipComm::ReceiveAck(){

        struct sockaddr_in FromAddr;
        socklen_t FromAddrLen=sizeof(FromAddr);

        char szBuf[1500]={0};
        int iReceived = recvfrom(m_iLogSockFd, szBuf, 1500, 0, (struct sockaddr *) &FromAddr, &FromAddrLen);
        if(iReceived>0){
                uint32_t uIP = ntohl(FromAddr.sin_addr.s_addr);
                uint16_t uPort = ntohs(FromAddr.sin_port);
                std::string sLog(szBuf, iReceived);
                std::cout<<"Ack received : "<< (uIP)<<":"<< uPort<<"#"<<sLog<<std::endl;

        }
return 0;
}

int CGossipComm::CreateLogSocket(){

        struct sockaddr_in FromAddr;
        int iRet;

        m_iLogSockFd = ::socket(AF_INET, SOCK_DGRAM, 0);
        if (m_iLogSockFd == -1){
                LOG("Could not create tcp socket.");
                return -1;
        }

        //reuse(m_iDaemonSockFd);

        bzero((char *) &FromAddr, sizeof(FromAddr));
        FromAddr.sin_family = AF_INET;
        FromAddr.sin_addr.s_addr = htonl(INADDR_ANY);
        FromAddr.sin_port = htons(0);

        iRet = bind(m_iLogSockFd, (struct sockaddr *) &FromAddr, sizeof(FromAddr));
        if(iRet<0){
                LOG("ERROR on binding in tcp sock");
                return -1;
        }

return 0;
}


void CGossipComm::DestroyLogSocket()
{
        if(m_iLogSockFd != SOCKET_ERROR){
                ::close(m_iLogSockFd);
                m_iLogSockFd = SOCKET_ERROR;
        }
}



void CGossipComm::StartLoggerThread(){

        int iRet = pthread_create(&m_LoggerThread, NULL, CGossipComm::LoggerThreadCallback, (void *)this);
        if (iRet){
                 DB_AND_LOG("ERROR in StartLoggerThread; return code from pthread_create() is "<<iRet);
                 exit(-1);
        }
}

void* CGossipComm::LoggerThreadCallback(void *arg){

        CGossipComm& GossipInt = *(CGossipComm*)arg;

        struct pollfd pollFd[1];

        pollFd[0].fd = GossipInt.GetLogSocketPair()[0];
        pollFd[0].events = POLLIN;


        while (1){
                int iStatus = poll(pollFd, 1, POLL_TIMEOUT );
       
                if( iStatus < 0){
                        DB_AND_LOG("poll failed in LoggerThreadCallback: "<<iStatus);
			LOG("poll error="<<strerror(errno));
                        //return NULL;
			continue;
                }

                if( 0 == iStatus){
                        //timeout
                }
                else if(pollFd[0].revents & POLLIN){
			//received the string and send
                        //GossipInt.CheckEvent(pollFd[0].fd);
			char szBuffer[MAXCHAR]={0};
			int iReceived = recv(pollFd[0].fd, (void*)szBuffer, MAXCHAR, 0 );
			std::string sLog(szBuffer, iReceived);
			if(sLog == "STOP"){
				DB_AND_LOG("Closing log thread");
				break;
			}
			GossipInt.SendToLogServer(sLog);

                }
                else{
                        LOG("Error in pollfd: "<<iStatus);
                        break;
                }

        }

return NULL;
}

int CGossipComm::InitLoggerSocketPair(){

        int rc = socketpair( AF_UNIX, SOCK_STREAM, 0, m_iLoggerSockPairFd );
        if ( rc < 0 ) {
                printf("Cannot open socketpair");
                return -1;
        }

return 0;
}

int CGossipComm::StartLogger(){

	if( CreateLogSocket() != 0 ){
		std::cout<<"Failed to create Log socket\n";
		return -1;
	}

	if(InitLoggerSocketPair() != 0 ){
		std::cout<<"Failed to create Log socketet pair\n";
		return -1;
	}

	StartLoggerThread();
	m_bLoggerInit = true;

return 0;
}

int CGossipComm::StopLogger(){

	if(m_bLoggerInit){
		DB("Sendign Log Shutdown Command");
		int iStatus = ::send(m_iLoggerSockPairFd[1], "STOP", 4, 0);
		if(iStatus<0){
			DB_AND_LOG("Failed sending shutdown command in Logger sockpair");
		}

		int iRet = pthread_join(m_LoggerThread, NULL);
		if(iRet!=0){
			std::cout<<"pthread_jon returns error for log thread\n";
		}
	}

	DestroyLogSocket();

        close(m_iLoggerSockPairFd[0]);
        close(m_iLoggerSockPairFd[1]);

	m_bLoggerInit = false;

return 0;
}
std::string CGossipComm::GetGMTime(void)
{
        // mm/dd/yyyy HH:MM:SS
        // 1234567891123456789
        char buf[50];

        time_t tNow = time(NULL);
        //strftime(buf, sizeof(buf), "%Y-%m-%d %T", ::localtime(&tNow));
        strftime(buf, sizeof(buf), "%Y-%m-%d %T", ::gmtime(&tNow));
	//strcpy(buf, asctime(gmtime(&tNow)));


        return buf;
}

void CGossipComm::Logging(const std::string& sLog){

	if(m_bLoggerInit){
                DB("Sending Log: "<<sLog);
                int iStatus = ::send(m_iLoggerSockPairFd[1], sLog.c_str(), sLog.size(), 0);
                if(iStatus<0){
                        DB_AND_LOG("Failed sending log command in Logger sockpair");
                }
	}
}

int CGossipComm::GetRandomIndex(){

	int iLimit = m_vBootAddrPort.size();
	int random_integer = rand();

return random_integer%iLimit;
}
